//============================================================================
// App.cpp - Application class
//============================================================================
#define WINVER 0x0500
#define _WIN32_WINNT 0x0500
#include "Plugin/Assert.h"
#include "Plugin/ScriptContext.h"
#include "Resource/Resource.h"
#include "ScriptManager/ScriptManager.h"
#include "TA2Stream/TA2Stream.h"
#include "StandardLib/StandardLib.h"
#include "Socket/Socket.h"
#include "HTTPSocket/HTTPSocket.h"
#include "App.h"
#include <shlobj.h>
#include <time.h>
#include <sstream>
#include <algorithm>

CApp* CApp::sm_pInstance = NULL;
int CApp::ms_nVersionMajor = 0;
int CApp::ms_nVersionMinor = 0;

// Plugin function pointers
typedef size_t (*LPGetNumExportedFunctions)(int nVersion);
typedef void (*LPGetExportedFunction)(size_t nIndex, ScriptFunction& funcInfo);

//============================================================================
static const int s_nDefaultInputHeight = 64;
static const int s_nMinInputHeight = 24;
static const int s_nDividerHeight = 2;
static const int s_nMinWidth = 64;
static const int s_nMinOutputHeight = 40;
static const int s_nawsMinInterval = 1;

static const unsigned int s_nTickTimerID = 1;

static const char* s_szDefaultWindowTitle = "DruinkClient v2";
static const char* s_szConfigFilename = "DruinkClient.cfg";
static const char* s_szAppDataFolderName = "\\DruinkClient\\";
static const char* s_szExtraConfigFilename = "Settings.cfg";

static const char* s_szExtraConfigHeader =
	"# This file is modified by DruinkClient. Do not manually edit this\n"
	"# file while DruinkClient is running!";

#ifdef BUILD_X64
	static const char* s_szClientVersion = "DruinkClient Version %d.%d x64";
#else
	static const char* s_szClientVersion = "DruinkClient Version %d.%d x86";
#endif

//============================================================================

inline bool GetHexColour(const char* szTag, COLORREF& rColour)
{
	rColour = 0;
	for(int i=0, nShift=0; i<3; i++, szTag+=2)
	{
		if((szTag[1] >= '0') && (szTag[1] <= '9')) rColour |= (szTag[1]-'0') << nShift;
		else if((szTag[1] >= 'a') && (szTag[1] <= 'f')) rColour |= (szTag[1]-'a'+10) << nShift;
		else if((szTag[1] >= 'A') && (szTag[1] <= 'F')) rColour |= (szTag[1]-'A'+10) << nShift;
		else return false;
		nShift += 4;
		if((szTag[0] >= '0') && (szTag[0] <= '9')) rColour |= (szTag[0]-'0') << nShift;
		else if((szTag[0] >= 'a') && (szTag[0] <= 'f')) rColour |= (szTag[0]-'a'+10) << nShift;
		else if((szTag[0] >= 'A') && (szTag[0] <= 'F')) rColour |= (szTag[0]-'A'+10) << nShift;
		else return false;
		nShift += 4;
	}

	return true;
}

inline std::string RemoveWhitespace(const std::string& str)
{
	std::string strRet;
	strRet.reserve(str.size());
	for(size_t i=0; i<str.length(); ++i)
	{
		char ch = str[i];
		if(ch > ' ')
			strRet += ch;
	}
	return strRet;
}

inline std::string Win32ErrorToString(DWORD dwErrorCode)
{
	char* szMsgBuffer;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS, NULL, dwErrorCode, 0, (LPTSTR)&szMsgBuffer, 0, NULL);
	std::string strRet(szMsgBuffer);
	LocalFree(szMsgBuffer);
	return strRet;
}

static void DebugLog(const char* szFormat, ...)
{
#ifdef _DEBUG
	char szBuff[1024];
	va_list arg;
	va_start(arg, szFormat);
	_vsnprintf(szBuff, sizeof(szBuff), szFormat, arg);
	va_end(arg);
	OutputDebugString(szBuff);
#else
	UNREFERENCED_PARAMETER(szFormat);
#endif
}

static const char* telnetCmdToString(BYTE _opt)
{
	switch(_opt)
	{
	case 0: return "Transmit-Binary";
	case 1: return "Echo";
	case 2: return "Reconnection";
	case 3: return "Suppress-Go-Ahead";
	case 4: return "Approx-Message-Size-Negotiation";
	case 5: return "Status";
	case 6: return "Timing-Mark";
	case 7: return "Remote-Controlled-Trans-and-Echo";
	case 8: return "Output-Line-Width";
	case 9: return "Output-Page-Size";
	case 10: return "NAOCRD";
	case 11: return "NAOHTS";
	case 12: return "NAOHTD";
	case 13: return "NAOFFD";
	case 14: return "NAOVTS";
	case 15: return "NAOVTD";
	case 16: return "NAOLFD";
	case 17: return "Extended-ASCII";
	case 18: return "Logout";
	case 19: return "Byte-Macro";
	case 20: return "Data-Entry-Terminal";
	case 21: return "SUPDUP";
	case 22: return "SUPDUP-Output";
	case 23: return "Send-Location";
	case 24: return "Terminal-Type";
	case 25: return "End-of-Record";
	case 26: return "TACACS-User-Identification";
	case 27: return "Output-Marking";
	case 28: return "TTYLOC";
	case 29: return "Telnet-3270-Regime";
	case 30: return "X.3-PAD";
	case 31: return "NAWS";
	case 32: return "Terminal-Speed";
	case 33: return "Toggle-Flow-Control";
	case 34: return "Linemode";
	case 35: return "X-Display-Location";
	case 36: return "Environment";
	case 37: return "Authentication";
	case 38: return "Encryption-Option";
	case 39: return "New-Environment";
	case 40: return "TN3270E";
	case 41: return "XAUTH";
	case 42: return "CHARSET";
	case 43: return "Telnet-Remote-Serial-Port";
	case 44: return "Com-Port-Control";
	case 45: return "Telnet-Suppress-Local-Echo";
	case 46: return "Telnet-Start-TLS";
	case 47: return "KERMIT";
	case 48: return "SEND-URL";
	case 49: return "FORWARD_X";
	case 70: return "MUD-Server-Status";
	case 85: return "Compress";
	case 86: return "Compress2";
	case 90: return "MUD-Sound-Protocol";
	case 91: return "MUD-eXtension-Protocol";
	case 93: return "Zenith-Mud-Protocol";
	case 138: return "TELOPT PRAGMA LOGON";
	case 139: return "TELOPT SSPI LOGON";
	case 140: return "TELOPT PRAGMA HEARTBEAT";
	case 201: return "GMCP";
	case 255: return "Extended-Options-List";
	}
	return "<Unknown>";
}

//============================================================================

CApp::CApp() :
	m_pScriptMgr(NULL),
	m_nInputHeight(s_nDefaultInputHeight),
	m_yClickOfs(0),
	m_nHighPerfModeCounter(0),
	m_bConfigDirty(false),
	m_bConnecting(false),
	m_bConnected(false),
	m_enableNAWS(false),
	m_sentNegotiation(false),
	m_compressedInput(false),
	m_compressedOutput(false),
	m_nawsWidth(-1),
	m_nawsHeight(-1),
	m_lastNawsTime(0),
	m_nawsDirty(true),
	m_nDefaultPort(23),
	m_nTickDelay(100),
	m_bWindowRectValid(false),
	m_hIcon(NULL),
	m_pSocket(0)
{
	sm_pInstance = this;

	memset(&m_inputZip, 0, sizeof(m_inputZip));
	memset(&m_outputZip, 0, sizeof(m_outputZip));

	m_rcDivider.left = 0;
	m_rcDivider.top = 0;
	m_rcDivider.right = 0;
	m_rcDivider.bottom = 0;

	// Add default commands
	m_theCommands["help"] = Command(&CApp::CmdHelp,
		"Don't you think you've had enough help?");
	m_theCommands["telnet"] = Command(&CApp::CmdTelnet,
		"Display negotiated telnet options and current state.");
	m_theCommands["connect"] = Command(&CApp::CmdConnect,
		"Connect to a server, optionally specifying a host and port.\r\n"
		"Usage:\r\n"
		"   /connect [host[:port]]");
//	m_theCommands["httpconnect"] = Command(&CApp::CmdConnectHTTP,
//		"Connect to a server supporting HTTP protocol, optionally specifying a host and port.\r\n"
//		"Usage:\r\n"
//		"   /httpconnect [host[:port]]");
	m_theCommands["disconnect"] = Command(&CApp::CmdDisconnect,
		"Disconnect from the server");

	m_theCommands["load"] = Command(&CApp::CmdLoadScript,
		"Load a script file from the scripts directory. The file extension is optional.\r\n"
		"Usage:\r\n"
		"   /load filename");
	m_theCommands["reload"] = Command(&CApp::CmdReloadScripts,
		"Reload all scripts.\r\n"
		"Usage:\r\n"
		"   /reload");
	m_theCommands["loadplugin"] = Command(&CApp::CmdLoadPlugin,
		"Load a plugin library.\r\n"
		"Usage:\r\n"
		"   /unload filename");
	m_theCommands["unloadplugin"] = Command(&CApp::CmdUnloadPlugin,
		"Unload a plugin library.\r\n"
		"Usage:\r\n"
		"   /unload filename");
	m_theCommands[":"] = Command(&CApp::CmdShell,
		"Execute a script line.\r\n"
		"Usage:\r\n"
		"   /: script code");
	m_theCommands["dump"] = Command(&CApp::CmdDump,
		"Debug dump script information.\r\n"
		"Usage:\r\n"
		"   /dump");
	m_theCommands["downloads"] = Command(&CApp::CmdDownloads,
		"List pending downloads.\r\n"
		"Usage:\r\n"
		"   /downloads");
	m_theCommands["checkupdates"] = Command(&CApp::CmdCheckUpdates,
		"Check for updates.\r\n"
		"Usage:\r\n"
		"   /checkupdates");
	m_theCommands["resetpml"] = Command(&CApp::CmdResetPML,
		"Close any open PML tags sent by the server. Should only be used as a result of server error.\r\n"
		"Usage:\r\n"
		"   /resetpml");
}

CApp::~CApp()
{
}

//============================================================================

bool CApp::Init(HINSTANCE hInstance)
{
	// Load version number from resource
	char szBuff[512];
	GetModuleFileName(NULL, szBuff, sizeof(szBuff));
	DWORD viHandle;
	DWORD versionSize = GetFileVersionInfoSize(szBuff, &viHandle);
	if(versionSize != 0)
	{
		BYTE* pData = new BYTE[versionSize];
		if(GetFileVersionInfo(szBuff, viHandle, versionSize, pData))
		{
			UINT len = 0;
			const char* pVersion;
			if(VerQueryValue(pData, "\\StringFileInfo\\080904b0\\ProductVersion",
				(void**)&pVersion, &len) && len > 0)
			{
				std::string strVersion(pVersion, pVersion+len);
				size_t nPos = strVersion.find('.');
				if(nPos != std::string::npos)
				{
					ms_nVersionMajor = atoi(strVersion.substr(0, nPos).c_str());
					ms_nVersionMinor = atoi(strVersion.substr(nPos+1).c_str());
				}
			}
		}
		delete[] pData;
	}

	// Get settings file directory and create DruinkClient directory if it
	// doesn't exist.
	if(SHGetFolderPath(NULL, CSIDL_APPDATA, NULL, SHGFP_TYPE_CURRENT, szBuff) == S_OK)
		m_strSettingsPath = szBuff;
	m_strSettingsPath += s_szAppDataFolderName;
	CreateDirectory(m_strSettingsPath.c_str(), NULL);
	m_strSettingsPath += s_szExtraConfigFilename;

	bool bConfigError = !LoadConfig(m_theConfig, s_szConfigFilename);
	LoadConfig(m_theExtraConfig, m_strSettingsPath);

	// Determine window size based on required client area size
	if(m_bWindowRectValid)
	{
		m_nWidth = m_rcWindowInitial.right - m_rcWindowInitial.left;
		m_nHeight = m_rcWindowInitial.bottom - m_rcWindowInitial.top;
		m_nXPos = m_rcWindowInitial.left;
		m_nYPos = m_rcWindowInitial.top;
	}
	else
	{
		RECT rcWnd;
		rcWnd.top = 0;
		rcWnd.left = 0;
		rcWnd.bottom = 320;
		rcWnd.right = m_deOutput.GetIdealWidth();
		AdjustWindowRect(&rcWnd, WS_OVERLAPPEDWINDOW, FALSE);

		m_nWidth = rcWnd.right - rcWnd.left;
		m_nHeight = rcWnd.bottom - rcWnd.top;
		m_nXPos = CW_USEDEFAULT;
		m_nYPos = CW_USEDEFAULT;
	}

	// Init window
	m_strWindowTitle = s_szDefaultWindowTitle;
	if(!CWindow::Init(hInstance))
		return false;

	// Setup timer for ticking
	SetTimer(m_hWnd, s_nTickTimerID, m_nTickDelay, NULL);

	RECT rcClient;
	GetClientRect(m_hWnd, &rcClient);
	int w = rcClient.right - rcClient.left;
	int h = rcClient.bottom - rcClient.top - (m_nInputHeight+s_nDividerHeight);

	// Create DruinkEdits
	if(!m_deOutput.Create(hInstance, m_hWnd, 0, 0, w, h, true))
		return false;
	if(!m_deInput.Create(hInstance, m_hWnd, 0, h+s_nDividerHeight, w, m_nInputHeight, false))
		return false;

	// Find location of divider rect
	m_rcDivider.left = rcClient.left;
	m_rcDivider.top = rcClient.bottom - (m_nInputHeight+s_nDividerHeight);
	m_rcDivider.right = rcClient.right;
	m_rcDivider.bottom = m_rcDivider.top + s_nDividerHeight;

	// Add welcome string
	CTA2Stream theStream;
	theStream << CTA2Stream::Colour(0x00ff00)
		<< "Welcome to DruinkClient v2\r\n"
		<< "Type /help for a list of commands\r\n"
		<< CTA2Stream::Reset()
		<< "---------------------------------\r\n";
	m_deOutput.AddText(theStream.str());
	if(bConfigError)
		ShowError("Failed to load DruinkClient.cfg");

	// Set focus
	ShowWindow(m_hWnd, SW_SHOW);
	SetFocus(m_deInput.GetWindow());

	// Initialise script manager
	m_pScriptMgr = new CScriptManager;
	if(!m_pScriptMgr->Init())
	{
		ShowError("Failed to initialise script system: " + m_pScriptMgr->GetError());
		delete m_pScriptMgr;
		m_pScriptMgr = NULL;
	}
	else
	{
		// Print warnings
		if(!m_pScriptMgr->GetError().empty())
		{
			CTA2Stream theStream;
			theStream << CTA2Stream::Colour(0xffff00) <<
				m_pScriptMgr->GetError() << "\r\n" << CTA2Stream::Reset();
			m_deOutput.AddText(theStream.str());
		}

		// Set default functions
		ScriptFunction func;
		func.pParam = this;
		func.pfnFunction = ScriptFlashWindow;
		strcpy(func.szName, "flashWindow");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptShowWindow;
		strcpy(func.szName, "showWindow");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptPlaySound;
		strcpy(func.szName, "playSound");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptBeep;
		strcpy(func.szName, "beep");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptPrint;
		strcpy(func.szName, "print");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptPrintLn;
		strcpy(func.szName, "println");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptSend;
		strcpy(func.szName, "send");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptSendLn;
		strcpy(func.szName, "sendln");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptClear;
		strcpy(func.szName, "clear");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptIsConnected;
		strcpy(func.szName, "isConnected");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptGetVersion;
		strcpy(func.szName, "getVersion");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptGetConfigValue;
		strcpy(func.szName, "getConfigValue");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptSetConfigValue;
		strcpy(func.szName, "setConfigValue");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptEnterHighPerformanceMode;
		strcpy(func.szName, "enterHighPerfMode");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptExitHighPerformanceMode;
		strcpy(func.szName, "exitHighPerfMode");
		m_pScriptMgr->Register(func);
		func.pfnFunction = ScriptSetPasswordInput;
		strcpy(func.szName, "setPasswordInput");
		m_pScriptMgr->Register(func);

		if(!RegisterStandardLibrary(*m_pScriptMgr))
		{
			ShowError("Failed to initialise standard library!");
			return false;
		}

		// Load plugin DLLs
		LoadAllPlugins();

		// Run autoexec script and return
		if(!m_pScriptMgr->LoadAutoexecScript())
			ShowError("Failed to load autoexec script: " + m_pScriptMgr->GetError());
	}

	// Create thread termination event
	m_hThreadTerminate = CreateEvent(NULL, FALSE, FALSE, NULL);
	if(!m_hThreadTerminate)
	{
		DWORD dwError = GetLastError();
		ShowError("Failed to create thread termination event: " + Win32ErrorToString(dwError));
	}

	// Auto-update check?
	std::string strDoUpdateCheck = m_theConfig.GetValue("DoUpdateCheck");
	if((_stricmp(strDoUpdateCheck.c_str(), "true")==0 ||
		_stricmp(strDoUpdateCheck.c_str(), "yes")==0 || 
		_stricmp(strDoUpdateCheck.c_str(), "1")==0))
	{
		CmdCheckUpdates("");
	}

	return true;
}

//============================================================================

void CApp::Shutdown()
{
	// Signal threads to terminate
	if(m_hThreadTerminate)
		SetEvent(m_hThreadTerminate);

	// Make sure config is saved!
	if(m_bConfigDirty)
	{
		m_theExtraConfig.Save(m_strSettingsPath, s_szExtraConfigHeader);
		m_bConfigDirty = false;
	}

	CleanupStandardLibrary(*m_pScriptMgr);

	for(PluginListIter it=m_thePlugins.begin(); it!=m_thePlugins.end(); ++it)
		UnloadPlugin(*it);
	m_thePlugins.clear();

	if(m_pSocket)
	{
		m_pSocket->Disconnect();
		m_pSocket->Destroy();
		m_pSocket = NULL;
	}

	m_deInput.Release();
	m_deOutput.Release();

	m_bConnecting = false;
	m_bConnected = false;

	KillTimer(m_hWnd, s_nTickTimerID);
	CWindow::Shutdown();

	if(m_pScriptMgr)
	{
		delete m_pScriptMgr;
		m_pScriptMgr = NULL;
	}

	// Wait on all downloads completing
	if(!m_vDownloads.empty())
	{
		std::vector<HANDLE> vThreadHandles;
		vThreadHandles.reserve(m_vDownloads.size());
		for(DownloadListIter it=m_vDownloads.begin(); it!=m_vDownloads.end(); ++it)
			vThreadHandles.push_back((*it)->hThread);
		WaitForMultipleObjects((DWORD)vThreadHandles.size(), &vThreadHandles[0], TRUE, INFINITE);
		for(DownloadListIter it=m_vDownloads.begin(); it!=m_vDownloads.end(); ++it)
		{
			HTTPGet* pGet = *it;
			CloseHandle(pGet->hThread);
			delete pGet;
		}
		m_vDownloads.clear();
	}

	// Close thread termination handle
	if(m_hThreadTerminate)
		CloseHandle(m_hThreadTerminate);
	m_hThreadTerminate = NULL;
}

//============================================================================

void CApp::SetIcon(HICON hIcon)
{
	bool isDefaultIcon = false;
	if(!hIcon)
	{
		hIcon = LoadIcon(m_hInstance, MAKEINTRESOURCE(IDI_MAIN));
		isDefaultIcon = true;
	}

	// Change icon
	SetClassLongPtr(m_hWnd, GCLP_HICON, (LONG_PTR)hIcon);
	SetClassLongPtr(m_hWnd, GCLP_HICONSM, (LONG_PTR)hIcon);

	// Free old icon and store this one
	if(m_hIcon)
		DestroyIcon(m_hIcon);
	m_hIcon = hIcon;
	if(isDefaultIcon)
		m_hIcon = NULL;
}

void CApp::SetTitle(const char* szTitle)
{
	if(!szTitle)
		szTitle = s_szDefaultWindowTitle;
	SetWindowText(m_hWnd, szTitle);
}

//============================================================================

void CApp::Send(const std::string& str)
{
	static int nRecursionCount = 0;

	++nRecursionCount;
	if(nRecursionCount > 16)
		RawSend(str);
	else
		SendLineImpl(str);
	--nRecursionCount;
}

void CApp::RawSend(const std::string& str)
{
	if(m_bConnected)
		SocketSend(str);
	else
		m_deOutput.AddText(str);
}

void CApp::SocketSend(const std::string& str)
{
	SocketSend((const BYTE*)str.c_str(), str.length());
}

void CApp::SocketSend(const BYTE* pData, size_t len)
{
	// Early out if not sending compressed output
	if(!m_compressedOutput)
	{
		m_pSocket->Send(pData, len);
		return;
	}

	static BYTE buff[4*1024];
	m_outputZip.next_in = (z_const Bytef*)pData;
	m_outputZip.avail_in = (uInt)len;
	do {
		m_outputZip.next_out = buff;
		m_outputZip.avail_out = sizeof(buff);
		int ret = deflate(&m_outputZip, Z_PARTIAL_FLUSH);
		if(ret != Z_OK)
		{
			char szBuff[256];
			sprintf_s(szBuff, sizeof(szBuff), "deflate() failed with error %d", ret);
			ShowError(szBuff);
			CmdDisconnect("");
			return;
		}
		size_t bytesCompressed = sizeof(buff) - m_outputZip.avail_out;
		m_pSocket->Send(buff, bytesCompressed);
	} while(m_outputZip.avail_out == 0);
}

//============================================================================

HICON CApp::GetIcon()
{
	return LoadIcon(m_hInstance, MAKEINTRESOURCE(IDI_MAIN));
}

HCURSOR CApp::GetCursor()
{
	return LoadCursor(NULL, IDC_SIZENS);
}

//============================================================================

LRESULT CApp::WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
		case WM_DESTROY:
		{
			// Save window size and pos
			RECT rcWindow;
			GetWindowRect(m_hWnd, &rcWindow);

			// X
			{
				std::ostringstream os;
				os << rcWindow.left;
				m_theExtraConfig.SetValue("Window.X", os.str());
			}

			// Y
			{
				std::ostringstream os;
				os << rcWindow.top;
				m_theExtraConfig.SetValue("Window.Y", os.str());
			}

			// W
			{
				std::ostringstream os;
				os << (rcWindow.right - rcWindow.left);
				m_theExtraConfig.SetValue("Window.W", os.str());
			}

			// H
			{
				std::ostringstream os;
				os << (rcWindow.bottom - rcWindow.top);
				m_theExtraConfig.SetValue("Window.H", os.str());
			}

			m_bConfigDirty = true;
		}
		break;

		case WM_CLOSE:
		{
			bool bDownloadsPending = false;
			for(DownloadListIter it=m_vDownloads.begin(); it!=m_vDownloads.end(); ++it)
			{
				HTTPGet* pGet = *it;
				if(pGet->nUserData == FLAG_StandardGet)
				{
					bDownloadsPending = true;
					break;
				}
			}
			if(bDownloadsPending)
			{
				int nRet = MessageBox(m_hWnd, "You have downloads in progress. Exit anyway?",
					"Downloads in progress", MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2);
				if(nRet == IDNO)
					return 0;
			}
		}
		break;

		case WM_SIZE:
		case WM_SIZING:
		{
			RECT rcClient;
			if(uMsg == WM_SIZE)
			{
				GetClientRect(m_hWnd, &rcClient);
				rcClient.right -= rcClient.left;
				rcClient.bottom -= rcClient.top;
				rcClient.left = 0;
				rcClient.top = 0;
			}
			else
			{
				rcClient = *(RECT*)lParam;
				RECT rcNull = {0, 0, 0, 0};
				AdjustWindowRect(&rcNull, (DWORD)GetWindowLongPtr(m_hWnd, GWL_STYLE), FALSE);
				rcClient.top += rcNull.bottom - rcNull.top;
				rcClient.left += rcNull.right - rcNull.left;
				rcClient.right -= rcClient.left;
				rcClient.bottom -= rcClient.top;
				rcClient.left = 0;
				rcClient.top = 0;
			}
			int w = rcClient.right;
			int h = rcClient.bottom - (m_nInputHeight+s_nDividerHeight);

			// Update location of divider rect
			m_rcDivider.left = rcClient.left;
			m_rcDivider.top = rcClient.bottom - (m_nInputHeight+s_nDividerHeight);
			m_rcDivider.right = rcClient.right;
			m_rcDivider.bottom = m_rcDivider.top + s_nDividerHeight;

			MoveWindow(m_deOutput.GetWindow(), 0, 0, w, h, FALSE);
			MoveWindow(m_deInput.GetWindow(), 0, h+s_nDividerHeight, w, m_nInputHeight, FALSE);

			HDC hDC = GetDC(m_hWnd);
			FillRect(hDC, &m_rcDivider, (HBRUSH)GetStockObject(COLOR_WINDOW+1));
			ReleaseDC(m_hWnd, hDC);
			if(uMsg == WM_SIZE && m_bConnected && m_enableNAWS)
				SendNAWS();
			return 0;
		}
		break;

		case WM_GETMINMAXINFO:
		{
			RECT rcMin;
			rcMin.left = 0;
			rcMin.top = 0;
			rcMin.right = s_nMinWidth;
			rcMin.bottom = m_nInputHeight + s_nDividerHeight + s_nMinOutputHeight;
			AdjustWindowRect(&rcMin, (DWORD)GetWindowLongPtr(m_hWnd, GWL_STYLE), FALSE);
			MINMAXINFO* pInfo = (MINMAXINFO*)lParam;
			pInfo->ptMinTrackSize.x = rcMin.right-rcMin.left;
			pInfo->ptMinTrackSize.y = rcMin.bottom-rcMin.top;
			return 0;
		}
		break;

		case WM_ERASEBKGND:
		{
			FillRect((HDC)wParam, &m_rcDivider, (HBRUSH)GetStockObject(COLOR_WINDOW+1));
			return 0;
		}
		break;

		case WM_ACTIVATE:
		{
			if(LOWORD(wParam) != WA_INACTIVE)
			{
				// Set focus to input box
				SetFocus(m_deInput.GetWindow());
				return 0;
			}
		}
		break;

		case WM_LBUTTONDOWN:
		{
			SetCapture(m_hWnd);
		}
		break;

		case WM_LBUTTONUP:
		{
			ReleaseCapture();
		}
		break;

		case WM_MOUSEMOVE:
		{
			if((wParam & MK_LBUTTON) == 0)
				break;

			RECT rcClient;
			GetClientRect(m_hWnd, &rcClient);
			rcClient.right -= rcClient.left;
			rcClient.bottom -= rcClient.top;
			rcClient.left = 0;
			rcClient.top = 0;

			m_nInputHeight = rcClient.bottom - HIWORD(lParam);
			if(m_nInputHeight < s_nMinInputHeight)
				m_nInputHeight = s_nMinInputHeight;
			else if(rcClient.bottom - (m_nInputHeight+s_nDividerHeight) < s_nMinOutputHeight)
				m_nInputHeight = rcClient.bottom - (s_nMinOutputHeight+s_nDividerHeight);

			int w = rcClient.right;
			int h = rcClient.bottom - (m_nInputHeight+s_nDividerHeight);

			// Update location of divider rect
			m_rcDivider.left = 0;
			m_rcDivider.top = rcClient.bottom - (m_nInputHeight+s_nDividerHeight);
			m_rcDivider.right = rcClient.right;
			m_rcDivider.bottom = m_rcDivider.top + s_nDividerHeight;

			MoveWindow(m_deOutput.GetWindow(), 0, 0, w, h, FALSE);
			MoveWindow(m_deInput.GetWindow(), 0, h+s_nDividerHeight, w, m_nInputHeight, FALSE);
		}
		break;

		case WM_TIMER:
		{
			if(wParam == s_nTickTimerID)
			{
				Tick();
			}
		}
		break;

		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			BeginPaint(m_hWnd, &ps);
			FillRect(ps.hdc, &m_rcDivider, (HBRUSH)GetStockObject(COLOR_WINDOW+1));
			EndPaint(m_hWnd, &ps);
		}
		break;

		case WM_COMMAND:
		{
			// Is this from the accelerator?
			if(HIWORD(wParam) != 1)
				break;

			if(m_deInput.IsFocused())
				m_deInput.OnCommand(LOWORD(wParam));
			else if(m_deOutput.IsFocused())
				m_deOutput.OnCommand(LOWORD(wParam));
		}
		break;
	}
	return CWindow::WindowProc(uMsg, wParam, lParam);
}

//============================================================================

void CApp::OnNotify(NMHDR* pHdr)
{
	CDruinkEdit* pControl;
	if(pHdr->hwndFrom == m_deInput.GetWindow())
		pControl = &m_deInput;
	else if(pHdr->hwndFrom == m_deOutput.GetWindow())
		pControl = &m_deOutput;
	else
		return;

	DruinkEditNotify* pNotify = (DruinkEditNotify*)pHdr;
	if(pHdr->code == DEN_NEWLINE)
	{
		// Unlock output
		m_deOutput.PauseOutput(false);

		// Grab the text
		std::string strText = pControl->GetText() + "\r\n";
		pControl->Clear();

		// Call initial input handler
		strText = OnPreparseText(strText);

		// Process line by line
		size_t nPos=0, nNext=0;
		do {
			nPos = strText.find('\n', nNext);
			if(nPos != std::string::npos)
			{
				std::string strLine = strText.substr(nNext, nPos-nNext+1);
				SendLineImpl(strLine);
				nNext = nPos+1;
				if(nNext >= strText.length())
					nPos = std::string::npos;
			}
			else
			{
				std::string strLine = strText.substr(nNext);
				SendLineImpl(strLine);
			}
		} while(nPos != std::string::npos);
	}
	else if(pHdr->code == DEN_ISREADONLY && pHdr->hwndFrom == m_deOutput.GetWindow())
	{
		// Change focus to input control
		SetFocus(m_deInput.GetWindow());

		// Pass message along
		m_deInput.DoKeypress(pNotify->nParam, pNotify->nParam2);
	}
	else if(pHdr->code == DEN_GOOGLE)
	{
		const char* szSearch = (const char*)pNotify->nParam;
		std::string strURL = m_strSearchURL + szSearch;
		pControl->DoShellExecute(strURL);
	}
	else if(pHdr->code == DEN_SAVEURL)
	{
		SaveURL((const char*)pNotify->nParam);
	}
}

//============================================================================

void CApp::Tick()
{
	// Process connecting
	if(m_bConnecting)
		TickConnecting();

	// Tick the socket
	if(m_bConnected)
		TickSocket();

	// Tick downloads
	for(DownloadListIter it=m_vDownloads.begin(); it!=m_vDownloads.end();)
	{
		HTTPGet* pGet = *it;
		DWORD dwRet = WaitForSingleObject(pGet->hThread, 0);
		if(dwRet == WAIT_OBJECT_0)
		{
			OnDownloadComplete(pGet);
			delete pGet;
			it = m_vDownloads.erase(it);
		}
		else if(dwRet == WAIT_FAILED || dwRet == WAIT_ABANDONED)
		{
			// Error
			DWORD dwError = GetLastError();
			ShowError("Failed to download \"" + pGet->strAddress + "\". Error: " + Win32ErrorToString(dwError));
			delete pGet;
			it = m_vDownloads.erase(it);
		}
		else
			++it;
	}

	// Handle PML
	while(m_deOutput.HasPMLWaiting())
	{
		std::string strTag = m_deOutput.GetNextPMLTag();
		if(m_pScriptMgr)
		{
			m_pScriptMgr->OnReceivePML(strTag);
			if(!m_pScriptMgr->GetError().empty())
				ShowError(m_pScriptMgr->GetError());
		}
	}

	// Tick standard lib
	if(m_pScriptMgr)
	{
		m_pScriptMgr->ResetError();
		TickStandardLibrary(*m_pScriptMgr);
		if(!m_pScriptMgr->GetError().empty())
			ShowError(m_pScriptMgr->GetError());

		// Tick script manager
		m_pScriptMgr->OnTick();
		if(!m_pScriptMgr->GetError().empty())
			ShowError(m_pScriptMgr->GetError());
	}

	// Need to flush config?
	if(m_bConfigDirty)
	{
		m_theExtraConfig.Save(m_strSettingsPath, s_szExtraConfigHeader);
		m_bConfigDirty = false;
	}

	// Time to refresh NAWS?
	if(m_bConnected && m_enableNAWS && m_nawsDirty)
		SendNAWS();
}

//============================================================================

void CApp::TickConnecting()
{
	CBaseSocket::ConnectStatus eStatus = m_pSocket->PollConnectComplete();
	switch(eStatus)
	{
		case CBaseSocket::STATUS_Connecting:
		{
			// Do nothing
		}
		break;

		case CBaseSocket::STATUS_ConnectError:
		{
			ShowError("Failed to connect to server. " + std::string(m_pSocket->GetError()));
			m_pSocket->ClearError();
			m_bConnecting = false;

			OnConnectFailed();
		}
		break;

		case CBaseSocket::STATUS_Connected:
		{
			CTA2Stream theStream;
			theStream << CTA2Stream::Colour(0x00ff00) << "Connected!\r\n" << CTA2Stream::Reset();
			m_deOutput.AddText(theStream.str());
			m_bConnecting = false;
			m_bConnected = true;

			OnConnected();
		}
		break;

		default:
		{
			// Must be STATUS_NotConnected. How did we get here?
			Assert(false, "Unhandled status");
		}
		break;
	}
}

//============================================================================
void CApp::TickSocket()
{
	CBaseSocket::SocketStatus eStatus = m_pSocket->Poll();
	switch(eStatus)
	{
		case CBaseSocket::STATUS_Ok:
		{
			// Any new data?
			if(m_pSocket->DataPending())
			{
				m_socketBuffer.insert(m_socketBuffer.end(), m_pSocket->GetBuffer(),
					m_pSocket->GetBuffer()+m_pSocket->GetBufferSize());
				m_pSocket->ClearBuffer();

				// Decompress data
				std::vector<char>* pData = &m_socketBuffer;
				if(m_compressedInput)
				{
					DecompressPendingData();
					pData = &m_decompressionBuffer;
				}

				// Any telnet data?
				if(ProcessTelnet(*pData))
				{
					std::string strLine;
					do {
						strLine = GetLine(*pData);
						if(!strLine.empty())
						{
							if(m_pScriptMgr)
								strLine = m_pScriptMgr->OnReceive(strLine);
							if(!strLine.empty())
								m_deOutput.AddText(strLine);
							if(m_pScriptMgr && !m_pScriptMgr->GetError().empty())
								ShowError(m_pScriptMgr->GetError());
						}
					} while(!strLine.empty());
				}
			}
		}
		break;

		case CBaseSocket::STATUS_Disconnected:
		{
			ShowError("Disconnected");

			m_bConnected = false;
			m_pSocket->Disconnect();
			OnDisconnected(false);
		}
		break;

		case CBaseSocket::STATUS_Error:
		{
			ShowError(m_pSocket->GetError());

			m_bConnected = false;
			m_pSocket->Disconnect();
			OnDisconnected(false);
		}
		break;

		default:
		{
			Assert(false, "Unknown state");
		}
		break;
	}
}

//============================================================================
void CApp::SendLineImpl(const std::string& str)
{
size_t nOfs = 0;

	// Check for commands
	if(!str.empty() && (str[0] == '/'))
	{
		// Is it a double slash? (No command)
		if((str.length() >= 2) && (str[1] == '/'))
			nOfs = 1;
		else
		{
			// It's a command
			ProcessCommand(str);
			return;
		}
	}

	// Send off to script manager, with "\r\n" stripped
	std::string strOutput = OnSendLine(str);

	// Add to socket if connected, or add to output if not
	if(!strOutput.empty())
	{
		if(m_bConnected)
			SocketSend(strOutput);
		else
			m_deOutput.AddText(ParseTA2Tags(strOutput));
	}
}

void CApp::ProcessCommand(const std::string& str)
{
	// Convert into command and parameters
	Assert(str[0] == '/', "Not a command");
	std::string strCommand, strParams;
	size_t nPos = str.find(' ', 1);
	if(nPos != std::string::npos)
	{
		strCommand = str.substr(1, nPos-1);
		strParams = str.substr(nPos+1);
		while(!strParams.empty() && (strParams.end()[-1]=='\r' || strParams.end()[-1]=='\n'))
			strParams.erase(strParams.end()-1);
	}
	else
	{
		strCommand = str.substr(1);
		while(!strCommand.empty() && (strCommand.end()[-1]=='\r' || strCommand.end()[-1]=='\n'))
			strCommand.erase(strCommand.end()-1);
	}

	// Convert command to lowercase
	std::transform(strCommand.begin(), strCommand.end(), strCommand.begin(), tolower);

	// Pass it off to the script manager to handle
	if(strCommand != "load")
	{
		if(m_pScriptMgr)
		{
			bool bRet = m_pScriptMgr->OnCommand(strCommand, strParams);
			if(!m_pScriptMgr->GetError().empty())
				ShowError(m_pScriptMgr->GetError());
			if(bRet)
				return;
		}
	}

	// Process the command
	CommandListIter itCommand = m_theCommands.find(strCommand);
	if(itCommand == m_theCommands.end())
	{
		// Unknown command
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0xff0000) << "Unknown command: " << CTA2Stream::Reset() <<
			strCommand << ". Type /help for a list of commands\r\n" << CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
	}
	else
	{
		// Call command
		(this->*itCommand->second.pfn)(strParams);
	}
}

//============================================================================
bool CApp::ProcessTelnet(std::vector<char>& data)
{
	if(data.empty())
		return false;

	for(size_t i=0; i+2<data.size(); ++i)
	{
		BYTE* pData = (BYTE*)&data[0];
		if(pData[i] != TELNET_IAC)
			continue;

		// Telnet IAC, valid data after it?
		BYTE cmd = pData[i+1];
		if(cmd == TELNET_WILL)
		{
			OnTelnetWillOrDo(pData[i+2], true);
			data.erase(data.begin()+i, data.begin()+i+3);
			--i;
		}
		else if(cmd == TELNET_WONT)
		{
			OnTelnetWontOrDont(pData[i+2], true);
			data.erase(data.begin()+i, data.begin()+i+3);
			--i;
		}
		else if(cmd == TELNET_DO)
		{
			OnTelnetWillOrDo(pData[i+2], false);
			data.erase(data.begin()+i, data.begin()+i+3);
			--i;
		}
		else if(cmd == TELNET_DONT)
		{
			OnTelnetWontOrDont(pData[i+2], false);
			data.erase(data.begin()+i, data.begin()+i+3);
			--i;
		}
		else if(cmd == TELNET_SB)
		{
			// The only thing we actually care about is COMPRESS / COMPRESS2
			if(data.size()-i < 5)
				return false; // Need more data

			// Compress subnegotiation?
			else if(pData[i+2] == TELNETOP_COMPRESS && pData[i+3] == TELNET_WILL && pData[i+4] == TELNET_SE)
			{
				// Compress
				DebugLog("GOT Compress SB\n");
				data.erase(data.begin()+i, data.begin()+i+5);
				OnCompressInputChanged(true, i);
				return false;
			}

			// Compress2 subnegotiation?
			else if(pData[i+2] == TELNETOP_COMPRESS2 && pData[i+3] == TELNET_IAC && pData[i+4] == TELNET_SE)
			{
				// Compress2
				DebugLog("GOT Compress2 SB\n");
				data.erase(data.begin()+i, data.begin()+i+5);
				OnCompressInputChanged(true, i);
				return false;
			}

			// Unknown subnegotiation; scan to "IAC SE"
			else
			{
				DebugLog("GOT unknown SB\n");
				bool gotEnd = false;
				for(size_t j=i+3; j<data.size()-1; ++j)
				{
					if(pData[j] == TELNET_IAC && pData[j+1] == TELNET_SE)
					{
						data.erase(data.begin()+i, data.begin()+j+2);
						--i;
						gotEnd = true;
						break;
					}
				}
				if(!gotEnd)
					return false; // Need more data
			}
		}
	}

	return true;
}

//============================================================================
void CApp::OnTelnetWillOrDo(BYTE cmd, bool _will)
{
	DebugLog("GOT %s %d (%s)\n", _will?"WILL":"DO", cmd, telnetCmdToString(cmd));

	// Do one-off announcement of our capabilities
	if(!m_sentNegotiation)
	{
		SendInitialNegotiation();
		m_sentNegotiation = true;
	}

	// Something we know about?
	if((cmd == TELNETOP_NAWS || cmd == TELNETOP_COMPRESS || cmd == TELNETOP_COMPRESS2) &&
		!m_respondedOptions[cmd].isNegotiated())
	{
		// Special case - refuse Compress if Compress2 already agreed
		if(cmd == TELNETOP_COMPRESS && m_respondedOptions[TELNETOP_COMPRESS2].isNegotiated())
		{
			DebugLog("Refusing compress because compress2 agreed (ignore next log)\n");
			OnTelnetWontOrDont(cmd, _will);
			return;
		}

		DebugLog("SEND %s %s\n", _will?"DO":"WILL", telnetCmdToString(cmd));
		BYTE buff[] = {TELNET_IAC, _will ? TELNET_DO : TELNET_WILL, cmd};
		SocketSend(buff, sizeof(buff));
		m_respondedOptions[cmd].op = _will ? TELNET_WILL : TELNET_DO;

		// If we just agreed to NAWS, send our current window size
		if(cmd == TELNETOP_NAWS)
		{
			m_enableNAWS = true;
			SendNAWS(true);
		}

		// If we just agreed to Compress / Compress2, compress output
		else if(cmd == TELNETOP_COMPRESS && !_will)
		{
			DebugLog("SEND SB Compress in response to DO Compress\n");
			BYTE buff[] = {TELNET_IAC, TELNET_SB, TELNETOP_COMPRESS, TELNET_WILL, TELNET_SE};
			SocketSend(buff, sizeof(buff));
			OnCompressOutputChanged(true, false);
		}

		// If we just agreed to Compress / Compress2, compress output
		else if(cmd == TELNETOP_COMPRESS2 && !_will)
		{
			DebugLog("SEND SB Compress2 in response to DO Compress2\n");
			BYTE buff[] = {TELNET_IAC, TELNET_SB, TELNETOP_COMPRESS2, TELNET_IAC, TELNET_SE};
			SocketSend(buff, sizeof(buff));
			OnCompressOutputChanged(true, false);
		}
	}
	else
		m_respondedOptions[cmd].op = _will ? TELNET_WILL : TELNET_DO;

	// Ignore option if we don't understand it
}

void CApp::OnTelnetWontOrDont(BYTE cmd, bool _wont)
{
	DebugLog("GOT %s %d (%s)\n", _wont?"WON'T":"DON'T", cmd, telnetCmdToString(cmd));

	// Do one-off announcement of our capabilities
	if(!m_sentNegotiation)
	{
		SendInitialNegotiation();
		m_sentNegotiation = true;
	}

	// Agree if we've not already
	if(!m_respondedOptions[cmd].isNegotiated())
	{
		BYTE buff[] = {TELNET_IAC, _wont ? TELNET_DONT : TELNET_WONT, cmd};
		SocketSend(buff, sizeof(buff));
		m_respondedOptions[cmd].op = _wont ? TELNET_WONT : TELNET_DONT;
	}
}

//============================================================================
void CApp::OnCompressInputChanged(bool compressed, size_t socketBuffOffset)
{
	// Early out if no change to compression state
	if(compressed == m_compressedInput)
		return;
	m_compressedInput = compressed;

	// Cleanup?
	if(!m_compressedInput)
	{
		inflateEnd(&m_inputZip);
		memset(&m_inputZip, 0, sizeof(m_inputZip));
		return;
	}

	// Setup new z_stream
	memset(&m_inputZip, 0, sizeof(m_inputZip));
	int err = inflateInit(&m_inputZip);
	if (err != Z_OK)
	{
		ShowError("inflateInit() failed!");
		CmdDisconnect("");
		return;
	}

	// Copy any pending already compressed data into the decompression buffer
	m_decompressionBuffer.insert(m_decompressionBuffer.begin(), m_socketBuffer.begin(),
		m_socketBuffer.begin()+socketBuffOffset);
	m_socketBuffer.erase(m_socketBuffer.begin(), m_socketBuffer.begin()+socketBuffOffset);
}

void CApp::OnCompressOutputChanged(bool compressed, bool gracefulFinish)
{
	// Early out if no change to compression state
	if(compressed == m_compressedOutput)
		return;
	m_compressedOutput = compressed;

	// Cleanup?
	if(!m_compressedOutput)
	{
		if(gracefulFinish)
		{
			static BYTE buff[1024];
			m_outputZip.next_in = (z_const Bytef*)"";
			m_outputZip.avail_in = 0;
			do {
				m_outputZip.next_out = buff;
				m_outputZip.avail_out = sizeof(buff);
				int ret = deflate(&m_outputZip, Z_FINISH);
				if(ret != Z_OK)
				{
					char szBuff[256];
					sprintf_s(szBuff, sizeof(szBuff), "deflate() failed with error %d", ret);
					ShowError(szBuff);
					CmdDisconnect("");
					return;
				}
				size_t bytesCompressed = sizeof(buff) - m_outputZip.avail_out;
				m_pSocket->Send(buff, bytesCompressed);
			} while(m_outputZip.avail_out == 0);
		}
		deflateEnd(&m_outputZip);
		memset(&m_outputZip, 0, sizeof(m_outputZip));
		return;
	}

	// Setup new z_stream
	memset(&m_outputZip, 0, sizeof(m_outputZip));
	int err = deflateInit(&m_outputZip, 9);
	if (err != Z_OK)
	{
		ShowError("deflateInit() failed!");
		CmdDisconnect("");
		return;
	}
}

void CApp::DecompressPendingData()
{
	static BYTE buff[64*1024];

	m_inputZip.next_in = (z_const Bytef*)&m_socketBuffer[0];
	m_inputZip.avail_in = (uInt)m_socketBuffer.size();
	bool eof = false;
	do {
		m_inputZip.next_out = buff;
		m_inputZip.avail_out = sizeof(buff);
		int ret = inflate(&m_inputZip, Z_NO_FLUSH);
		if(ret == Z_NEED_DICT)
			ret = Z_DATA_ERROR;
		else if(ret == Z_STREAM_END)
			eof = true;

		if(ret < 0)
		{
			char szBuff[256];
			sprintf_s(szBuff, sizeof(szBuff), "inflate() failed with error %d", ret);
			ShowError(szBuff);
			CmdDisconnect("");
			return;
		}
		size_t bytesDecompressed = sizeof(buff) - m_inputZip.avail_out;
		m_decompressionBuffer.insert(m_decompressionBuffer.end(), buff, buff+bytesDecompressed);
	} while(m_inputZip.avail_out == 0);
	m_socketBuffer.erase(m_socketBuffer.begin(), m_socketBuffer.begin()+(m_socketBuffer.size() - m_inputZip.avail_in));

	// End of stream?
	if(eof)
	{
		// Poke the decompressed data into the socket buffer
		OnCompressInputChanged(false, 0);
		m_socketBuffer.insert(m_socketBuffer.begin(), m_decompressionBuffer.begin(), m_decompressionBuffer.end());
		m_decompressionBuffer.clear();
	}
}

std::string CApp::GetLine(std::vector<char>& data)
{
	for(size_t i=0; i<data.size(); ++i)
	{
		if(data[i] == '\n')
		{
			std::string strRet(i+1, '*');
			memcpy(&strRet[0], &data[0], i+1);
			data.erase(data.begin(), data.begin()+i+1);
			return strRet;
		}
	}
	return "";
}

//============================================================================
void CApp::SendInitialNegotiation()
{
	DebugLog("Sending initial telnet negotiation\n");

	BYTE enabledOptions[] = {TELNETOP_NAWS, TELNETOP_COMPRESS2, TELNETOP_COMPRESS};
	for(size_t i=0; i<sizeof(enabledOptions)/sizeof(enabledOptions[0]); ++i)
	{
		BYTE buff[] = {TELNET_IAC, TELNET_WILL, enabledOptions[i]};
		SocketSend(buff, sizeof(buff));
	}
}

void CApp::SendNAWS(bool sendNow)
{
	// Get window size in characters
	SIZE outSize = m_deOutput.GetSizeInCharacters();
	int widthChars = outSize.cx;
	int heightChars = outSize.cy;
	if(widthChars == 0 || heightChars == 0)
	{
		// Don't send zero window size
		return;
	}
	if(widthChars != m_nawsWidth || heightChars != m_nawsHeight)
	{
		m_nawsWidth = widthChars;
		m_nawsHeight = heightChars;
		m_nawsDirty = true;
	}

	time_t currTime = time(NULL);
	if((m_nawsDirty && currTime-m_lastNawsTime >= s_nawsMinInterval) || sendNow)
	{
		// RFC1073: IAC SB NAWS <16-bit value> <16-bit value> IAC SE
		DebugLog("Sending NAWS\n");
		BYTE buff[] = {TELNET_IAC, TELNET_SB, TELNETOP_NAWS, (BYTE)(widthChars >> 8), (BYTE)(widthChars&0xff),
			(BYTE)(heightChars >> 8), (BYTE)(heightChars&0xff), TELNET_IAC, TELNET_SE};
		SocketSend(buff, sizeof(buff));
		m_nawsDirty = false;
		m_lastNawsTime = currTime;
	}
}

//============================================================================
std::string CApp::OnPreparseText(const std::string& str)
{
	if(!m_pScriptMgr)
		return str;

	std::string strRet = m_pScriptMgr->OnPreparseText(str);
	if(!m_pScriptMgr->GetError().empty())
		ShowError(m_pScriptMgr->GetError());
	return strRet;
}

std::string CApp::OnSendLine(const std::string& str)
{
	if(!m_pScriptMgr)
		return str;

	std::string strRet = m_pScriptMgr->OnSendLine(str);
	if(!m_pScriptMgr->GetError().empty())
		ShowError(m_pScriptMgr->GetError());
	return strRet;
}

void CApp::OnConnected()
{
	if(!m_pScriptMgr)
		return;

	m_pScriptMgr->OnConnected();
	if(!m_pScriptMgr->GetError().empty())
		ShowError(m_pScriptMgr->GetError());
}

void CApp::OnConnectFailed()
{
	if(!m_pScriptMgr)
		return;

	m_pScriptMgr->OnConnectFailed();
	if(!m_pScriptMgr->GetError().empty())
		ShowError(m_pScriptMgr->GetError());
}

void CApp::OnDisconnected(bool bUserRequested)
{
	if(!m_pScriptMgr)
		return;

	m_pScriptMgr->OnDisconnected(bUserRequested);
	if(!m_pScriptMgr->GetError().empty())
		ShowError(m_pScriptMgr->GetError());

	// Reset telnet negotiation state
	for(size_t i=0; i<sizeof(m_respondedOptions)/sizeof(m_respondedOptions[0]); ++i)
		m_respondedOptions[i].reset();
	m_enableNAWS = false;
	m_sentNegotiation = false;
	OnCompressInputChanged(false, 0);
	OnCompressOutputChanged(false, false);
	m_decompressionBuffer.clear();
}

//============================================================================
void CApp::CmdHelp(const std::string& strParams)
{
	// If there are no parameters, display command list
	if(strParams.empty())
	{
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0xffff00) << "Registered commands:\r\n" << CTA2Stream::Reset();
		for(CommandListIter it=m_theCommands.begin(); it!=m_theCommands.end(); ++it)
			theStream << "   /" << it->first << "\r\n" << CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
		return;
	}

	// Find the command
	std::string strCommand(strParams);
	std::transform(strCommand.begin(), strCommand.end(), strCommand.begin(), tolower);
	CommandListIter itCommand = m_theCommands.find(strCommand);
	if(itCommand == m_theCommands.end())
	{
		// Unknown command
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0xff0000) << "Unknown command: " << CTA2Stream::Reset() <<
			strCommand << ". Type /help for a list of commands\r\n" << CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
	}
	else
	{
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0xffff00) << "/" << strCommand << ":\r\n"
			<< itCommand->second.strHelp << "\r\n" << CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
	}
}

void CApp::CmdTelnet(const std::string& strParams)
{
	UNREFERENCED_PARAMETER(strParams);

	if(!m_bConnected)
	{
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0xff0000) << "Not connected.\r\n" << CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
		return;
	}

	CTA2Stream theStream;
	theStream << "Negotiated options:\r\n";
	for(size_t i=0; i<sizeof(m_respondedOptions)/sizeof(m_respondedOptions[0]); ++i)
	{
		TelnetOption op = m_respondedOptions[i];
		if(op.isNegotiated())
		{
			bool isActive = false;
			if(i == TELNETOP_NAWS)
				isActive = m_enableNAWS;
			else if(i == TELNETOP_COMPRESS && m_respondedOptions[TELNETOP_COMPRESS2].isNegotiated())
				isActive = false;
			else if(i == TELNETOP_COMPRESS2)
				isActive = m_compressedInput || m_compressedOutput;
			if(isActive)
				theStream << CTA2Stream::Colour(0x00ff00);
			else if(op.op == TELNET_WONT || op.op == TELNET_DONT)
				theStream << CTA2Stream::Colour(0xff0000);
			else
				theStream << CTA2Stream::Colour(0xffff00);
			switch(op.op)
			{
			case TELNET_WILL: theStream << "   WILL "; break;
			case TELNET_WONT: theStream << "   WON'T "; break;
			case TELNET_DO: theStream << "   DO "; break;
			case TELNET_DONT: theStream << "   DON'T "; break;
			}
			theStream << telnetCmdToString((BYTE)i);
			char szBuff[32];
			sprintf_s(szBuff, sizeof(szBuff), " (%d)\r\n", i);
			theStream << szBuff << CTA2Stream::Reset();
		}
	}
	if(m_enableNAWS)
		theStream << CTA2Stream::Colour(0x00ff00) << "Server is aware of current window size.\r\n" << CTA2Stream::Reset();
	if(m_compressedInput)
		theStream << CTA2Stream::Colour(0x00ff00) << "Data from server is being received compressed.\r\n" << CTA2Stream::Reset();
	if(m_compressedOutput)
		theStream << CTA2Stream::Colour(0x00ff00) << "Data to server is being sent compressed.\r\n" << CTA2Stream::Reset();

	m_deOutput.AddText(theStream.str());
}

void CApp::CmdConnect(const std::string& strParams)
{
	// Setup socket
	m_pSocket = new CSocket;
	if(!DoConnect(strParams))
	{
		m_pSocket->Destroy();
		m_pSocket = 0;
	}
}

void CApp::CmdConnectHTTP(const std::string& strParams)
{
	// Setup socket
	m_pSocket = new CHTTPSocket;
	if(!DoConnect(strParams))
	{
		m_pSocket->Destroy();
		m_pSocket = 0;
	}
}

bool CApp::DoConnect(const std::string& strParams)
{
	unsigned short wPort = (unsigned short)m_nDefaultPort;

	// Is there colon or space, to siginify a port?
	size_t nPortStart = 0;
	size_t nPos = strParams.find(':');
	if(nPos == std::string::npos)
	{
		nPos = strParams.find(' ');
		if(nPos != std::string::npos)
		{
			nPortStart = nPos;
			while((nPos < strParams.length()) && (strParams[nPos] == ' ')) ++nPos;
		}
	}
	else
	{
		nPortStart = nPos;
		++nPos;
	}

	// Convert port number to an int, and extract host
	std::string strHost;
	if(nPos != std::string::npos)
	{
		std::istringstream is(strParams.substr(nPos));
		is >> wPort;
		if(is.fail())
		{
			CTA2Stream theStream;
			theStream << CTA2Stream::Colour(0xff0000) << "Invalid port: " << CTA2Stream::Reset() <<
				strParams.substr(nPos) << "\r\n" << CTA2Stream::Reset();
			m_deOutput.AddText(theStream.str());
			return false;
		}

		strHost = strParams.substr(0, nPortStart);
	}
	else
		strHost = strParams;

	// Use default values if none specified
	if(strHost.empty())
		strHost = m_strDefaultServer;

	// Connect
	if(!m_pSocket->StartConnect(strHost.c_str(), wPort))
	{
		ShowError("Failed to connect to server. " + std::string(m_pSocket->GetError()));
		m_pSocket->ClearError();
		return false;
	}
	m_bConnecting = true;
	return true;
}

void CApp::CmdDisconnect(const std::string& strParams)
{
	UNREFERENCED_PARAMETER(strParams);

	if(!m_bConnected)
	{
		ShowError("Not connected");
		return;
	}
	else
	{
		ShowError("Disconnected");

		m_bConnected = false;
		m_pSocket->Disconnect();
		OnDisconnected(true);
	}
}

void CApp::CmdLoadScript(const std::string& strParams)
{
	// Display help if there's not enough parameters
	if(strParams.empty())
	{
		CmdHelp("Load");
		return;
	}
	else if(!m_pScriptMgr)
	{
		ShowError("No script system loaded");
		return;
	}

	// Pass off to the script manager
	if (m_pScriptMgr->LoadScript(strParams) != SLR_Success)
	{
		ShowError("Failed to load script: " + m_pScriptMgr->GetError());
	}
	else
	{
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0x00ff00) << "Script loaded\r\n" << CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
	}
}

void CApp::CmdReloadScripts(const std::string& strParams)
{
	// Display help if there's too many parameters
	if(!strParams.empty())
	{
		CmdHelp("Reload");
		return;
	}
	else if(!m_pScriptMgr)
	{
		ShowError("No script system loaded");
		return;
	}

	// Pass off to the script manager
	if(!m_pScriptMgr->ReloadScripts())
	{
		ShowError("Failed to reload scripts: " + m_pScriptMgr->GetError());
	}
	else
	{
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0x00ff00) << "Scripts reloaded\r\n" << CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
	}
}

void CApp::CmdLoadPlugin(const std::string& strParams)
{
	// Display help if there's too many parameters
	if(strParams.empty())
	{
		CmdHelp("load");
		return;
	}

	// Convert plugin name to lower case
	std::string strName(strParams);
	std::transform(strName.begin(), strName.end(), strName.begin(), tolower);

	// See if plugin already exists
	for(PluginListIter it=m_thePlugins.begin(); it!=m_thePlugins.end(); ++it)
	{
		if(it->strName == strName)
		{
			// Plugin already exists
			ShowError("Plugin \"" + strName + "\" already exists. Use /unload to unload it first.");
			return;
		}
	}

	// Plugin doesn't exist
	SetLastError(0);
	if(!LoadPlugin(strParams))
	{
		// Get windows error for if the DLL load fails
		char* szMsgBuffer;
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL, GetLastError(), 0, (LPTSTR)&szMsgBuffer, 0, NULL);
		std::string strError = szMsgBuffer;
		LocalFree(szMsgBuffer);

		// Remove newlines from string
		size_t nPos = strError.find('\r');
		if(nPos != std::string::npos)
			strError.erase(strError.begin()+nPos);
		nPos = strError.find('\n');
		if(nPos != std::string::npos)
			strError.erase(strError.begin()+nPos);

		// Convert "%1" to parameter name
		nPos = strError.find("%1");
		if(nPos != std::string::npos)
		{
			strError.erase(strError.begin()+nPos, strError.begin()+nPos+2);
			strError.insert(nPos, strParams);
		}

		ShowError("Could not load plugin \"" + strName + "\". " + strError);
	}
	else
	{
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0x00ff00) << strName << " loaded OK\r\n"
			<< CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
	}
}

void CApp::CmdUnloadPlugin(const std::string& strParams)
{
	// Display help if there's too many parameters
	if(strParams.empty())
	{
		CmdHelp("unloadplugin");
		return;
	}

	// Convert plugin name to lower case
	std::string strName(strParams);
	std::transform(strName.begin(), strName.end(), strName.begin(), tolower);

	// Find plugin
	for(PluginListIter it=m_thePlugins.begin(); it!=m_thePlugins.end(); ++it)
	{
		if(it->strName == strName)
		{
			UnloadPlugin(*it);

			CTA2Stream theStream;
			theStream << CTA2Stream::Colour(0x00ff00) << strName << " unloaded OK\r\n"
				<< CTA2Stream::Reset();
			m_deOutput.AddText(theStream.str());

			m_thePlugins.erase(it);
			return;
		}
	}

	// Plugin not found
	ShowError("Failed to find plugin:" + strName);
}

void CApp::CmdShell(const std::string& strParams)
{
	if(!m_pScriptMgr)
	{
		ShowError("No script system loaded");
		return;
	}
	if(!m_pScriptMgr->LoadScriptFromString(strParams))
	{
		ShowError("Failed to execute: " + m_pScriptMgr->GetError());
	}
	else
	{
		// Plugin not found
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0x00ff00) << "Exec OK\r\n" << CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
	}
}

void CApp::CmdDump(const std::string& strParams)
{
	UNREFERENCED_PARAMETER(strParams);

	if(!m_pScriptMgr)
	{
		ShowError("No script system loaded");
		return;
	}
	std::string strDump = m_pScriptMgr->DebugDump();
	m_deOutput.AddText(ParseTA2Tags(strDump));
}

void CApp::CmdDownloads(const std::string& strParams)
{
	UNREFERENCED_PARAMETER(strParams);

	CTA2Stream theStream;
	theStream << CTA2Stream::Colour(0x00ff00) << "Active downloads:\r\n";
	bool bAnyDownloads = false;
	for(DownloadListIter it=m_vDownloads.begin(); it!=m_vDownloads.end(); ++it)
	{
		HTTPGet* pGet = *it;
		if(pGet->nUserData != FLAG_StandardGet)
			continue;
		bAnyDownloads = true;

		if(!pGet->bHeaderComplete)
		{
			theStream << "+ " << pGet->strAddress << ": Connecting...\r\n";
		}
		else if(pGet->nContentLength == 0)
		{
			std::ostringstream os;
			int nTransferred = (int)(pGet->nBytesTransferred / 1024);
			int nSpeed = pGet->nTransferSpeed / 1024;
			os << "+ " << pGet->strAddress << ": " << nTransferred << "KB (Unknown size) at " <<
				nSpeed << "KB/s\r\n";
			theStream << os.str();
		}
		else
		{
			std::ostringstream os;
			int nTransferred = (int)(pGet->nBytesTransferred / 1024);
			int nTotal = pGet->nContentLength / 1024;
			int nPercent = (nTransferred * 100) / nTotal;
			int nSpeed = pGet->nTransferSpeed / 1024;
			os << "+ " << pGet->strAddress << ": " << nTransferred << "KB ("
				<< nPercent << "%) at " << nSpeed << "KB/s\r\n";
			theStream << os.str();
		}
	}
	if(!bAnyDownloads)
	{
		theStream << "+ No downloads in progress.\r\n";
	}
	theStream << CTA2Stream::Reset();
	m_deOutput.AddText(theStream.str());
}

void CApp::CmdCheckUpdates(const std::string& strParams)
{
	UNREFERENCED_PARAMETER(strParams);

	// Auto-update check in progress?
	for(DownloadListIter it=m_vDownloads.begin(); it!=m_vDownloads.end(); ++it)
	{
		HTTPGet* pGet = *it;
		if(pGet->nUserData == FLAG_AutoUpdateRequest)
		{
			CTA2Stream theStream;
			theStream << CTA2Stream::Colour(0xffff00) << "Update check already in progress.\r\n" <<
				CTA2Stream::Reset();
			m_deOutput.AddText(theStream.str());
			return;
		}
	}

	// Auto-update check?
	std::string strUpdateCheckURL = m_theConfig.GetValue("UpdateCheckURL");
	if(strUpdateCheckURL.empty())
	{
		ShowError("Failed to check for updates: Invalid \"UpdateCheckURL\" in config file.");
		return;
	}

	// Create HTTP GET request
	std::string strError;
	HTTPGet* pGet = StartHTTPGet(strUpdateCheckURL, "", m_hThreadTerminate, strError);
	if(!pGet)
	{
		ShowError("Failed to check for updates: " + strError);
		return;
	}
	// Mark this as the auto-update request
	pGet->nUserData = FLAG_AutoUpdateRequest;

	// Save in download list
	m_vDownloads.push_back(pGet);
}

void CApp::CmdResetPML(const std::string& strParams)
{
	UNREFERENCED_PARAMETER(strParams);

	m_deOutput.ResetPML();

	CTA2Stream theStream;
	theStream << CTA2Stream::Colour(0x00ff00) << "PML tags reset.\r\n" <<
		CTA2Stream::Reset();
	m_deOutput.AddText(theStream.str());
}

//============================================================================

bool CApp::LoadConfig(CConfig& cfg, const std::string& strFilename)
{
	// Load config
	if(!cfg.Load(strFilename))
		return false;

	// Set DruinkEdit font
	std::string strBuff;
	int nBuff;
	COLORREF crColour = 0;
	strBuff = cfg.GetValue("DefaultFontFace");
	std::istringstream is(cfg.GetValue("DefaultFontSize"));
	is >> nBuff;
	if(!is.fail())
	{
		m_deInput.SetFont(strBuff, nBuff);
		m_deOutput.SetFont(strBuff, nBuff);
	}

	// Set DruinkEdit text colour
	strBuff = cfg.GetValue("DefaultText");
	if(strBuff.length() >= 8 &&
		strBuff.substr(0, 2)=="0x" &&
		GetHexColour(strBuff.c_str()+2, crColour))
	{
		m_deInput.SetDefaultTextColour(crColour);
		m_deOutput.SetDefaultTextColour(crColour);
	}

	// Set DruinkEdit back colour
	strBuff = cfg.GetValue("DefaultBackground");
	if(strBuff.length() >= 8 &&
		strBuff.substr(0, 2)=="0x" &&
		GetHexColour(strBuff.c_str()+2, crColour))
	{
		m_deInput.SetDefaultBackColour(crColour);
		m_deOutput.SetDefaultBackColour(crColour);
	}

	// Set URL formatting
	strBuff = cfg.GetValue("URLTextColour");
	if(strBuff.length() >= 8 &&
		strBuff.substr(0, 2)=="0x" &&
		GetHexColour(strBuff.c_str()+2, crColour))
	{
		m_deOutput.SetURLTextColour(crColour);
	}
	strBuff = cfg.GetValue("URLBackColour");
	if(strBuff.length() >= 8 &&
		strBuff.substr(0, 2)=="0x" &&
		GetHexColour(strBuff.c_str()+2, crColour))
	{
		m_deOutput.SetURLBackColour(crColour);
	}
	strBuff = ParseTA2Tags(cfg.GetValue("URLFormat"));
	BYTE byURLTags = 0;
	for(size_t i=0; i<strBuff.length(); ++i)
		byURLTags |= strBuff[i];
	if(!strBuff.empty())
		m_deOutput.SetURLFormat(byURLTags);

	// Set URL prefixes
	strBuff = m_theConfig.GetValue("URLPrefixes");
	if(!strBuff.empty())
		SetURLPrefixes(strBuff);

	// Set default server and port
	strBuff = m_theConfig.GetValue("DefaultServer");
	if(!strBuff.empty())
		m_strDefaultServer = strBuff;
	{
		unsigned short wPort;
		std::istringstream is(m_theConfig.GetValue("DefaultPort"));
		is >> wPort;
		if(!is.fail())
			m_nDefaultPort = wPort;
	}

	// Set tick timer frequency
	int nDelay;
	{
		std::istringstream is(cfg.GetValue("TickDelay"));
		is >> nDelay;
		if(!is.fail())
			m_nTickDelay = nDelay;
	}

	// Set search URL
	std::string str = cfg.GetValue("SearchURLPrefix");
	if(!str.empty())
		m_strSearchURL = str;

	// Window size and pos
	m_bWindowRectValid = true;
	{
		std::istringstream is(cfg.GetValue("Window.X"));
		is >> m_rcWindowInitial.left;
		if(is.fail())
			m_bWindowRectValid = false;
	}
	{
		std::istringstream is(cfg.GetValue("Window.Y"));
		is >> m_rcWindowInitial.top;
		if(is.fail())
			m_bWindowRectValid = false;
	}
	{
		std::istringstream is(cfg.GetValue("Window.W"));
		is >> m_rcWindowInitial.right;
		if(is.fail())
			m_bWindowRectValid = false;
		else
			m_rcWindowInitial.right += m_rcWindowInitial.left;
	}
	{
		std::istringstream is(cfg.GetValue("Window.H"));
		is >> m_rcWindowInitial.bottom;
		if(is.fail())
			m_bWindowRectValid = false;
		else
			m_rcWindowInitial.bottom += m_rcWindowInitial.top;
	}

	// Done
	return true;
}

//============================================================================

void CApp::LoadAllPlugins()
{
	// Search for all DLLs in the current path
	WIN32_FIND_DATA theFind;
	HANDLE hFind = FindFirstFile("*.dll", &theFind);
	if(hFind == INVALID_HANDLE_VALUE)
		return;

	do {
		if(LoadPlugin(theFind.cFileName))
		{
			CTA2Stream theStream;
			theStream << CTA2Stream::Colour(0x00ff00) << "Loaded plugin library \"" << theFind.cFileName << "\".\r\n"
				<< CTA2Stream::Reset();
			m_deOutput.AddText(theStream.str());
		}
	} while(FindNextFile(hFind, &theFind));
	FindClose(hFind);
}

bool CApp::LoadPlugin(const std::string& strName)
{
	if(!m_pScriptMgr)
	{
		ShowError("No script system loaded");
		return false;
	}

	// Load the DLL
	HMODULE hDll = LoadLibrary(strName.c_str());
	if(!hDll)
		return false;

	// Find the "GetNumExportedFunctions" function
	LPGetNumExportedFunctions pfnGetNumExportedFunctions =
		(LPGetNumExportedFunctions)GetProcAddress(hDll, "GetNumExportedFunctions");
	if(!pfnGetNumExportedFunctions)
	{
		DWORD dwError = GetLastError();
		FreeLibrary(hDll);
		SetLastError(dwError);
		return false;
	}

	// Find the "GetExportedFunction" function
	LPGetExportedFunction pfnGetExportedFunction =
		(LPGetExportedFunction)GetProcAddress(hDll, "GetExportedFunction");
	if(!pfnGetExportedFunction)
	{
		DWORD dwError = GetLastError();
		FreeLibrary(hDll);
		SetLastError(dwError);
		return false;
	}

	// Fill in plugin struct
	Plugin thePlugin;
	thePlugin.hDll = hDll;
	thePlugin.strName = strName;

	// Trim off ".dll" and convert to lower case
	size_t nPos = thePlugin.strName.rfind('.');
	if(nPos != std::string::npos)
		thePlugin.strName.erase(thePlugin.strName.begin()+nPos, thePlugin.strName.end());
	std::transform(thePlugin.strName.begin(), thePlugin.strName.end(), thePlugin.strName.begin(), tolower);

	// Register script functions
	size_t nFunctions = pfnGetNumExportedFunctions(PLUGINAPI_VERSION);
	for(size_t i=0; i<nFunctions; ++i)
	{
		ScriptFunction func;
		func.pfnFunction = NULL;
		func.pParam = NULL;
		func.szName[0] = 0;
		pfnGetExportedFunction(i, func);
		if(m_pScriptMgr->Register(func))
		{
			thePlugin.vFunctions.push_back(func.szName);
		}
	}
	m_thePlugins.push_back(thePlugin);
	return true;
}

void CApp::UnloadPlugin(const Plugin& plugin)
{
	if(!m_pScriptMgr)
	{
		ShowError("No script system loaded");
		return;
	}

	for(size_t i=0; i<plugin.vFunctions.size(); ++i)
	{
		m_pScriptMgr->Unregister(plugin.vFunctions[i]);
	}
	FreeLibrary(plugin.hDll);
}

//============================================================================

void CApp::SaveURL(const char* szUrl)
{
	// Determine file extension
	std::string strUrl, strExtension;
	if(strncmp(szUrl, "http://", 7) == 0)
		strUrl = szUrl+7;
	else
		strUrl = szUrl;
	size_t nPos = strUrl.rfind('/');
	if(nPos != std::string::npos)
	{
		nPos = strUrl.find('.', nPos+1);
		if(nPos != std::string::npos)
			strExtension = strUrl.substr(nPos);
	}

	// Get file type from extension and build filter string
	std::string strFiletype = GetFileType(strExtension);
	char szFilter[512];
	char* pNext = szFilter;
	if(!strFiletype.empty())
	{
		strcpy(pNext, strFiletype.c_str());
		pNext += strFiletype.length();
		strcpy(pNext, " (*");
		pNext += 3;
		strcpy(pNext, strExtension.c_str());
		pNext += strExtension.length();
		memcpy(pNext, ")\0*", 3);
		pNext += 3;
		strcpy(pNext, strExtension.c_str());
		pNext += strExtension.length()+1;
	}
	memcpy(pNext, "All Files\0*.*\0", 15);

	// Get filename
	OPENFILENAME ofn;
	char szFilename[512];
	szFilename[0] = 0;
	memset(&ofn, 0, sizeof(ofn));
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = m_hWnd;
	ofn.hInstance = NULL;
	ofn.lpstrFilter = szFilter;
	ofn.lpstrFile = szFilename;
	ofn.nMaxFile = sizeof(szFilename)-1;
	ofn.Flags = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;
	ofn.lpstrDefExt = strExtension.c_str()+1;
	if(!GetSaveFileName(&ofn))
		return;

	// Create HTTP GET request
	std::string strError;
	HTTPGet* pGet = StartHTTPGet(szUrl, szFilename, m_hThreadTerminate, strError);
	if(!pGet)
	{
		ShowError("Failed to save URL: " + strError);
		return;
	}
	pGet->nUserData = FLAG_StandardGet;

	// Save in download list
	m_vDownloads.push_back(pGet);
	CTA2Stream theStream;
	theStream << CTA2Stream::Colour(0x00ff00) <<
		"Download in progress. Type /downloads to view current downloads.\r\n" << CTA2Stream::Reset();
	m_deOutput.AddText(theStream.str());
}

//============================================================================

std::string CApp::GetFileType(const std::string& strExtension)
{
	// Lookup file extension to get content type key
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CLASSES_ROOT, strExtension.c_str(), 0, KEY_READ, &hKey))
		return "";
	char szBuff[256];
	DWORD dwSize = (DWORD)sizeof(szBuff);
	if(RegQueryValueEx(hKey, NULL, NULL, NULL, (BYTE*)szBuff, &dwSize)!=0 || dwSize==0)
	{
		RegCloseKey(hKey);
		return "";
	}
	RegCloseKey(hKey);

	// Look up content type key to get string name
	if(RegOpenKeyEx(HKEY_CLASSES_ROOT, szBuff, 0, KEY_READ, &hKey))
		return "";
	dwSize = (DWORD)sizeof(szBuff);
	if(RegQueryValueEx(hKey, NULL, NULL, NULL, (BYTE*)szBuff, &dwSize)!=0 || dwSize==0)
	{
		RegCloseKey(hKey);
		return "";
	}
	RegCloseKey(hKey);
	return szBuff;
}

//============================================================================

bool CApp::ScriptFlashWindow(ScriptContext&, void* pParam)
{
	CApp* pApp = (CApp*)pParam;
	FLASHWINFO flashInfo;
	flashInfo.cbSize = sizeof(flashInfo);
	flashInfo.hwnd = pApp->m_hWnd;
	flashInfo.dwFlags = FLASHW_ALL | FLASHW_TIMERNOFG;
	flashInfo.uCount = 0;
	flashInfo.dwTimeout = 0;
	FlashWindowEx(&flashInfo);
	return true;
}

bool CApp::ScriptShowWindow(ScriptContext&, void* pParam)
{
	CApp* pApp = (CApp*)pParam;
	AllowSetForegroundWindow(GetCurrentProcessId());
	SetForegroundWindow(pApp->m_hWnd);
	return true;
}

bool CApp::ScriptPlaySound(ScriptContext& context, void*)
{
	if(context.GetNumArgs() != 1)
		return false;
	ScriptVar var = context.GetArg(0);
	if(var.GetType() != ScriptVar::TYPE_String)
		return false;

	PlaySound(var.ToString(), NULL, SND_FILENAME | SND_ASYNC);
	return true;
}

bool CApp::ScriptBeep(ScriptContext&, void*)
{
	MessageBeep(MB_OK);
	return true;
}

bool CApp::ScriptPrint(ScriptContext& context, void* pParam)
{
	CApp* pApp = (CApp*)pParam;
	size_t nArgs = context.GetNumArgs();
	for(size_t i=0; i<nArgs; ++i)
	{
		ScriptVar text = context.GetArg(i);
		if(text.GetType() != ScriptVar::TYPE_String)
			continue;

		pApp->m_deOutput.AddText(ParseTA2Tags(text.ToString()));
	}
	return true;
}

bool CApp::ScriptPrintLn(ScriptContext& context, void* pParam)
{
	CApp* pApp = (CApp*)pParam;
	size_t nArgs = context.GetNumArgs();
	for(size_t i=0; i<nArgs; ++i)
	{
		ScriptVar text = context.GetArg(i);
		if(text.GetType() != ScriptVar::TYPE_String)
			continue;

		pApp->m_deOutput.AddText(ParseTA2Tags(text.ToString()));
	}
	CTA2Stream theStream;
	theStream << "\r\n" << CTA2Stream::Reset();
	pApp->m_deOutput.AddText(theStream.str());
	return true;
}

bool CApp::ScriptSend(ScriptContext& context, void* pParam)
{
	size_t nArgs = context.GetNumArgs();
	std::string str;
	for(size_t i=0; i<nArgs; ++i)
	{
		ScriptVar text = context.GetArg(i);
		if(text.GetType() != ScriptVar::TYPE_String)
			continue;

		str += text.ToString();
	}

	CApp* pApp = (CApp*)pParam;
	pApp->Send(str);
	return true;
}

bool CApp::ScriptSendLn(ScriptContext& context, void* pParam)
{
	size_t nArgs = context.GetNumArgs();
	std::string str;
	for(size_t i=0; i<nArgs; ++i)
	{
		ScriptVar text = context.GetArg(i);
		if(text.GetType() != ScriptVar::TYPE_String)
			continue;

		str += text.ToString();
	}

	CApp* pApp = (CApp*)pParam;
	pApp->Send(str + "\r\n");
	return true;
}

bool CApp::ScriptClear(ScriptContext&, void* pParam)
{
	CApp* pApp = (CApp*)pParam;
	pApp->m_deOutput.Clear();
	return true;
}

bool CApp::ScriptIsConnected(ScriptContext& context, void* pParam)
{
	CApp* pApp = (CApp*)pParam;
	context.AddRetval(pApp->m_bConnected ? 1 : 0);
	return true;
}

bool CApp::ScriptGetVersion(ScriptContext& context, void*)
{
	char szBuff[256];
	sprintf_s(szBuff, sizeof(szBuff), s_szClientVersion, ms_nVersionMajor, ms_nVersionMinor);
	context.AddRetval(szBuff);
	return true;
}

bool CApp::ScriptGetConfigValue(ScriptContext& context, void* pParam)
{
	if(context.GetNumArgs() != 1)
	{
		context.SetError("getConfigValue expects 1 parameter");
		return false;
	}
	ScriptVar key = context.GetArg(0);
	if(key.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("getConfigValue expects a string as the first parameter");
		return false;
	}

	// Try the extra config file first, then fall back to the main config
	CApp* pApp = (CApp*)pParam;
	std::string strBuff = pApp->m_theExtraConfig.GetValue(key.ToString());
	if(strBuff.empty())
		strBuff = pApp->m_theConfig.GetValue(key.ToString());
	context.AddRetval(strBuff.c_str());
	return true;
}

bool CApp::ScriptSetConfigValue(ScriptContext& context, void* pParam)
{
	if(context.GetNumArgs() != 2)
	{
		context.SetError("setConfigValue expects 2 parameters");
		return false;
	}
	ScriptVar key = context.GetArg(0);
	ScriptVar value = context.GetArg(1);
	if(key.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("setConfigValue expects a string as the first parameter");
		return false;
	}
	if(value.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("setConfigValue expects a string as the second parameter");
		return false;
	}


	CApp* pApp = (CApp*)pParam;
	pApp->m_theExtraConfig.SetValue(key.ToString(), value.ToString());
	pApp->m_bConfigDirty = true;
	return true;
}

bool CApp::ScriptEnterHighPerformanceMode(ScriptContext& context, void* pParam)
{
	CApp* pApp = (CApp*)pParam;

	if(context.GetNumArgs() != 0)
		context.SetError("enterHighPerfMode expects no parameters. Parameters ignored");

	++pApp->m_nHighPerfModeCounter;
	return true;
}

bool CApp::ScriptExitHighPerformanceMode(ScriptContext &context, void *pParam)
{
	CApp* pApp = (CApp*)pParam;

	if(context.GetNumArgs() != 0)
		context.SetError("exitHighPerfMode expects no parameters. Parameters ignored");

	if(pApp->m_nHighPerfModeCounter > 0)
		--pApp->m_nHighPerfModeCounter;
	return true;
}

bool CApp::ScriptSetPasswordInput(ScriptContext &context, void *pParam)
{
	if(context.GetNumArgs() != 1)
	{
		context.SetError("setPasswordInput expects 1 parameter");
		return false;
	}
	ScriptVar isPassword = context.GetArg(0);
	if(isPassword.GetType() != ScriptVar::TYPE_Int && isPassword.GetType() != ScriptVar::TYPE_Float)
	{
		context.SetError("setPasswordInput expects a boolean as the first parameter");
		return false;
	}

	CApp* pApp = (CApp*)pParam;
	pApp->m_deInput.SetIsPassword(isPassword.ToInt() != 0);
	return true;
}

//============================================================================

void CApp::OnDownloadComplete(HTTPGet* pGet)
{
	// Auto-update download?
	if(pGet->nUserData == FLAG_AutoUpdateRequest)
	{
		if(!pGet->strError.empty())
		{
			ShowError("Error checking for updates: " + pGet->strError);
		}
		else if(pGet->nResponseCode != 200)
		{
			CTA2Stream theStream;
			theStream << CTA2Stream::Colour(0xff0000) <<
				"Error checking for updates: Remote server responsed with:\r\n" <<
				CTA2Stream::Colour(0xffff00) << pGet->strHeader << "\r\n" << CTA2Stream::Reset();
			m_deOutput.AddText(theStream.str());
		}
		else
		{
			OnCheckForUpdates(pGet->strContent);
		}
		return;
	}

	// Download complete. Display message
	CTA2Stream theStream;
	if(!pGet->strError.empty())
	{
		ShowError("Error downloading \"" + pGet->strAddress + "\": " + pGet->strError);
	}
	else if(pGet->nResponseCode != 200)
	{
		theStream << CTA2Stream::Colour(0xff0000) << "Error downloading \"" <<
			pGet->strAddress << "\": Remote server responsed with:\r\n" <<
			CTA2Stream::Colour(0xffff00) << pGet->strHeader << "\r\n" << CTA2Stream::Reset();
	}
	else
	{
		theStream << CTA2Stream::Colour(0x00ff00) << "\"" <<
			pGet->strAddress << "\" downloaded successfully\r\n" << CTA2Stream::Reset();
	}
	m_deOutput.AddText(theStream.str());
}

void CApp::OnCheckForUpdates(const std::string& strVersionFile)
{
	// Get the first line of the version file (Current file version)
	size_t nPos = strVersionFile.find('\n');
	if(nPos == std::string::npos)
	{
		ShowError("Error checking for updates: Remote server responsed with incorrectly "
			"formatted response file");
		return;
	}
	std::string strVersion = RemoveWhitespace(strVersionFile.substr(0, nPos));
	std::string strPaths = strVersionFile.substr(nPos+1);

	// Find the . in the version file
	nPos = strVersion.find('.');
	if(nPos == std::string::npos)
	{
		ShowError("Error checking for updates: Remote server responsed with incorrectly "
			"formatted response file");
		return;
	}

	// Extract major and minor version
	int nMajor = atoi(strVersion.substr(0, nPos).c_str());
	int nMinor = atoi(strVersion.substr(nPos+1).c_str());
	if(nMajor <= ms_nVersionMajor && nMinor <= ms_nVersionMinor)
	{
		CTA2Stream theStream;
		theStream << CTA2Stream::Colour(0x00ff00) << "No new updates available.\r\n" <<
			CTA2Stream::Reset();
		m_deOutput.AddText(theStream.str());
#ifdef _DEBUG
		if(nMajor < ms_nVersionMajor || nMinor < ms_nVersionMinor)
		{
			CTA2Stream theStream;
			theStream << CTA2Stream::Colour(0xff0000) << "Server version is lower than this version. "
				"Need to update version number on server!\r\n" <<
				CTA2Stream::Reset();
			m_deOutput.AddText(theStream.str());
		}
#endif // _DEBUG
		return;
	}

	// New version available, get URLs of download locations
	nPos = strPaths.find('\n');
	if(nPos == std::string::npos)
	{
		ShowError("Error checking for updates: Remote server responsed with incorrectly "
			"formatted response file");
		return;
	}
	std::string strPathx86 = RemoveWhitespace(strPaths.substr(0, nPos));
	std::string strPathx64 = RemoveWhitespace(strPaths.substr(nPos+1));

	#ifdef BUILD_X64
		std::string strPath = strPathx64;
	#else
		std::string strPath = strPathx86;
	#endif

	CTA2Stream theStream;
	theStream << CTA2Stream::Colour(0x00ff00) << "A new version of DruinkClient (Version " <<
		strVersion << ") is now available to download from " << strPath << "\r\n" << CTA2Stream::Reset();
	m_deOutput.AddText(theStream.str());
	return;
}

//============================================================================

void CApp::ShowError(const std::string& strError)
{
	CTA2Stream theStream;
	theStream << CTA2Stream::Colour(0xff0000) << strError << "\r\n" << CTA2Stream::Reset();
	m_deOutput.AddText(theStream.str());
}

//============================================================================

void CApp::SetURLPrefixes(const std::string& strPrefixes)
{
	std::vector<std::string> vPrefixes;
	std::string strBuff(strPrefixes);

	while(!strBuff.empty())
	{
		// Trim leading whitespace
		while(!strBuff.empty() && strBuff[0]==' ')
			strBuff.erase(strBuff.begin());

		// Find start of URL prefix
		if(strBuff.empty() || strBuff[0]!='\"')
		{
			ShowError("Invalid URLPrefixes config option");
			return;
		}

		// Find end of URL prefix
		size_t nEnd = strBuff.find('\"', 1);
		if(nEnd == std::string::npos)
		{
			ShowError("Invalid URLPrefixes config option");
			return;
		}

		// Extract it
		vPrefixes.push_back(strBuff.substr(1, nEnd-1));
		strBuff.erase(strBuff.begin(), strBuff.begin()+nEnd+1);

		// Trim leading whitespace
		while(!strBuff.empty() && strBuff[0]==' ')
			strBuff.erase(strBuff.begin());

		// Is there a comma?
		if(!strBuff.empty() && strBuff[0]==',')
			strBuff.erase(strBuff.begin());
	}

	if(!vPrefixes.empty())
		m_deOutput.SetURLPrefixes(vPrefixes);
}

//============================================================================
