//============================================================================
// App.h - Main application class
//============================================================================

#ifndef __APP_H__
#define __APP_H__

#include "Window/Window.h"
#include "DruinkEdit/DruinkEdit.h"
#include "Config/Config.h"
#include "ScriptManager/ScriptManager.h"
#include "zlib/zlib.h"
#include <vector>
#include <list>
#include <map>

class ScriptContext;
class CBaseSocket;
struct HTTPGet;

//============================================================================
// Derived application
class CApp : public CWindow
{
public:
	CApp();
	virtual ~CApp();

	virtual bool Init(HINSTANCE hInstance);
	virtual void Shutdown();

	static CApp* GetInstance() { return sm_pInstance; }

	void SetIcon(HICON hIcon);
	void SetTitle(const char* szTitle);

protected:
	virtual LRESULT WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam);
	virtual void Tick();
	virtual HICON GetIcon();
	virtual HCURSOR GetCursor();
	virtual bool InHighPerfMode() const { return m_nHighPerfModeCounter>0; }
	virtual void OnNotify(NMHDR* pHdr);

private:
	void Send(const std::string& str);
	void RawSend(const std::string& str);
	void SocketSend(const std::string& str);
	void SocketSend(const BYTE* pData, size_t len);

	// Called from Tick()
	void TickConnecting();
	void TickSocket();

	// Misc. internal functions
	void SendLineImpl(const std::string& str);
	void ProcessCommand(const std::string& str);
	bool ProcessTelnet(std::vector<char>& data);

	// Telnet negotiation handlers
	void OnTelnetWillOrDo(BYTE cmd, bool _will);
	void OnTelnetWontOrDont(BYTE cmd, bool _wont);
	void OnCompressInputChanged(bool compressed, size_t socketBuffOffset);
	void OnCompressOutputChanged(bool compressed, bool gracefulFinish);
	void DecompressPendingData();
	std::string GetLine(std::vector<char>& data);
	void SendInitialNegotiation();
	void SendNAWS(bool sendNow=false);

	// Various event handlers
	std::string OnPreparseText(const std::string& str);
	std::string OnSendLine(const std::string& str);
	void OnConnected();
	void OnConnectFailed();
	void OnDisconnected(bool bUserRequested);

	// Exposed script functions
	static bool ScriptFlashWindow(ScriptContext& context, void* pParam);
	static bool ScriptShowWindow(ScriptContext& context, void* pParam);
	static bool ScriptPlaySound(ScriptContext& context, void* pParam);
	static bool ScriptBeep(ScriptContext& context, void* pParam);
	static bool ScriptPrint(ScriptContext& context, void* pParam);
	static bool ScriptPrintLn(ScriptContext& context, void* pParam);
	static bool ScriptSend(ScriptContext& context, void* pParam);
	static bool ScriptSendLn(ScriptContext& context, void* pParam);
	static bool ScriptClear(ScriptContext& context, void* pParam);
	static bool ScriptIsConnected(ScriptContext& context, void* pParam);
	static bool ScriptGetVersion(ScriptContext& context, void* pParam);
	static bool ScriptGetConfigValue(ScriptContext& context, void* pParam);
	static bool ScriptSetConfigValue(ScriptContext& context, void* pParam);
	static bool ScriptEnterHighPerformanceMode(ScriptContext &context, void* pParam);
	static bool ScriptExitHighPerformanceMode(ScriptContext &context, void* pParam);
	static bool ScriptSetPasswordInput(ScriptContext &context, void* pParam);

private:
	// std::map of commands
	typedef void (CApp::*LPCommand)(const std::string& strParams);
	struct Command
	{
		Command() {}
		Command(LPCommand _pfn, const std::string& str) : pfn(_pfn), strHelp(str) {}

		LPCommand pfn;
		std::string strHelp;
	};

	// Plugin functions struct
	struct Plugin
	{
		HMODULE hDll;
		std::string strName;
		std::vector<std::string> vFunctions;
	};

	struct TelnetOption
	{
		TelnetOption() { reset(); }
		bool isNegotiated() const { return op != 0; }
		void reset() { op=0; }

		BYTE op; // DO, DON'T, WILL, WON'T
	};

	typedef std::map<std::string, Command> CommandList;
	typedef CommandList::iterator CommandListIter;
	typedef std::vector<Plugin> PluginList;
	typedef PluginList::iterator PluginListIter;
	typedef std::list<HTTPGet*> DownloadList;
	typedef DownloadList::iterator DownloadListIter;

	static const UINT_PTR FLAG_StandardGet = 0x00000000;
	static const UINT_PTR FLAG_AutoUpdateRequest = 0x00000001;

	static const BYTE TELNETOP_NAWS = 31;
	static const BYTE TELNETOP_COMPRESS = 85;
	static const BYTE TELNETOP_COMPRESS2 = 86;

	static const BYTE TELNET_SE = 0xf0;
	static const BYTE TELNET_SB = 0xfa;
	static const BYTE TELNET_WILL = 0xfb;
	static const BYTE TELNET_WONT = 0xfc;
	static const BYTE TELNET_DO = 0xfd;
	static const BYTE TELNET_DONT = 0xfe;
	static const BYTE TELNET_IAC = 0xff;

	//========================================================================

	// Commands
	void CmdHelp(const std::string& strParams);
	void CmdTelnet(const std::string& strParams);
	void CmdConnect(const std::string& strParams);
	void CmdConnectHTTP(const std::string& strParams);
	void CmdDisconnect(const std::string& strParams);
	void CmdLoadScript(const std::string& strParams);
	void CmdReloadScripts(const std::string& strParams);
	void CmdLoadPlugin(const std::string& strParams);
	void CmdUnloadPlugin(const std::string& strParams);
	void CmdShell(const std::string& strParams);
	void CmdDump(const std::string& strParams);
	void CmdDownloads(const std::string& strParams);
	void CmdCheckUpdates(const std::string& strParams);
	void CmdResetPML(const std::string& strParams);

	bool DoConnect(const std::string& strParams);
	bool LoadConfig(CConfig& cfg, const std::string& strFilename);
	void LoadAllPlugins();
	bool LoadPlugin(const std::string& strName);
	void UnloadPlugin(const Plugin& plugin);

	void SaveURL(const char* szUrl);
	std::string GetFileType(const std::string& strExtension);
	void OnDownloadComplete(HTTPGet* pGet);
	void OnCheckForUpdates(const std::string& strVersionFile);
	void ShowError(const std::string& strError);
	void SetURLPrefixes(const std::string& strPrefixes);

	//========================================================================

	CScriptManager*	m_pScriptMgr;
	HANDLE			m_hThreadTerminate;

	int				m_nInputHeight;
	int				m_yClickOfs;
	int				m_nHighPerfModeCounter;
	static int		ms_nVersionMajor;
	static int		ms_nVersionMinor;

	CommandList		m_theCommands;
	CConfig			m_theConfig;
	CConfig			m_theExtraConfig;
	bool			m_bConfigDirty;
	PluginList		m_thePlugins;

	CDruinkEdit		m_deOutput;			// History window
	CDruinkEdit		m_deInput;			// Input window

	CBaseSocket*	m_pSocket;			// Socket connection to the server
	bool			m_bConnecting;		// Are we connecting to a server?
	bool			m_bConnected;		// Do we think we're connected?

	RECT			m_rcDivider;		// Area of the divider between input and output windows

	DownloadList	m_vDownloads;

	// Telnet negotiation stuff
	TelnetOption	m_respondedOptions[256];
	bool			m_enableNAWS;
	bool			m_sentNegotiation;
	bool			m_compressedInput;
	bool			m_compressedOutput;
	z_stream		m_inputZip;
	z_stream		m_outputZip;
	std::vector<char> m_socketBuffer;
	std::vector<char> m_decompressionBuffer;

	// Window size negotiation
	int				m_nawsWidth;		// Last width sent through NAWS
	int				m_nawsHeight;		// Last height sent through NAWS
	time_t			m_lastNawsTime;
	bool			m_nawsDirty;

	// Config options
	std::string		m_strSettingsPath;
	std::string		m_strDefaultServer;
	int				m_nDefaultPort;
	std::string		m_strSearchURL;
	int				m_nTickDelay;
	RECT			m_rcWindowInitial;
	bool			m_bWindowRectValid;

	// Configurable values
	HICON			m_hIcon;

	// Static instance pointer
	static CApp*	sm_pInstance;
};

#endif // __APP_H__
