//==========================================================================//
// Mutex.h - Critical Section object wrapper								//
//==========================================================================//

#ifndef __MUTEX_H__
#define __MUTEX_H__

#include <windows.h>
#include "Plugin/Assert.h"

//==========================================================================//
// Main class																//
//==========================================================================//
class Mutex
{
public:
	class Lock
	{
	public:
#ifdef _DEBUG
		Lock(Mutex& rMutex, const char* szFunc, const char* szFile, int nLine) :
			m_rMutex(rMutex) {rMutex.Enter(szFunc, szFile, nLine);}
#else
		Lock(Mutex& rMutex) : m_rMutex(rMutex) {rMutex.Enter();}
#endif
		~Lock() {m_rMutex.Leave();}

	private:
		Lock(const Lock& rhs) : m_rMutex(rhs.m_rMutex) {}
		Lock& operator=(const Lock&) {return *this;}

		Mutex& m_rMutex;
	};

public:
	void Init()
	{
		InitializeCriticalSection(&m_cs);
#ifdef _DEBUG
		m_dwHoldingThread = 0;
		m_szFunc = "";
		m_szFile = "";
		m_nLine = 0;
#endif
	}

	void Delete()
	{
		DeleteCriticalSection(&m_cs);
	}

#ifdef _DEBUG
	void Enter(const char* szFunc, const char* szFile, int nLine)
	{
		Assert(m_dwHoldingThread != GetCurrentThreadId(), "Mutex entered twice");
		EnterCriticalSection(&m_cs);
		m_dwHoldingThread = GetCurrentThreadId();
		m_szFunc = szFunc;
		m_szFile = szFile;
		m_nLine = nLine;
	}
#else

	void Enter()
	{
		EnterCriticalSection(&m_cs);
	}
#endif

	void Leave()
	{
#ifdef _DEBUG
		Assert(GetCurrentThreadId() == m_dwHoldingThread,
			"Leaving mutex from other thread than one that entered");
		m_dwHoldingThread = 0;
		m_szFunc = "";
		m_szFile = "";
		m_nLine = 0;
#endif
		LeaveCriticalSection(&m_cs);
	}

#ifdef _DEBUG
	bool IsHeld() const {return m_dwHoldingThread != 0;}
#else
	bool IsHeld() const {return false;}
#endif

private:
	CRITICAL_SECTION	m_cs;
#ifdef _DEBUG
	DWORD				m_dwHoldingThread;
	const char*			m_szFunc;
	const char*			m_szFile;
	int					m_nLine;
#endif
};

#ifdef _DEBUG
#	define MUTEX_LOCK(mut)		Mutex::Lock theLock(mut, __FUNCTION__, __FILE__, __LINE__)
#	define MUTEX_ENTER(mut)		(mut).Enter(__FUNCTION__, __FILE__, __LINE__);
#else
#	define MUTEX_LOCK(mut)		Mutex::Lock theLock(mut)
#	define MUTEX_ENTER(mut)		(mut).Enter();
#endif

#endif // __MUTEX_H__
