//============================================================================
// DruinkEdit.h - Overridden RichEdit window class
//============================================================================

#ifndef __DRUINKEDIT_H__
#define __DRUINKEDIT_H__

#include "ShellExecuteThread.h"
#include <windows.h>
#include <list>
#include <string>
#include <vector>

#define TA2BIT				0x80
#define TA2BIT_ENABLE		0x40
#define TA2BIT_BOLD			0x20
#define TA2BIT_ITALIC		0x10
#define TA2BIT_UNDERLINE	0x08
#define TA2BIT_TEXTCOLOUR	0x04
#define TA2BIT_STRIKEOUT	0x02
#define TA2BIT_URL			0x01	// Used internally

//============================================================================
// WM_NOTIFY messages / structures
struct DruinkEditNotify
{
	NMHDR hdr;
	UINT_PTR nParam;
	UINT_PTR nParam2;
};

// Enter was pressed on an input DruinkEdit control
#define DEN_NEWLINE						0x00000001

// A key was pressed on an output DruinkEdit control
// nParam: The virtual key code of the key pressed
// nParam2: The LPARAM passed to WM_KEYDOWN
#define DEN_ISREADONLY					0x00000002

// "Google" was selected.
// nParam: The hilighted text, as a const char*
#define DEN_GOOGLE						0x00000003

// "Save URL" was selected.
// nParam: The URL, as a const char*
#define DEN_SAVEURL						0x00000004

//============================================================================
// Main class
class CDruinkEdit
{
public:
	CDruinkEdit();
	~CDruinkEdit() {Release();}
	void Release();

	// Return the ideal width for this control (100 characters in current font)
	int GetIdealWidth();

	// Only valid to call before control is created
	void SetDefaultTextColour(COLORREF crText);
	void SetDefaultBackColour(COLORREF crBack);

	// General functions
	bool Create(HINSTANCE hInstance, HWND hWndParent, DWORD x, DWORD y,
		DWORD w, DWORD h, bool bReadOnly);
	bool AddText(const std::string& strText, bool bInsert=false);
	void Clear();

	// Set current font or URL formatting
	void SetFont(const std::string& strFont, int nSize);
	void SetURLTextColour(COLORREF crText);
	void SetURLBackColour(COLORREF crBack);
	void SetURLFormat(BYTE byTA2Bits);
	void SetURLPrefixes(const std::vector<std::string>& vPrefixes);

	// Set the control as a password field - doesn't store history, and will obscure
	// characters
	void SetIsPassword(bool isPassword);

	// Pause/unpause output
	void PauseOutput(bool bPause);

	// Get/set text and next PML tag extracted
	std::string GetText() const;
	void SetText(const std::string& str);
	std::string GetNextPMLTag();
	void ResetPML();

	// Accessors
	HWND GetWindow() const {return m_hWnd;}
	bool HasPMLWaiting() const {return !m_vPMLTags.empty();}
	bool IsFocused() const { return m_bFocused; }

	// Helper functions
	std::string ConvertRTFString(const std::string& str);

	// Make a ShellExecute() call - threaded so as not to block main thread
	void DoShellExecute(const std::string& strCmd);

	// "Fake" a keypress
	void DoKeypress(WPARAM nChar, LPARAM nParams);

	// Process a WM_COMMAND message
	void OnCommand(WORD nID);

	// Get the width and height of this window, in characters
	SIZE GetSizeInCharacters() const;

private:
	// Window proc
	static LRESULT CALLBACK StaticWindowProc(HWND hWnd, UINT uMsg,
		WPARAM wParam, LPARAM lParam);
	LRESULT WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam);

	// Message handlers
	void OnDestroy();
	void OnEraseBackground(HDC hDC);
	bool OnKeyDown(int nVirtKey, LPARAM lKeyData);
	bool OverURL(WORD x, WORD y);

	// Internal functions
	void AddTextImpl(const std::string& strText, int nInsertPoint, bool bSetSelection);
	void AddTextImpl(const std::wstring& strText, int nInsertPoint, bool bSetSelection);
	void ProcessURLsInBuffer();
	void ShowPopupMenu(int x, int y);
	void InitPopupMenu(HMENU hMenu);
	bool Copy(bool bCopyTaggedPlaintext=false);
	bool CopyURL();
	bool Paste(bool bPlaintextOnly=false);
	bool GoogleSearch();
	void OnFormat(char chFormat); // chFormat should be 'B', 'I', 'U', 'S'
	void UpdateCursor();
	void OnWindowPosChanged(WINDOWPOS* pNewPos);
	static DWORD GetBaseWindowStyle();
	static DWORD GetThisWindowStyle(bool bReadOnly);
	bool IsHTTPUrl(const char* szUrl);

private:
	static const int ms_nURLBufferSize = 1024;

	static bool		s_bLoadedDlls;
	HINSTANCE		m_hInstance;
	HWND			m_hWndParent;
	HWND			m_hWnd;
	HWND			m_hWndPasteBuffer;
	WNDPROC			m_pfnOldProc;
	bool			m_bReadOnly;
	std::string		m_strFont;
	int				m_nFontSize;
	bool			m_bOutputHeld;
	bool			m_bMinimized;
	bool			m_bFocused;
	bool			m_bPassword;

	COLORREF		m_crDefaultText;
	COLORREF		m_crDefaultBack;
	HBRUSH			m_hBackgroundBrush;
	std::string		m_strBuffer;

	bool			m_bOverURL;
	HCURSOR			m_hDefaultCursor;
	HCURSOR			m_hLinkCursor;
	HCURSOR			m_hArrowCursor;
	char			m_szURL[ms_nURLBufferSize];
	DWORD			m_dwFormatBeforeURL;
	COLORREF		m_crColourBeforeURLText;
	COLORREF		m_crColourBeforeURLBack;
	DWORD			m_dwURLFormat;
	COLORREF		m_crURLText;
	COLORREF		m_crURLBack;

	UINT			m_nRTFFormatID;
	bool			m_bMenuVisible;

	std::vector<std::string>	m_vURLPrefixes;
	std::vector<std::string>	m_vPMLTags;

	std::list<std::string>				m_liBuffer;
	std::list<std::string>::iterator	m_itBufferPos;

	ShellExecuteThread m_shellExecuteThread;
};

#endif // __DRUINKEDIT_H__
