//-------------------------------------------------------------------------------------------------
// HTTPSocket.cpp - Socket that communicates over psuedo-HTTP
//-------------------------------------------------------------------------------------------------
#include "HTTPSocket.h"
#include <algorithm>
#include <sstream>

// /httpconnect localhost:80

static inline std::string Trim(const std::string& src)
{
	std::string ret(src);
	while(!ret.empty() && ret[ret.length()-1] <= ' ')
		ret.erase(ret.end()-1);
	return ret;
}

static inline std::string LowerCase(const std::string& src)
{
	std::string ret(src);
	std::transform(ret.begin(), ret.end(), ret.begin(), tolower);
	return ret;
}

//-------------------------------------------------------------------------------------------------

CHTTPSocket::CHTTPSocket() :
	m_state(STATE_Idle),
	m_connectStatus(STATUS_NotConnected),
	m_socketStatus(STATUS_Ok),
	m_port(0)
{
}

CHTTPSocket::~CHTTPSocket()
{
}

void CHTTPSocket::Destroy()
{
	delete this;
}

//-------------------------------------------------------------------------------------------------

bool CHTTPSocket::StartConnect(const char* szHost, unsigned short wPort)
{
	// Disconnect if connected
	Disconnect();

	// Do initial connection
	if(!m_sock.StartConnect(szHost, wPort))
	{
		m_error = m_sock.GetError();
		return false;
	}

	// Store host and port
	m_host = szHost;
	m_port = wPort;
	m_connectStatus = STATUS_Connecting;
	m_state = STATE_ConnectingHTTP;
	return true;
}

CBaseSocket::ConnectStatus CHTTPSocket::PollConnectComplete()
{
	Tick();
	return m_connectStatus;
}

void CHTTPSocket::Disconnect()
{
	m_sock.Disconnect();
	m_buffer.clear();
	m_cookie.clear();
	m_connectStatus = STATUS_NotConnected;
	m_socketStatus = STATUS_Ok;
}

void CHTTPSocket::ClearError()
{
	m_sock.ClearError();
	m_error.clear();

	if(m_connectStatus == STATUS_ConnectError)
		m_connectStatus = STATUS_NotConnected;
	m_socketStatus = STATUS_Ok;
}

//-------------------------------------------------------------------------------------------------

CBaseSocket::SocketStatus CHTTPSocket::Poll()
{
	Tick();
	return m_socketStatus;
}

//-------------------------------------------------------------------------------------------------

const char* CHTTPSocket::GetLine()
{
	// If the data contains a NULL, we need to do things the hard way
	for(size_t i=0; i<m_buffer.size(); ++i)
	{
		if(m_buffer[i] == '\n')
		{
			m_lineBuff.resize(i+1, '*');
			memcpy(&m_lineBuff[0], &m_buffer[0], i+1);
			m_buffer.erase(m_buffer.begin(), m_buffer.begin()+i+1);
			return m_lineBuff.c_str();
		}
	}

	return "";
}

//-------------------------------------------------------------------------------------------------

bool CHTTPSocket::Send(const void* buff, size_t len)
{
	m_sendBuffer.insert(m_sendBuffer.end(), (const char*)buff, (const char*)buff + len);
	return true;
}

//-------------------------------------------------------------------------------------------------

void CHTTPSocket::Tick()
{
	switch(m_state)
	{
	case STATE_Idle: break;
	case STATE_ConnectingHTTP: TickConnecting(); break;
	case STATE_WaitResponse: TickWaitResponse(); break;
	}
	return;
}

//-------------------------------------------------------------------------------------------------

void CHTTPSocket::TickConnecting()
{
	// Poll socket status and bail if still connecting
	ConnectStatus status = m_sock.PollConnectComplete();
	if(status == STATUS_Connecting)
		return;

	// Error?
	if(status == STATUS_ConnectError)
	{
		// Initial connect error?
		if(m_connectStatus == STATUS_Connecting)
		{
			m_error = m_sock.GetError();
			m_connectStatus = STATUS_ConnectError;
			m_state = STATE_Idle;
		}
		else
		{
			// Subsequent connect error
			m_error = "Failed to re-establish HTTP connection: ";
			m_error += m_sock.GetError();
			m_socketStatus = STATUS_Error;
			m_state = STATE_Idle;
		}
		return;
	}

	// Send HTTP POST request with pending send buffer
	std::ostringstream request;
	request << "POST / HTTP/1.1\r\n";
	request << "Host: " << m_host << "\r\n";
	request << "User-Agent: DruinkClientV2 HTTP connect\r\n";
	if(!m_cookie.empty())
		request << "Cookie: " << m_cookie << "\r\n";
	request << "Connection: Close\r\n";
	request << "Content-Type: application/octet-stream\r\n";
	request << "Content-Length: " << m_sendBuffer.size() << "\r\n";
	request << "\r\n";
	std::string strRequest(request.str());
	m_sock.Send(strRequest.c_str(), strRequest.length());
	if(!m_sendBuffer.empty())
	{
		m_sock.Send(&m_sendBuffer[0], m_sendBuffer.size());
		m_sendBuffer.clear();
	}

	// Now waiting on HTTP response with server data in it
	m_state = STATE_WaitResponse;
}

//-------------------------------------------------------------------------------------------------

void CHTTPSocket::TickWaitResponse()
{
	// Poll the socket and respond to error condition
	SocketStatus status = m_sock.Poll();
	if(status == STATUS_Error)
	{
		m_error = m_sock.GetError();
		if(m_connectStatus == STATUS_Connecting)
		{
			m_error = "HTTP Error: Connecting to the server failed. Error: ";
			m_error += m_sock.GetError();
			m_connectStatus = STATUS_ConnectError;
		}
		else
		{
			m_error = "HTTP Error: Reconnecting to the server failed. Error: ";
			m_error += m_sock.GetError();
		}
		m_socketStatus = STATUS_Error;
		m_state = STATE_Idle;
		return;
	}

	// Early out if socket not yet closed
	if(status == STATUS_Ok)
	{
		// If we've got data to send, do so now
		if(!m_sendBuffer.empty())
		{
			m_sock.Send(&m_sendBuffer[0], m_sendBuffer.size());
			m_sendBuffer.clear();
		}
		return;
	}

	// Server has sent full buffer and closed the connection; parse the HTTP response header
	std::string origLine = Trim(m_sock.GetLine());
	if(origLine.empty())
	{
		// If the server closes the socket with no data sent, this means that the remote server
		// closed the socket.
		Disconnect();
		m_socketStatus = STATUS_Disconnected;
		return;
	}
	std::string line = LowerCase(origLine);
	std::string totalHeader = origLine;
	size_t pos = line.find(' ');
	if(strncmp(line.c_str(), "http/", 5) != 0 || pos == std::string::npos ||
		strncmp(line.substr(pos+1).c_str(), "200 ", 4) != 0)
	{
		m_error = "HTTP Error: Server responded with \"" + origLine +
			"\" while executing HTTP request.";
		m_socketStatus = STATUS_Error;
		m_state = STATE_Idle;
		return;
	}

	// Search for cookie and response token and scan till end of headers
	std::string magicToken;
	do {
		origLine = Trim(m_sock.GetLine());
		line = LowerCase(origLine);
		totalHeader += "\r\n" + origLine;
		if(strncmp(line.c_str(), "set-cookie: ", 12) == 0)
		{
			m_cookie = origLine.substr(12);
		}
		else if(strncmp(line.c_str(), "x-http-tunnel: ", 15) == 0)
		{
			magicToken = origLine.substr(15);
		}
	} while(!line.empty());

	// Did we get the token?
	if(magicToken.empty())
	{
		m_error = "HTTP Error: Server is not an HTTP tunneling proxy. Full HTTP response header:\r\n";
		m_error += totalHeader;
		m_socketStatus = STATUS_Error;
		m_state = STATE_Idle;
		m_sendBuffer.clear();
		return;
	}

	// Output token on first connection
	if(m_connectStatus == STATUS_Connecting)
	{
		std::string msg = "Connected to HTTP tunnelling proxy to " + magicToken + "\r\n\r\n";
		m_buffer.insert(m_buffer.end(), msg.begin(), msg.end());
	}
	m_connectStatus = STATUS_Connected;

	// Remainder of buffer is actual server data
	m_buffer.insert(m_buffer.end(), m_sock.GetBuffer(), m_sock.GetBuffer()+m_sock.GetBufferSize());

	// Reconnect socket
	if(!m_sock.StartConnect(m_host.c_str(), m_port))
	{
		m_error = "HTTP Error: Reconnecting to the server failed. Error: ";
		m_error += m_sock.GetError();
		m_socketStatus = STATUS_Error;
		m_state = STATE_Idle;
		m_sendBuffer.clear();
		return;
	}

	// Change state
	m_connectStatus = STATUS_Connected;
	m_state = STATE_ConnectingHTTP;
}

//-------------------------------------------------------------------------------------------------
