//=================================================================================================
// Memory.cpp - Custom memory manager
//=================================================================================================
#include "Memory.h"
#ifdef USE_MMGR
#include <stdio.h>
#include <new>

#ifdef USE_STACKTRACE
	#include <dbghelp.h>
	#pragma comment(lib,"dbghelp.lib")
#endif // USE_STACKTRACE

// Assert macro
#ifdef _DEBUG
#	define Assert(exp, msg, ...) \
		__pragma(warning(disable:4127)) \
		do { \
			if (!(exp)) \
				DbgAssertFailed(#exp, __FILE__, __LINE__, __FUNCTION__, msg, __VA_ARGS__); \
		} while(0) \
		__pragma(warning(default:4127))
#else
#	define Assert(exp, msg, ...) ((void)0)
#endif
static void DbgAssertFailed(const char* szExp, const char* szFile, unsigned int nLine,
	const char* szFunc, const char* szMsg, ...);

// Sentinal fill values
#ifdef BUILD_X64
	static const size_t s_nHeadSentinalFill = 0xFDFDFDFDFDFDFDFD;
	static const size_t s_nTailSentinalFill = 0xFDFDFDFDFDFDFDFD;
#else
	static const size_t s_nHeadSentinalFill = 0xFDFDFDFD;
	static const size_t s_nTailSentinalFill = 0xFDFDFDFD;
#endif

Memory* Memory::ms_pInstance = NULL;

//=================================================================================================

static void Log(FILE* pFile, const char* psz)
{
	OutputDebugStringA(psz);
	if(pFile)
		fwrite(psz, 1, strlen(psz), pFile);
}

static void Format(FILE* pFile, const char* szFormat, ...)
{
	char szBuff[1024];
	va_list arg;
	va_start(arg, szFormat);
	_vsnprintf(szBuff, sizeof(szBuff), szFormat, arg);
	va_end(arg);
	Log(pFile, szBuff);
}

//=================================================================================================

static void DbgAssertFailed(const char* szExp, const char* szFile, unsigned int nLine,
	const char* szFunc, const char* szMsg, ...)
{
	// Trim path from filename
	char szBuff[1024];
	if(!szFile) szFile = "??";
	const char* szRealFile = strrchr(szFile,'\\');
	if(!szRealFile) szRealFile = szFile;
	else szRealFile++;

	// Build vaargs
	va_list arg;
	va_start(arg, szMsg);
	_vsnprintf_s(szBuff, sizeof(szBuff), sizeof(szBuff), szMsg, arg);
	va_end(arg);

	// Compose string
	char szOutput[1536];
	sprintf_s(szOutput, sizeof(szOutput), "Assertion failed:\n"
		"%s\n"
		"\n"
		"File: %s:%d (%s)\n"
		"Expression:\n"
		"%s\n", szBuff, szRealFile, nLine, szFunc, szExp);

	// Log assert then display messagebox
	OutputDebugStringA(szOutput);

	// Only allow one message box to be active at once - if an assert is put in a window message
	// loop or other callback, we might get multiple message boxes.
	static bool bMessageBoxActive = false;
	if(!bMessageBoxActive)
	{
		strcat_s(szOutput, sizeof(szOutput), "\nDebug?\n\n");
		bMessageBoxActive = true;
		if(MessageBoxA(NULL, szOutput, "Assertion failed!", MB_YESNO|MB_ICONEXCLAMATION) == IDYES)
		{
			__debugbreak();
		}
		bMessageBoxActive = false;
	}
}

//=================================================================================================

void* operator new(size_t nSize)
{
	void* pMemory = Memory::Get().Allocate(nSize);
	Assert(pMemory, "Memory allocator failed!");
	return pMemory;
}

void* operator new[](size_t nSize)
{
	void* pMemory = Memory::Get().AllocateArray(nSize);
	Assert(pMemory, "Memory allocator failed!");
	return pMemory;
}

void operator delete(void* pAddress)
{
	Memory::Get().Deallocate(pAddress);
}

void operator delete[](void* pAddress)
{
	Memory::Get().DeallocateArray(pAddress);
}

//=================================================================================================

Memory::Memory() :
	m_pHead(NULL),
	m_pTail(NULL),
	m_nAllocations(0),
	m_nAllocationsSinceMarker(0),
	m_nAllocatedBytes(0),
	m_nNextID(1),
	m_nBreakOnAllocID(0),
	m_nHighWatermark(0),
	m_nAllocationTag(0)
{
	// Get RtlCaptureContext() function pointer dynamically to avoid compile-time
	// dependency on WinXP+
	HMODULE hKernel32 = LoadLibrary("kernel32.dll");
	m_pfnCaptureContext = (LPRtlCaptureContext)GetProcAddress(hKernel32, "RtlCaptureContext");
	FreeLibrary(hKernel32);

	InitializeCriticalSection(&m_theMutex);
#ifdef USE_STACKTRACE
	SymInitialize(GetCurrentProcess(), NULL, TRUE);
#endif
}

Memory::~Memory()
{
	Format(NULL, "MEMORY  : High watermark was %d bytes (%dKB, %dMB)\n", m_nHighWatermark,
		m_nHighWatermark/1024, m_nHighWatermark/(1024*1024));
	if(m_nAllocations != 0)
	{
		Log(NULL, "MEMORY  : Memory leaks detected!\n");
		DumpAllAllocations(true);
	}
	else
		Log(NULL, "MEMORY  : No memory leaks detected\n");

	DeleteCriticalSection(&m_theMutex);
	Assert(m_nAllocations == 0, "Memory leaks were found");
}

void Memory::Create()
{
	void* pBuff = malloc(sizeof(Memory));
	Memory* pMemory = new(pBuff) Memory();
	ms_pInstance = pMemory;
}

void Memory::Destroy()
{
	if(ms_pInstance)
	{
		ms_pInstance->~Memory();
		free(ms_pInstance);
		ms_pInstance = 0;
	}
}

//=================================================================================================

void Memory::DumpAllAllocations(bool bDumpToFile)
{
	FILE* pFile = bDumpToFile ? fopen("memleaks.log", "w") : NULL;

	EnterCriticalSection(&m_theMutex);
	Format(pFile, "%d active allocations totalling %d bytes:\n",
		m_nAllocations, m_nAllocatedBytes);
	AllocHeader* pAlloc = m_pHead;
	while(pAlloc)
	{
#ifdef USE_STACKTRACE
		Format(pFile, "+ ID %08d, tag ID 0x%08x: 0x%p %d bytes [%s:%d (%s)]\n",
			pAlloc->nAllocID, pAlloc->nTagID, ((unsigned char*)pAlloc)+sizeof(AllocHeader)+sizeof(size_t),
			pAlloc->nLength, pAlloc->szFile, pAlloc->nLine, pAlloc->szFunc);
#else
		Format(pFile, "+ ID %08d, tag ID 0x%08x: 0x%p (%d bytes)\n",
			pAlloc->nAllocID, pAlloc->nTagID, ((unsigned char*)pAlloc)+sizeof(AllocHeader)+sizeof(size_t),
			pAlloc->nLength);
#endif
		pAlloc = pAlloc->pNext;
	}
	Log(pFile, "End of allocations\n");
	LeaveCriticalSection(&m_theMutex);
}

//=================================================================================================

void* Memory::Allocate(size_t nSize)
{
	return InternalAlloc(nSize, ALLOC_New);
}

void* Memory::AllocateArray(size_t nSize)
{
	return InternalAlloc(nSize, ALLOC_NewArray);
}

void Memory::Deallocate(void* pAddress)
{
	InternalDealloc(pAddress, ALLOC_New);
}

void Memory::DeallocateArray(void* pAddress)
{
	InternalDealloc(pAddress, ALLOC_NewArray);
}

//=================================================================================================

void* Memory::InternalAlloc(size_t nSize, Alloc eType)
{
	// Break on this alloc? Increase alloc ID first in case we re-enter this func
	++m_nNextID;
	Assert(m_nNextID-1 != m_nBreakOnAllocID, "Break on this allocation requested");

	// Allocate the memory block and get pointers
	unsigned char* pRawData = (unsigned char*)malloc(nSize + sizeof(AllocHeader) + sizeof(size_t)*2);
	if(!pRawData)
	{
		Assert(false, "Memory manager out of memory");
		return NULL;
	}

	AllocHeader* pHeader = (AllocHeader*)pRawData;
	size_t* pHeadSentinal = (size_t*)(pRawData + sizeof(AllocHeader));
	size_t* pTailSentinal = (size_t*)(pRawData + sizeof(AllocHeader) + sizeof(size_t) + nSize);

	// Set the sentinals
	*pHeadSentinal = s_nHeadSentinalFill;
	*pTailSentinal = s_nTailSentinalFill;

	// Determine caller
	DetermineCaller(pHeader);

	// Enter mutex for insertion to allocation list
	EnterCriticalSection(&m_theMutex);
	{
		// Setup the allocation header
		pHeader->pNext = NULL;
		pHeader->pPrev = m_pTail;
		pHeader->eType = eType;
		pHeader->nLength = nSize;
		pHeader->nAllocID = m_nNextID;
		pHeader->nTagID = m_nAllocationTag;

		// Update stats
		++m_nAllocations;
		++m_nAllocationsSinceMarker;
		m_nAllocatedBytes += nSize;
		if(m_nAllocatedBytes > m_nHighWatermark)
			m_nHighWatermark = m_nAllocatedBytes;

		if(!m_pHead) m_pHead = pHeader;
		if(m_pTail)
			m_pTail->pNext = pHeader;
		m_pTail = pHeader;
	}
	LeaveCriticalSection(&m_theMutex);
	return pRawData + sizeof(AllocHeader) + sizeof(size_t);
}

void Memory::InternalDealloc(void* p, Alloc eType)
{
	// Do nothing for null pointers
	if(!p)
		return;

	// Grab the header and sentinal positions for this block
	unsigned char* pRawData = (unsigned char*)p;
	size_t* pHeadSentinal = (size_t*)(pRawData - sizeof(size_t));
	AllocHeader* pHeader = (AllocHeader*)(pRawData - sizeof(size_t) - sizeof(AllocHeader));
	size_t* pTailSentinal = (size_t*)(pRawData + pHeader->nLength);

	// Validate sentinals and header
	Assert(*pHeadSentinal == s_nHeadSentinalFill, "Head sentinal damaged!");
	Assert(pHeader->eType == eType, "Incorrect deallocation used");
	Assert(*pTailSentinal == s_nTailSentinalFill, "Tail sentinal damaged!");

	// Shut up about unreferenced parameters in release builds
	UNREFERENCED_PARAMETER(pHeadSentinal);
	UNREFERENCED_PARAMETER(eType);
	UNREFERENCED_PARAMETER(pTailSentinal);

	// Unlink block
	EnterCriticalSection(&m_theMutex);
	{
		if(pHeader->pPrev)
			pHeader->pPrev->pNext = pHeader->pNext;
		else
		{
			Assert(pHeader == m_pHead, "Allocator broken!");
			m_pHead = pHeader->pNext;
		}

		if(pHeader->pNext)
			pHeader->pNext->pPrev = pHeader->pPrev;
		else
		{
			Assert(pHeader == m_pTail, "Allocator broken!");
			m_pTail = pHeader->pPrev;
		}
	}
	LeaveCriticalSection(&m_theMutex);

	// Update stats
	--m_nAllocations;
	m_nAllocatedBytes -= pHeader->nLength;

	// Deallocate memory
	free(pHeader);
}

//=================================================================================================

void Memory::DetermineCaller(AllocHeader* pAllocation)
{
#ifdef USE_STACKTRACE
	// Sanity check we got RtlCaptureContext()
	if(!m_pfnCaptureContext)
	{
		strcpy(pAllocation->szFile, "??");
		pAllocation->nLine = 0;
		strcpy(pAllocation->szFunc, "??");
		return;
	}

	// Get the current thread context
	DWORD dwMachineType;
	CONTEXT theContext;
	memset(&theContext, 0, sizeof(theContext));
	m_pfnCaptureContext(&theContext);

	// Init the stack frame for this function
	STACKFRAME64 theStackFrame;
	memset(&theStackFrame, 0, sizeof(theStackFrame));
#ifdef _M_IX86
	dwMachineType = IMAGE_FILE_MACHINE_I386;
	theStackFrame.AddrPC.Offset = theContext.Eip;
	theStackFrame.AddrPC.Mode = AddrModeFlat;
	theStackFrame.AddrFrame.Offset = theContext.Ebp;
	theStackFrame.AddrFrame.Mode = AddrModeFlat;
	theStackFrame.AddrStack.Offset = theContext.Esp;
	theStackFrame.AddrStack.Mode = AddrModeFlat;
#elif _M_X64
	dwMachineType = IMAGE_FILE_MACHINE_AMD64;
	theStackFrame.AddrPC.Offset = theContext.Rip;
	theStackFrame.AddrPC.Mode = AddrModeFlat;
	theStackFrame.AddrFrame.Offset = theContext.Rsp;
	theStackFrame.AddrFrame.Mode = AddrModeFlat;
	theStackFrame.AddrStack.Offset = theContext.Rsp;
	theStackFrame.AddrStack.Mode = AddrModeFlat;
#elif _M_IA64
	dwMachineType = IMAGE_FILE_MACHINE_IA64;
	theStackFrame.AddrPC.Offset = theContext.StIIP;
	theStackFrame.AddrPC.Mode = AddrModeFlat;
	theStackFrame.AddrFrame.Offset = theContext.IntSp;
	theStackFrame.AddrFrame.Mode = AddrModeFlat;
	theStackFrame.AddrBStore.Offset = theContext.RsBSP;
	theStackFrame.AddrBStore.Mode = AddrModeFlat;
	theStackFrame.AddrStack.Offset = theContext.IntSp;
	theStackFrame.AddrStack.Mode = AddrModeFlat;
#else
#	error "Platform not supported!"
#endif

	// Initialise allocation source
	strcpy(pAllocation->szFile, "??");
	pAllocation->nLine = 0;

	// Walk up the stack to the caller
	for(int i=0; i<64; ++i)
	{
		if(!StackWalk64(dwMachineType, GetCurrentProcess(), GetCurrentThread(), &theStackFrame,
			&theContext, NULL, SymFunctionTableAccess64, SymGetModuleBase64, NULL))
		{
			// Failed to walk up the stack enough, this might not need an assert
			Assert(false, "Failed to find caller");
			return;
		}

		// Get function name
		unsigned char byBuffer[sizeof(IMAGEHLP_SYMBOL64) + AllocHeader::cnBufferSize];
		IMAGEHLP_SYMBOL64* pSymbol = (IMAGEHLP_SYMBOL64*)byBuffer;
		DWORD64 dwDisplacement;
		memset(pSymbol, 0, sizeof(IMAGEHLP_SYMBOL64) + AllocHeader::cnBufferSize);
		pSymbol->SizeOfStruct = sizeof(IMAGEHLP_SYMBOL64);
		pSymbol->MaxNameLength = AllocHeader::cnBufferSize;
		if(!SymGetSymFromAddr64(GetCurrentProcess(), theStackFrame.AddrPC.Offset, &dwDisplacement, pSymbol))
			strcpy(pAllocation->szFunc, "??");
		else
		{
			pSymbol->Name[AllocHeader::cnBufferSize-1] = '\0';
			// See if we need to go further up the stack
			if(strncmp(pSymbol->Name, "Memory::", 9) == 0)
			{
				// In Memory, keep going...
			}
			else if(strncmp(pSymbol->Name, "operator new", 12) == 0)
			{
				// In operator new or new[], keep going...
			}
			else if(strncmp(pSymbol->Name, "std::", 5) == 0)
			{
				// In STL code, keep going...
			}
			else
			{
				// Found the allocator (Or near to it)
				strcpy(pAllocation->szFunc, pSymbol->Name);
				break;
			}
		}
	}

	// Get file/line number
	{
		IMAGEHLP_LINE64 theLine;
		DWORD dwDisplacement;
		memset(&theLine, 0, sizeof(theLine));
		theLine.SizeOfStruct = sizeof(theLine);
		if(!SymGetLineFromAddr64(GetCurrentProcess(), theStackFrame.AddrPC.Offset, &dwDisplacement, &theLine))
		{
			strcpy(pAllocation->szFile, "??");
			pAllocation->nLine = 0;
			strcpy(pAllocation->szFunc, "??");
		}
		else
		{
			const char* pszFile = strrchr(theLine.FileName, '\\');
			if(!pszFile) pszFile = theLine.FileName;
			else ++pszFile;
			strncpy(pAllocation->szFile, pszFile, AllocHeader::cnBufferSize);
			pAllocation->nLine = theLine.LineNumber;
		}
	}

#else
	UNREFERENCED_PARAMETER(pAllocation);
#endif // USE_STACKTRACE
}

//=================================================================================================

// Ensure the memory manager is constructed extremely early on
#pragma warning(disable:4074) // C4074: initializers put in compiler reserved initialization area
#pragma init_seg(compiler)
struct MemoryInitialiser
{
	MemoryInitialiser() { Memory::Create(); }
	~MemoryInitialiser() { Memory::Destroy(); }
} g_theMemoryInitialiser;

//=================================================================================================

#endif // USE_MMGR
