//============================================================================
// ScriptVar.cpp - A single script variable
//
// The reason for the lack of STL objects, or new/delete is so this code can
// function properly across DLL boundaries, where the code on each size of
// the boundary uses a different STL version or operator new/delete.
// A script var keeps track of a pointer to the free() function, which matches
// the malloc() function that was used to allocate memory.
//============================================================================
#include "ScriptVar.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#pragma warning(disable:4127) // warning C4127: conditional expression is constant

//============================================================================
// Inline assert-esque function
#ifdef _DEBUG
#include <stdarg.h>
#include <windows.h>
static void DbgAssertFailed(const char* szFile, unsigned int nLine, const char* szFunc,
						const char* szMsg, ...)
{
	// Trim path from filename
	if(!szFile) szFile = "??";
	const char* szRealFile = strrchr(szFile,'\\');
	if(!szRealFile) szRealFile = szFile;
	else szRealFile++;

	// Build vaargs
	char szBuff[1024];
	va_list arg;
	va_start(arg, szMsg);
	_vsnprintf(szBuff, sizeof(szBuff), szMsg, arg);
	va_end(arg);

	// Compose string
	char szOutput[512];
	_snprintf(szOutput, sizeof(szOutput), "Debug error:\n"
		"%s\n"
		"\n"
		"File: %s:%d (%s)\n"
		"\n"
		"Debug?", szBuff, szRealFile, nLine, szFunc);

	// Display messagebox (0x34 == MB_YESNO|MB_ICONEXCLAMATION, 6==IDYES
	if(MessageBoxA(NULL, szOutput, "Debug error", 0x34) == 6)
	{
		__debugbreak();
	}
}
#define DebugError(msg, ...) \
				DbgAssertFailed( __FILE__, __LINE__, __FUNCTION__, msg, __VA_ARGS__)
#else
#	define DebugError(msg, ...) ((void)0)
#endif

//============================================================================

ScriptVar::ScriptVar() :
	m_eType(TYPE_Int),
	m_nArrayLen(0),
	m_pfnFree(free)
{
	m_primVal.nValue = 0;
}

ScriptVar::ScriptVar(Type eType) :
	m_eType(eType),
	m_nArrayLen(0),
	m_pfnFree(free)
{
	m_primVal.pValue = 0;
}

ScriptVar::ScriptVar(float f) :
	m_eType(TYPE_Float),
	m_nArrayLen(0),
	m_pfnFree(free)
{
	m_primVal.fValue = f;
}

ScriptVar::ScriptVar(int n) :
	m_eType(TYPE_Int),
	m_nArrayLen(0),
	m_pfnFree(free)
{
	m_primVal.nValue = n;
}

ScriptVar::ScriptVar(const char* sz) :
	m_eType(TYPE_String),
	m_nArrayLen(0),
	m_pfnFree(free)
{
	if(!sz) sz="";
	size_t nLen = strlen(sz)+1;
	m_primVal.szValue = (char*)malloc(nLen);
	memcpy(m_primVal.szValue, sz, nLen);
}

ScriptVar::ScriptVar(const ScriptVar* pArray, size_t nLength) :
	m_eType(TYPE_Array),
	m_nArrayLen(nLength),
	m_pfnFree(free)
{
	m_primVal.pValue = (ScriptVar*)malloc(sizeof(ScriptVar)*nLength);
	for(size_t i=0; i<nLength; ++i)
	{
		m_primVal.pValue[i].m_eType = TYPE_Int; // Prevent deallocation
		m_primVal.pValue[i] = pArray[i];
	}
}

ScriptVar::ScriptVar(const ScriptVar& rhs) :
	m_eType(rhs.m_eType),
	m_nArrayLen(rhs.m_nArrayLen),
	m_pfnFree(free)
{
	if(m_eType == TYPE_String)
	{
		const char* psz = rhs.m_primVal.szValue;
		if(!psz) psz="";
		size_t nLen = strlen(psz)+1;
		m_primVal.szValue = (char*)malloc(nLen);
		memcpy(m_primVal.szValue, psz, nLen);
		m_pfnFree = free;
	}
	else if(m_eType == TYPE_Array)
	{
		if(rhs.m_primVal.pValue)
		{
			m_primVal.pValue = (ScriptVar*)malloc(sizeof(ScriptVar)*m_nArrayLen);
			m_pfnFree = free;
			for(size_t i=0; i<m_nArrayLen; ++i)
			{
				m_primVal.pValue[i].m_eType = TYPE_Int;
				m_primVal.pValue[i] = rhs.m_primVal.pValue[i];
			}
		}
		else
		{
			m_primVal.pValue = 0;
			m_nArrayLen = 0;
		}
	}
	else
	{
		m_primVal = rhs.m_primVal;
	}
}

ScriptVar::~ScriptVar()
{
	if(m_eType == TYPE_String)
		m_pfnFree(m_primVal.szValue);
	else if(m_eType == TYPE_Array)
	{
		for(size_t i=0; i<m_nArrayLen; ++i)
			m_primVal.pValue[i].~ScriptVar();
		m_pfnFree(m_primVal.pValue);
	}
}

//============================================================================

int ScriptVar::ToInt() const
{
	switch(m_eType)
	{
	case TYPE_Float: return (int)m_primVal.fValue;
	case TYPE_Int: return m_primVal.nValue;
	case TYPE_String: return atoi(m_primVal.szValue);
	case TYPE_Array: DebugError("Cannot convert from array to int"); return 0;
	}

	DebugError("Unhandled type");
	return 0;
}

float ScriptVar::ToFloat() const
{
	switch(m_eType)
	{
	case TYPE_Float: return m_primVal.fValue;
	case TYPE_Int: return (float)m_primVal.nValue;
	case TYPE_String: return (float)atof(m_primVal.szValue);
	case TYPE_Array: DebugError("Cannot convert from array to float"); return 0.0f;
	}

	DebugError("Unhandled type");
	return 0.0f;
}

const char* ScriptVar::ToString() const
{
	static char szBuff[512];

	switch(m_eType)
	{
		case TYPE_Float:
		{
			sprintf(szBuff, "%f", m_primVal.fValue);
			return szBuff;
		}
		case TYPE_Int:
		{
			sprintf(szBuff, "%d", m_primVal.nValue);
			return szBuff;
		}

		case TYPE_String:
		{
			return m_primVal.szValue;
		}

		case TYPE_Array:
		{
			DebugError("Cannot convert from array to string");
			return "";
		}
	}

	DebugError("Unhandled type");
	return "";
}

//============================================================================

const ScriptVar& ScriptVar::GetArrayItem(size_t nIdx) const
{
	if(m_eType != TYPE_Array)
	{
		DebugError("Not an array");
		return *this;
	}
	if(nIdx >= m_nArrayLen)
	{
		DebugError("Invalid array index");
		return *this;
	}
	return m_primVal.pValue[nIdx];
}

ScriptVar& ScriptVar::GetArrayItem(size_t nIdx)
{
	if(m_eType != TYPE_Array)
	{
		DebugError("Not an array");
		return *this;
	}
	if(nIdx >= m_nArrayLen)
	{
		DebugError("Invalid array index");
		return *this;
	}
	return m_primVal.pValue[nIdx];
}

void ScriptVar::AddArrayItem(const ScriptVar& val)
{
	if(m_eType != TYPE_Array)
	{
		DebugError("Not an array");
		return;
	}

	ScriptVar* pOld = m_primVal.pValue;
	++m_nArrayLen;
	m_primVal.pValue = (ScriptVar*)malloc(sizeof(ScriptVar)*m_nArrayLen);
	m_pfnFree = free;
	if(pOld)
	{
		for(size_t i=0; i<m_nArrayLen-1; ++i)
		{
			m_primVal.pValue[i].m_eType = TYPE_Int;
			m_primVal.pValue[i] = pOld[i];
		}
	}
	m_primVal.pValue[m_nArrayLen-1].m_eType = TYPE_Int;
	m_primVal.pValue[m_nArrayLen-1] = val;
	for(size_t i=0; i<m_nArrayLen-1; ++i)
		pOld[i].~ScriptVar();
	m_pfnFree(pOld);
}

void ScriptVar::RemoveArrayItem(size_t nIdx)
{
	if(m_eType != TYPE_Array)
		DebugError("Not an array");
	else if(nIdx >= m_nArrayLen)
		DebugError("Invalid array index");
	else
	{
		--m_nArrayLen;
		for(size_t i=nIdx; i<m_nArrayLen; ++i)
			m_primVal.pValue[i] = m_primVal.pValue[i+1];
		m_primVal.pValue[m_nArrayLen] = 0;
	}
}

//============================================================================

ScriptVar& ScriptVar::operator=(const ScriptVar& rhs)
{
	if(rhs == *this)
		return *this;

	// Destroy this
	if(m_eType == TYPE_String)
		m_pfnFree(m_primVal.szValue);
	else if(m_eType == TYPE_Array)
	{
		for(size_t i=0; i<m_nArrayLen; ++i)
			m_primVal.pValue[i].~ScriptVar();
		m_pfnFree(m_primVal.pValue);
	}

	// Assign
	m_eType = rhs.m_eType;
	m_nArrayLen = rhs.m_nArrayLen;
	if(m_eType == TYPE_String)
	{
		const char* psz = rhs.m_primVal.szValue;
		if(!psz) psz="";
		size_t nLen = strlen(psz)+1;
		m_primVal.szValue = (char*)malloc(nLen);
		m_pfnFree = free;
		memcpy(m_primVal.szValue, psz, nLen);
	}
	else if(m_eType == TYPE_Array)
	{
		if(rhs.m_primVal.pValue)
		{
			m_primVal.pValue = (ScriptVar*)malloc(sizeof(ScriptVar)*m_nArrayLen);
			m_pfnFree = free;
			for(size_t i=0; i<m_nArrayLen; ++i)
			{
				m_primVal.pValue[i].m_eType = TYPE_Int;
				m_primVal.pValue[i] = rhs.m_primVal.pValue[i];
			}
		}
		else
		{
			m_primVal.pValue = 0;
			m_nArrayLen = 0;
		}
	}
	else
	{
		m_primVal = rhs.m_primVal;
	}
	return *this;
}

ScriptVar& ScriptVar::operator=(float f)
{
	// Destroy this
	if(m_eType == TYPE_String)
		m_pfnFree(m_primVal.szValue);
	else if(m_eType == TYPE_Array)
	{
		for(size_t i=0; i<m_nArrayLen; ++i)
			m_primVal.pValue[i].~ScriptVar();
		m_pfnFree(m_primVal.pValue);
	}

	// Assign
	m_eType = TYPE_Float;
	m_primVal.fValue = f;
	m_nArrayLen = 0;
	return *this;
}

ScriptVar& ScriptVar::operator=(int n)
{
	// Destroy this
	if(m_eType == TYPE_String)
		m_pfnFree(m_primVal.szValue);
	else if(m_eType == TYPE_Array)
	{
		for(size_t i=0; i<m_nArrayLen; ++i)
			m_primVal.pValue[i].~ScriptVar();
		m_pfnFree(m_primVal.pValue);
	}

	// Assign
	m_eType = TYPE_Int;
	m_primVal.nValue = n;
	m_nArrayLen = 0;
	return *this;
}

ScriptVar& ScriptVar::operator=(const char* sz)
{
	// Destroy this
	if(m_eType == TYPE_String)
	{
		if(sz == m_primVal.szValue)
			return *this;
		m_pfnFree(m_primVal.szValue);
	}
	else if(m_eType == TYPE_Array)
	{
		for(size_t i=0; i<m_nArrayLen; ++i)
			m_primVal.pValue[i].~ScriptVar();
		m_pfnFree(m_primVal.pValue);
	}

	// Assign
	m_eType = TYPE_String;
	m_nArrayLen = 0;
	if(!sz) sz="";
	size_t nLen = strlen(sz)+1;
	m_primVal.szValue = (char*)malloc(nLen);
	m_pfnFree = free;
	memcpy(m_primVal.szValue, sz, nLen);
	return *this;
}

ScriptVar& ScriptVar::AssignToArray(const ScriptVar* pArray, size_t nLength)
{
	// Destroy this
	if(m_eType == TYPE_String)
		m_pfnFree(m_primVal.szValue);
	else if(m_eType == TYPE_Array)
	{
		if(pArray == m_primVal.pValue)
			return *this;
		for(size_t i=0; i<m_nArrayLen; ++i)
			m_primVal.pValue[i].~ScriptVar();
		m_pfnFree(m_primVal.pValue);
	}

	// Assign
	m_eType = TYPE_Array;
	m_nArrayLen = nLength;
	if(pArray)
	{
		m_primVal.pValue = (ScriptVar*)malloc(sizeof(ScriptVar)*nLength);
		m_pfnFree = free;
		for(size_t i=0; i<nLength; ++i)
		{
			m_primVal.pValue[i].m_eType = TYPE_Int;
			m_primVal.pValue[i] = pArray[i];
		}
	}
	else
	{
		m_primVal.pValue = 0;
		m_nArrayLen = 0;
	}
	return *this;
}

bool ScriptVar::operator==(const ScriptVar& rhs) const
{
	if(m_eType != rhs.m_eType)
		return false;
	if(m_eType == TYPE_String)
	{
		return strcmp(m_primVal.szValue, rhs.m_primVal.szValue)==0;
	}
	else if(m_eType == TYPE_Array)
	{
		for(size_t i=0; i<m_nArrayLen; ++i)
		{
			if(m_primVal.pValue[i] != rhs.m_primVal.pValue[i])
				return false;
		}
		return true;
	}
	else
	{
		return m_primVal.nValue == rhs.m_primVal.nValue;
	}
}

//============================================================================
