//============================================================================
// ScriptManager.cpp - Script manager. Handles scripts
//============================================================================
#include "Plugin/Assert.h"
#include "ScriptManager.h"
#include <windows.h>
#include <algorithm>
#include <sstream>

#define REGISTER_FUNCTION(name, func) \
	{ \
		ScriptFunction func; \
		func.pfnFunction = Static##func; \
		func.pParam = this; \
		strcpy(func.szName, name); \
		for(size_t i=0; i<m_vScriptSystems.size(); ++i) \
			m_vScriptSystems[i].pContext->Register(Static##func, this, name); \
	}

typedef ScriptContextInterface* (*LPCreateScriptInterface)(int nAPIVersion);

//============================================================================

CScriptManager::CScriptManager()
{
}

CScriptManager::~CScriptManager()
{
	Cleanup();
}

//============================================================================

bool CScriptManager::Init()
{
	// Enumerate all DLLs in the current directory, and find script ones
	m_strError = "";
	std::string strWarnings;
	WIN32_FIND_DATA theFind;
	HANDLE hFind = FindFirstFile("*.dll", &theFind);
	if(hFind != INVALID_HANDLE_VALUE)
	{
		do {
			// Load the DLL
			HMODULE hModule = LoadLibrary(theFind.cFileName);
			if(!hModule)
			{
				DWORD dwError = GetLastError();
				if(!m_strError.empty()) m_strError += "\r\n";
				m_strError += "* Could not load " + std::string(theFind.cFileName) + ": ";

				// Get windows error for if the DLL load fails
				char* szMsgBuffer;
				FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL, dwError, 0, (LPTSTR)&szMsgBuffer, 0, NULL);
				std::string strError = szMsgBuffer;
				LocalFree(szMsgBuffer);

				// Remove newlines from string
				size_t nPos = strError.find('\r');
				if(nPos != std::string::npos)
					strError.erase(strError.begin()+nPos);
				nPos = strError.find('\n');
				if(nPos != std::string::npos)
					strError.erase(strError.begin()+nPos);

				// Convert "%1" to parameter name
				nPos = strError.find("%1");
				if(nPos != std::string::npos)
				{
					strError.erase(strError.begin()+nPos, strError.begin()+nPos+2);
					strError.insert(nPos, theFind.cFileName);
				}

				m_strError += strError;
				continue;
			}
			LPCreateScriptInterface pfn = (LPCreateScriptInterface)GetProcAddress(hModule,
				"CreateScriptInterface");
			if(!pfn)
			{
				if(!m_strError.empty()) m_strError += "\r\n";
				m_strError += "* CreateScriptInterface does not exist in " +
					std::string(theFind.cFileName);
				continue;
			}

			// Call the function
			ScriptSystem system;
			system.pContext = pfn(SCRIPT_API_VERSION);
			if(!system.pContext)
			{
				if(!m_strError.empty()) m_strError += "\r\n";
				m_strError += "* CreateScriptInterface failed for " +
					std::string(theFind.cFileName);
				continue;
			}
			system.strExtension = system.pContext->GetScriptType();
			system.strFilename = theFind.cFileName;

			// Check extension not already registered
			bool bOk = true;
			for(size_t i=0; i<m_vScriptSystems.size(); ++i)
			{
				if(m_vScriptSystems[i].strExtension == system.strExtension)
				{
					if(!strWarnings.empty()) strWarnings += "\r\n";
					strWarnings += "* Script extension ." + system.strExtension +
						" registered by both " + m_vScriptSystems[i].strFilename +
						" and " + system.strFilename + ". Using " + m_vScriptSystems[i].strFilename;
					bOk = false;
				}
			}

			// Add to list
			if(bOk)
			{
				system.hDll = hModule;
				m_vScriptSystems.push_back(system);
			}
			else
			{
				system.pContext->Destroy();
				FreeLibrary(hModule);
			}
		} while(FindNextFile(hFind, &theFind));
		FindClose(hFind);
	}

	// Error out if we didn't find any suitable systems
	if(m_vScriptSystems.empty())
	{
		if(m_strError.empty())
			m_strError = "Could not find any suitable scripting systems";
		else
			m_strError = "Could not find any suitable scripting systems:\r\n" + m_strError;
		return false;
	}
	else
		m_strError = "";

	// Register basic functions
	REGISTER_FUNCTION("load", HAPI_LoadScript);
	REGISTER_FUNCTION("register_command", HAPI_RegisterCommand);
	REGISTER_FUNCTION("unregister_command", HAPI_UnregisterCommand);
	REGISTER_FUNCTION("register_event", HAPI_RegisterEvent);
	REGISTER_FUNCTION("unregister_event", HAPI_UnregisterEvent);
	REGISTER_FUNCTION("register_timer", HAPI_RegisterTimer);
	REGISTER_FUNCTION("unregister_timer", HAPI_UnregisterTimer);

	// Save warnings
	if(!strWarnings.empty())
		m_strError = "Script system reports the following warnings:\r\n" + strWarnings;

	// Done
	return true;
}

void CScriptManager::Cleanup()
{
	m_strError = "";

	for(size_t i=0; i<m_vScriptSystems.size(); ++i)
	{
		m_vScriptSystems[i].pContext->Destroy();
		FreeLibrary(m_vScriptSystems[i].hDll);
	}
	m_vScriptSystems.clear();

	m_mapCommands.clear();
	for(size_t i=0; i<EVENT_Count; ++i)
		m_vCallbacks[i].clear();
	m_vTimers.clear();
}

//============================================================================

ScriptLoadResult CScriptManager::LoadScript(const std::string& strFilename)
{
	// Get file extension
	m_strError = "";
	size_t nPos = strFilename.find('.');
	if(nPos == std::string::npos)
	{
		m_strError = "Script \""+strFilename+"\" has no extension; unable to determine script system to use";
		return SLR_UnknownError;
	}
	std::string strExtension = strFilename.substr(nPos+1);
	std::transform(strExtension.begin(), strExtension.end(), strExtension.begin(), tolower);

	// Find the script system matching this extension
	ScriptContextInterface* pSystem = GetScriptSystem(strExtension);
	if(!pSystem)
		return SLR_UnknownError;

	// Load the script
	ScriptLoadResult result = pSystem->LoadScript(strFilename.c_str());
	if(result != SLR_Success)
		m_strError = pSystem->GetError();
	return result;
}

bool CScriptManager::LoadAutoexecScript()
{
	// Load the script for each system
	m_strError = "";
	for(size_t i=0; i<m_vScriptSystems.size(); ++i)
	{
		// Load the script
		std::string strName = "autoexec.";
		strName += m_vScriptSystems[i].pContext->GetScriptType();
		ScriptLoadResult result = m_vScriptSystems[i].pContext->LoadScript(strName.c_str());
		if(result != SLR_Success)
		{
			if(!m_strError.empty()) m_strError += "\r\n";
			m_strError += m_vScriptSystems[i].pContext->GetError();
		}
	}

	return m_strError.empty();
}

bool CScriptManager::LoadScriptFromString(const std::string& strScript)
{
	// Go through each script system, trying it in turn
	m_strError = "";
	for(size_t i=0; i<m_vScriptSystems.size(); ++i)
	{
		if(!m_vScriptSystems[i].pContext->LoadScript(strScript.c_str(), strScript.length()))
		{
			if(!m_strError.empty()) m_strError += "\r\n";
			m_strError += m_vScriptSystems[i].pContext->GetError();
		}
		else
		{
			m_strError = "";
			return true;
		}
	}

	return false;
}

bool CScriptManager::ReloadScripts()
{
	// Reset state
	m_mapCommands.clear();
	for(size_t i=0; i<EVENT_Count; ++i)
		m_vCallbacks[i].clear();
	m_vTimers.clear();

	m_strError = "";
	for(size_t i=0; i<m_vScriptSystems.size(); ++i)
	{
		if(!m_vScriptSystems[i].pContext->ReloadScripts())
		{
			if(!m_strError.empty()) m_strError += "\r\n";
			m_strError += m_vScriptSystems[i].pContext->GetError();
		}
	}
	return m_strError.empty();
}

//============================================================================

bool CScriptManager::Register(const ScriptFunction& func)
{
	m_strError = "";
	for(size_t i=0; i<m_vScriptSystems.size(); ++i)
	{
		if(!m_vScriptSystems[i].pContext->Register(func.pfnFunction, func.pParam, func.szName))
		{
			if(!m_strError.empty()) m_strError += "\r\n";
			m_strError += m_vScriptSystems[i].pContext->GetError();
		}
	}
	return m_strError.empty();
}

bool CScriptManager::Unregister(const std::string& strFuncName)
{
	m_strError = "";
	for(size_t i=0; i<m_vScriptSystems.size(); ++i)
	{
		if(!m_vScriptSystems[i].pContext->Unregister(strFuncName.c_str()))
		{
			if(!m_strError.empty()) m_strError += "\r\n";
			m_strError += m_vScriptSystems[i].pContext->GetError();
		}
	}
	return m_strError.empty();
}

bool CScriptManager::Call(const std::string& strScriptType, const std::string& strFunction,
						const std::vector<ScriptVar>& args, std::vector<ScriptVar>& retvals)
{
	// Find the script system matching this extension
	ScriptContextInterface* pSystem = GetScriptSystem(strScriptType);
	if(!pSystem)
		return false;

	// Call function
	size_t nRetvals = 0;
	ScriptVar* pRetvals;
	if(!pSystem->Call(strFunction.c_str(), args.size(), &args[0], &pRetvals, nRetvals))
	{
		m_strError = pSystem->GetError();
		return false;
	}

	// Convert return values to vector
	retvals.clear();
	retvals.reserve(nRetvals);
	for(size_t i=0; i<nRetvals; ++i)
		retvals.push_back(pRetvals[i]);
	return true;
}

//============================================================================

std::string CScriptManager::DebugDump() const
{
	std::string strRet;

	// Dump commands
	if(!m_mapCommands.empty())
	{
		strRet += "<-b>Registered commands:<-/b>\n";
		for(CommandListConstIter it=m_mapCommands.begin(); it!=m_mapCommands.end(); ++it)
		{
			strRet += "   <-c00ff00>/" + it->first + " calls " + it->second.strCallback + " (" +
				it->second.strScriptType + " script)\n";
		}
		strRet += "\n";
	}

	// Dump events:
	for(size_t i=0; i<EVENT_Count; ++i)
	{
		if(m_vCallbacks[i].empty())
			continue;

		strRet += "<-b>Registered ";
		switch(i)
		{
		case EVENT_OnReceive: strRet += "onReceive"; break;
		case EVENT_OnSendLine: strRet += "onSendLine"; break;
		case EVENT_OnPreparseText: strRet += "onPreParse"; break;
		case EVENT_OnReceivePML: strRet += "onReceivePML"; break;
		case EVENT_OnConnected: strRet += "onConnected"; break;
		case EVENT_OnConnectFail: strRet += "onConnectFailed"; break;
		case EVENT_OnDisconnect: strRet += "onDisconnect"; break;
		case EVENT_OnTick: strRet += "onTick"; break;
		}
		strRet += " callbacks:<-/b>\n";

		for(size_t j=0; j<m_vCallbacks[i].size(); ++j)
		{
			strRet += "   <-c00ff00>"+m_vCallbacks[i][j].strCallback + " (" +
				m_vCallbacks[i][j].strScriptType + " script)\n";
		}
	}

	// Dump timers
	if(!m_vTimers.empty())
	{
		strRet += "<-b>Registered timers:<-/b>\n";
		for(size_t i=0; i<m_vTimers.size(); ++i)
		{
			const TimerFunc& timer = m_vTimers[i];
			std::string strFrequency, strRemain;
			{
				std::ostringstream os;
				os << timer.nTimerFrequency;
				strFrequency = os.str();
			}
			{
				std::ostringstream os;
				os << (timer.dwNextTick - GetTickCount());
				strRemain = os.str();
			}
			strRet += "   <-c00ff00>" + timer.strCallback + " (" + timer.strScriptType +
				" script) called every " + strFrequency + "ms, next call in " + strRemain + "ms\n";
		}
	}
	return strRet;
}

//============================================================================
// Host API functions
bool CScriptManager::HAPI_LoadScript(ScriptContext& context)
{
	// Validate number of parameters
	if(context.GetNumArgs() != 1)
	{
		context.SetError("load expects 1 parameter");
		return false;
	}

	// Param 1: string: Script name
	ScriptVar script = context.GetArg(0);
	if(script.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("load expects parameter 1 to be a string (script name)");
		return false;
	}

	// Load
	ScriptLoadResult result = LoadScript(script.ToString());
	if(result != SLR_Success && result != SLR_AlreadyLoaded)
	{
		context.SetError(m_strError.c_str());
		return false;
	}

	return true;
}

bool CScriptManager::HAPI_RegisterCommand(ScriptContext& context)
{
	// Validate number of parameters
	if(context.GetNumArgs() != 2)
	{
		context.SetError("register_command expects 2 parameters");
		return false;
	}

	// Param 1: string: Command name
	ScriptVar command = context.GetArg(0);
	if(command.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("register_command expects parameter 1 to be a string (command name)");
		return false;
	}

	// Param 2: string: Callback function name
	ScriptVar callback = context.GetArg(1);
	if(callback.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("register_command expects parameter 2 to be a string (callback function name)");
		return false;
	}

	// Validate command name
	std::string strCommand = command.ToString();
	for(const char* p=strCommand.c_str(); *p; ++p)
	{
		if((*p == ' ') || (*p == '/'))
		{
			std::string strError = "Invalid name for command: \"" + strCommand + "\"";
			context.SetError(strError.c_str());
			return false;
		}
	}

	// Add to command list and return
	m_mapCommands[strCommand] = CallbackInfo(context.GetScriptType(), callback.ToString());
	return true;
}

bool CScriptManager::HAPI_UnregisterCommand(ScriptContext& context)
{
	// Validate number of parameters
	if(context.GetNumArgs() != 1)
	{
		context.SetError("unregister_command expects 1 parameter");
		return false;
	}

	// Param 1: string: Callback function
	ScriptVar callback = context.GetArg(0);
	if(callback.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("unregister_command expects parameter 1 to be a string (callback function name)");
		return false;
	}

	// Find and remove command
	std::string strCallback = callback.ToString();
	for(CommandListIter it=m_mapCommands.begin(); it!=m_mapCommands.end(); ++it)
	{
		if(it->second.strCallback==strCallback && it->second.strScriptType==context.GetScriptType())
		{
			m_mapCommands.erase(it);
			return true;
		}
	}

	// Could not find command
	std::string strError = "unregister_command could not find callback function " + strCallback;
	context.SetError(strError.c_str());
	return false;
}

bool CScriptManager::HAPI_RegisterEvent(ScriptContext& context)
{
	// Validate number of parameters
	if(context.GetNumArgs() != 2)
	{
		context.SetError("register_event expects 2 parameters");
		return false;
	}

	// Param 1: string: Event name
	ScriptVar eventName = context.GetArg(0);
	if(eventName.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("register_event expects parameter 1 to be a string (event name)");
		return false;
	}

	// Param 2: string: Callback function name
	ScriptVar callback = context.GetArg(1);
	if(callback.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("register_event expects parameter 2 to be a string (callback function name)");
		return false;
	}

	// Convert event name to enum
	std::string strEvent = eventName.ToString();
	std::transform(strEvent.begin(), strEvent.end(), strEvent.begin(), tolower);
	Event eEvent;
	if(strEvent == "onpreparse") eEvent = EVENT_OnPreparseText;
	else if(strEvent == "onpreparsetext") eEvent = EVENT_OnPreparseText;
	else if(strEvent == "onsendline") eEvent = EVENT_OnSendLine;
	else if(strEvent == "onconnect") eEvent = EVENT_OnConnected;
	else if(strEvent == "onconnected") eEvent = EVENT_OnConnected;
	else if(strEvent == "onconnectfailed") eEvent = EVENT_OnConnectFail;
	else if(strEvent == "ondisconnect") eEvent = EVENT_OnDisconnect;
	else if(strEvent == "ontick") eEvent = EVENT_OnTick;
	else if(strEvent == "onreceive") eEvent = EVENT_OnReceive;
	else if(strEvent == "onreceivepml") eEvent = EVENT_OnReceivePML;
	else
	{
		std::string strError = "Invalid event for register_event: " + strEvent;
		context.SetError(strError.c_str());
		return false;
	}

	// Add to event list and return
	m_vCallbacks[eEvent].push_back(CallbackInfo(context.GetScriptType(), callback.ToString()));
	return true;
}

bool CScriptManager::HAPI_UnregisterEvent(ScriptContext& context)
{
	// Validate number of parameters
	if(context.GetNumArgs() != 1)
	{
		context.SetError("unregister_event expects 1 parameter");
		return false;
	}

	// Param 1: string: Callback function
	ScriptVar callback = context.GetArg(0);
	if(callback.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("unregister_event expects parameter 1 to be a string (callback function name)");
		return false;
	}

	// Find and remove event
	std::string strCallback = callback.ToString();
	for(size_t i=0; i<EVENT_Count; ++i)
	{
		for(size_t j=0; j<m_vCallbacks[i].size(); ++j)
		{
			if(m_vCallbacks[i][j].strCallback==strCallback &&
				m_vCallbacks[i][j].strScriptType==context.GetScriptType())
			{
				m_vCallbacks[i].erase(m_vCallbacks[i].begin()+j);
				return true;
			}
		}
	}

	// Could not find event
	std::string strError = "unregister_event could not find callback function " + strCallback;
	context.SetError(strError.c_str());
	return true;
}

bool CScriptManager::HAPI_RegisterTimer(ScriptContext& context)
{
	// Validate number of parameters
	if(context.GetNumArgs() != 2)
	{
		context.SetError("register_timer expects 2 parameters");
		return false;
	}

	// Param 1: string: Timer frequency in ms
	ScriptVar frequency = context.GetArg(0);
	if(frequency.GetType() != ScriptVar::TYPE_Float && frequency.GetType() != ScriptVar::TYPE_Int)
	{
		context.SetError("register_timer expects parameter 1 to be an integer (timer frequency)");
		return false;
	}
	int nTime = frequency.ToInt();
	if(nTime == 0)
	{
		context.SetError("Timer frequency for register_timer cannot be < 1ms");
		return false;
	}

	// Param 2: string: Callback function name
	ScriptVar callback = context.GetArg(1);
	if(callback.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("register_timer expects parameter 2 to be a string (callback function name)");
		return false;
	}

	// Add to timer list and return
	TimerFunc theTimer;
	theTimer.strScriptType = context.GetScriptType();
	theTimer.strCallback = callback.ToString();
	theTimer.nTimerFrequency = nTime;
	theTimer.dwNextTick = GetTickCount() + nTime;
	m_vTimers.push_back(theTimer);
	return true;
}

bool CScriptManager::HAPI_UnregisterTimer(ScriptContext& context)
{
	// Validate number of parameters
	if(context.GetNumArgs() != 1)
	{
		context.SetError("unregister_timer expects 1 parameter");
		return false;
	}

	// Param 1: string: Event name
	ScriptVar callback = context.GetArg(0);
	if(callback.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("unregister_timer expects parameter 1 to be a string (callback function name)");
		return false;
	}

	// Find and remove timer
	std::string strCallback = callback.ToString();
	for(size_t i=0; i<m_vTimers.size(); ++i)
	{
		if(m_vTimers[i].strCallback==strCallback && m_vTimers[i].strScriptType==context.GetScriptType())
		{
			m_vTimers.erase(m_vTimers.begin()+i);
			return true;
		}
	}

	// Could not find timer
	std::string strError = "unregister_timer could not find timer callback " + strCallback;
	context.SetError(strError.c_str());
	return false;
}

//============================================================================
// Event handlers
std::string CScriptManager::OnPreparseText(const std::string& str)
{
	// Process
	m_strError = "";
	std::string strText = str;
	for(size_t i=0; i<m_vCallbacks[EVENT_OnPreparseText].size(); ++i)
	{
		// Find the script system matching this extension
		ScriptContextInterface* pSystem = GetScriptSystem(m_vCallbacks[EVENT_OnPreparseText][i].strScriptType);
		if(!pSystem)
		{
			m_vCallbacks[EVENT_OnPreparseText].erase(m_vCallbacks[EVENT_OnPreparseText].begin()+i);
			--i;
			continue;
		}

		// Call callback
		ScriptVar arg(strText.c_str());
		size_t nRetvals = 0;
		ScriptVar* pRetval;
		if(!pSystem->Call(m_vCallbacks[EVENT_OnPreparseText][i].strCallback.c_str(),
			1, &arg, &pRetval, nRetvals))
		{
			m_strError = "Failed to call callback function \""+
				m_vCallbacks[EVENT_OnPreparseText][i].strCallback+"\" for onPreParse: "+pSystem->GetError();
			m_vCallbacks[EVENT_OnPreparseText].erase(m_vCallbacks[EVENT_OnPreparseText].begin()+i);
			--i;
			continue;
		}
		if(nRetvals != 1)
		{
			m_strError = "Callback function \""+
				m_vCallbacks[EVENT_OnPreparseText][i].strCallback+"\" for onPreParse should return 1 return value";
			m_vCallbacks[EVENT_OnPreparseText].erase(m_vCallbacks[EVENT_OnPreparseText].begin()+i);
			--i;
		}
		else if(pRetval->GetType() != ScriptVar::TYPE_String)
		{
			m_strError = "Callback function \""+
				m_vCallbacks[EVENT_OnPreparseText][i].strCallback+"\" for onPreParse should return a string";
			m_vCallbacks[EVENT_OnPreparseText].erase(m_vCallbacks[EVENT_OnPreparseText].begin()+i);
			--i;
		}
		else
		{
			strText = pRetval->ToString();
		}
	}

	// Return
	return strText;
}

std::string CScriptManager::OnSendLine(const std::string& str)
{
	// Process
	m_strError = "";
	std::string strText = str;
	for(size_t i=0; i<m_vCallbacks[EVENT_OnSendLine].size(); ++i)
	{
		// Find the script system matching this extension
		ScriptContextInterface* pSystem = GetScriptSystem(m_vCallbacks[EVENT_OnSendLine][i].strScriptType);
		if(!pSystem)
		{
			m_vCallbacks[EVENT_OnSendLine].erase(m_vCallbacks[EVENT_OnSendLine].begin()+i);
			--i;
			continue;
		}

		// Call callback
		ScriptVar arg(strText.c_str());
		size_t nRetvals = 0;
		ScriptVar* pRetval;
		if(!pSystem->Call(m_vCallbacks[EVENT_OnSendLine][i].strCallback.c_str(),
			1, &arg, &pRetval, nRetvals))
		{
			m_strError = "Failed to call callback function \""+
				m_vCallbacks[EVENT_OnSendLine][i].strCallback+"\" for onSendLine: "+pSystem->GetError();
			m_vCallbacks[EVENT_OnSendLine].erase(m_vCallbacks[EVENT_OnSendLine].begin()+i);
			--i;
			continue;
		}
		if(nRetvals != 1)
		{
			m_strError = "Callback function \""+
				m_vCallbacks[EVENT_OnSendLine][i].strCallback+"\" for onSendLine should return 1 return value";
			m_vCallbacks[EVENT_OnSendLine].erase(m_vCallbacks[EVENT_OnSendLine].begin()+i);
			--i;
		}
		else if(pRetval->GetType() != ScriptVar::TYPE_String)
		{
			m_strError = "Callback function \""+
				m_vCallbacks[EVENT_OnSendLine][i].strCallback+"\" for onSendLine should return a string";
			m_vCallbacks[EVENT_OnSendLine].erase(m_vCallbacks[EVENT_OnSendLine].begin()+i);
			--i;
		}
		else
		{
			strText = pRetval->ToString();
		}
	}

	// Return
	return strText;
}

void CScriptManager::OnConnected()
{
	// Process
	m_strError = "";
	for(size_t i=0; i<m_vCallbacks[EVENT_OnConnected].size(); ++i)
	{
		// Find the script system matching this extension
		ScriptContextInterface* pSystem = GetScriptSystem(m_vCallbacks[EVENT_OnConnected][i].strScriptType);
		if(!pSystem)
		{
			m_vCallbacks[EVENT_OnConnected].erase(m_vCallbacks[EVENT_OnConnected].begin()+i);
			--i;
			continue;
		}

		// Call callback
		ScriptVar* pRetval;
		size_t nRetvals = 0;
		if(!pSystem->Call(m_vCallbacks[EVENT_OnConnected][i].strCallback.c_str(), 0, NULL, &pRetval, nRetvals))
		{
			m_strError = "Failed to call callback function \""+
				m_vCallbacks[EVENT_OnConnected][i].strCallback+"\" for onConnected: "+pSystem->GetError();
			m_vCallbacks[EVENT_OnConnected].erase(m_vCallbacks[EVENT_OnConnected].begin()+i);
			--i;
			continue;
		}
	}
}

void CScriptManager::OnConnectFailed()
{
	// Process
	m_strError = "";
	for(size_t i=0; i<m_vCallbacks[EVENT_OnConnectFail].size(); ++i)
	{
		// Find the script system matching this extension
		ScriptContextInterface* pSystem = GetScriptSystem(m_vCallbacks[EVENT_OnConnectFail][i].strScriptType);
		if(!pSystem)
		{
			m_vCallbacks[EVENT_OnConnectFail].erase(m_vCallbacks[EVENT_OnConnectFail].begin()+i);
			--i;
			continue;
		}

		// Call callback
		ScriptVar* pRetval;
		size_t nRetvals = 0;
		if(!pSystem->Call(m_vCallbacks[EVENT_OnConnectFail][i].strCallback.c_str(), 0, NULL, &pRetval, nRetvals))
		{
			m_strError = "Failed to call callback function \""+
				m_vCallbacks[EVENT_OnConnectFail][i].strCallback+"\" for onConnectFailed: "+pSystem->GetError();
			m_vCallbacks[EVENT_OnConnectFail].erase(m_vCallbacks[EVENT_OnConnectFail].begin()+i);
			--i;
			continue;
		}
	}
}

void CScriptManager::OnDisconnected(bool bUserRequested)
{
	// Process
	m_strError = "";
	for(size_t i=0; i<m_vCallbacks[EVENT_OnDisconnect].size(); ++i)
	{
		// Find the script system matching this extension
		ScriptContextInterface* pSystem = GetScriptSystem(m_vCallbacks[EVENT_OnDisconnect][i].strScriptType);
		if(!pSystem)
		{
			m_vCallbacks[EVENT_OnDisconnect].erase(m_vCallbacks[EVENT_OnDisconnect].begin()+i);
			--i;
			continue;
		}

		// Call callback
		ScriptVar arg(bUserRequested ? 1 : 0);
		ScriptVar* pRetval;
		size_t nRetvals = 0;
		if(!pSystem->Call(m_vCallbacks[EVENT_OnDisconnect][i].strCallback.c_str(), 1, &arg, &pRetval, nRetvals))
		{
			m_strError = "Failed to call callback function \""+
				m_vCallbacks[EVENT_OnDisconnect][i].strCallback+"\" for onDisconnect: "+pSystem->GetError();
			m_vCallbacks[EVENT_OnDisconnect].erase(m_vCallbacks[EVENT_OnDisconnect].begin()+i);
			--i;
			continue;
		}
	}
}

void CScriptManager::OnTick()
{
	// Call onTick handlers
	m_strError = "";
	for(size_t i=0; i<m_vCallbacks[EVENT_OnTick].size(); ++i)
	{
		// Find the script system matching this extension
		ScriptContextInterface* pSystem = GetScriptSystem(m_vCallbacks[EVENT_OnTick][i].strScriptType);
		if(!pSystem)
		{
			m_vCallbacks[EVENT_OnTick].erase(m_vCallbacks[EVENT_OnTick].begin()+i);
			--i;
			continue;
		}

		// Call callback
		ScriptVar* pRetval;
		size_t nRetvals = 0;
		if(!pSystem->Call(m_vCallbacks[EVENT_OnTick][i].strCallback.c_str(), 0, NULL, &pRetval, nRetvals))
		{
			m_strError = "Failed to call callback function \""+
				m_vCallbacks[EVENT_OnTick][i].strCallback+"\" for onTick: "+pSystem->GetError();
			m_vCallbacks[EVENT_OnTick].erase(m_vCallbacks[EVENT_OnTick].begin()+i);
			--i;
			continue;
		}
	}

	// Process timers
	DWORD dwTime = GetTickCount();
	for(size_t i=0; i<m_vTimers.size(); ++i)
	{
		if(m_vTimers[i].dwNextTick <= dwTime)
		{
			// Find the script system matching this extension
			ScriptContextInterface* pSystem = GetScriptSystem(m_vTimers[i].strScriptType);
			if(!pSystem)
			{
				m_vTimers.erase(m_vTimers.begin()+i);
				--i;
				continue;
			}

			// Store details of current timer incase the callback removes it
			TimerFunc timer = m_vTimers[i];

			// Call callback
			ScriptVar* pRetval;
			size_t nRetvals = 0;
			if(!pSystem->Call(timer.strCallback.c_str(), 0, NULL, &pRetval, nRetvals))
			{
				m_strError = "Failed to call callback function \""+
					timer.strCallback+"\" for time event: "+pSystem->GetError();
				m_vTimers.erase(m_vTimers.begin()+i);
				--i;
				continue;
			}
			else
			{
				if(i < m_vTimers.size() && m_vTimers[i]==timer)
					m_vTimers[i].dwNextTick = dwTime + m_vTimers[i].nTimerFrequency;
			}
		}
	}
}

std::string CScriptManager::OnReceive(const std::string& str)
{
	// Process
	m_strError = "";
	std::string strText = str;
	for(size_t i=0; i<m_vCallbacks[EVENT_OnReceive].size(); ++i)
	{
		// Find the script system matching this extension
		ScriptContextInterface* pSystem = GetScriptSystem(m_vCallbacks[EVENT_OnReceive][i].strScriptType);
		if(!pSystem)
		{
			m_vCallbacks[EVENT_OnReceive].erase(m_vCallbacks[EVENT_OnReceive].begin()+i);
			--i;
			continue;
		}

		// Call callback
		ScriptVar arg(strText.c_str());
		size_t nRetvals = 0;
		ScriptVar* pRetval;
		if(!pSystem->Call(m_vCallbacks[EVENT_OnReceive][i].strCallback.c_str(),
			1, &arg, &pRetval, nRetvals))
		{
			m_strError = "Failed to call callback function \""+
				m_vCallbacks[EVENT_OnReceive][i].strCallback+"\" for onReceive: "+pSystem->GetError();
			m_vCallbacks[EVENT_OnReceive].erase(m_vCallbacks[EVENT_OnReceive].begin()+i);
			--i;
			continue;
		}
		if(nRetvals != 1)
		{
			m_strError = "Callback function \""+
				m_vCallbacks[EVENT_OnReceive][i].strCallback+"\" for onReceive should return 1 return value";
			m_vCallbacks[EVENT_OnReceive].erase(m_vCallbacks[EVENT_OnReceive].begin()+i);
			--i;
		}
		else if(pRetval->GetType() != ScriptVar::TYPE_String)
		{
			m_strError = "Callback function \""+
				m_vCallbacks[EVENT_OnReceive][i].strCallback+"\" for onReceive should return a string";
			m_vCallbacks[EVENT_OnReceive].erase(m_vCallbacks[EVENT_OnReceive].begin()+i);
			--i;
		}
		else
		{
			strText = pRetval->ToString();
		}
	}

	// Return
	return strText;
}

bool CScriptManager::OnCommand(const std::string& strCommand, const std::string& strParams)
{
	// Try to find this command
	m_strError = "";
	CommandListIter it = m_mapCommands.find(strCommand);
	if(it == m_mapCommands.end())
		return false;

	// Find the script system matching this extension
	ScriptContextInterface* pSystem = GetScriptSystem(it->second.strScriptType);
	if(!pSystem)
		return true;

	// Call callback
	size_t nRetvals = 0;
	ScriptVar* pRetvals;
	ScriptVar arg(strParams.c_str());
	if(!pSystem->Call(it->second.strCallback.c_str(), 1, &arg, &pRetvals, nRetvals))
	{
		m_strError = "Failed to call callback function \""+it->second.strCallback+"\" for command: "+
			strCommand+": "+pSystem->GetError();
		m_mapCommands.erase(it);
		return true;
	}

	m_strError = "";
	return true;
}

void CScriptManager::OnReceivePML(const std::string& str)
{
	// Process
	m_strError = "";
	for(size_t i=0; i<m_vCallbacks[EVENT_OnReceivePML].size(); ++i)
	{
		// Find the script system matching this extension
		ScriptContextInterface* pSystem = GetScriptSystem(m_vCallbacks[EVENT_OnReceivePML][i].strScriptType);
		if(!pSystem)
		{
			m_vCallbacks[EVENT_OnReceivePML].erase(m_vCallbacks[EVENT_OnReceivePML].begin()+i);
			--i;
			continue;
		}

		// Call callback
		ScriptVar arg(str.c_str());
		size_t nRetvals = 0;
		ScriptVar* pRetval;
		if(!pSystem->Call(m_vCallbacks[EVENT_OnReceivePML][i].strCallback.c_str(),
			1, &arg, &pRetval, nRetvals))
		{
			m_strError = "Failed to call callback function \""+
				m_vCallbacks[EVENT_OnReceivePML][i].strCallback+"\" for onReceivePML: "+pSystem->GetError();
			m_vCallbacks[EVENT_OnReceivePML].erase(m_vCallbacks[EVENT_OnReceivePML].begin()+i);
			--i;
			continue;
		}
	}
}

//============================================================================

ScriptContextInterface* CScriptManager::GetScriptSystem(const std::string& strSystemName)
{
	for(size_t i=0; i<m_vScriptSystems.size(); ++i)
	{
		if(m_vScriptSystems[i].strExtension == strSystemName)
		{
			return m_vScriptSystems[i].pContext;
		}
	}

	m_strError = "No registered script system handles ."+strSystemName+" files.";
	return NULL;
}

//============================================================================
