//============================================================================
// ScriptManager.h - Script manager. Handles scripts
//============================================================================

#ifndef __SCRIPTMANAGER_H__
#define __SCRIPTMANAGER_H__

#include "ScriptContextInterface/ScriptContextInterface.h"
#include "Plugin/Plugin.h"
#include <windows.h>
#include <string>
#include <vector>
#include <map>

#define DECLARE_FUNCTION(name) \
	bool name(ScriptContext& context); \
	static bool Static##name(ScriptContext& context, void* pMgr) \
		{ return ((CScriptManager*)pMgr)->name(context); }

//============================================================================
// Manager class
class CScriptManager
{
public:
	CScriptManager();
	~CScriptManager();

	bool Init();
	void Cleanup();

	// Reasons for invoking script functions
	enum Event
	{
		EVENT_OnReceive,			// Data arrived on the socket
		EVENT_OnSendLine,			// When a line of text is sent
		EVENT_OnPreparseText,		// As soon as the user hits enter in the input box
		EVENT_OnReceivePML,			// PML tag parsed out from data
		EVENT_OnConnected,			// When the socket is connected successfully
		EVENT_OnConnectFail,		// When the socket fails to connect
		EVENT_OnDisconnect,			// When the socket becomes disconnected
		EVENT_OnTick,				// Every client tick

		EVENT_Count					// Reserved
	};

	// Load a script file / string
	ScriptLoadResult LoadScript(const std::string& strFilename);
	bool LoadAutoexecScript();
	bool LoadScriptFromString(const std::string& strScript);
	bool ReloadScripts();

	// Error information
	void SetError(const std::string& str) { m_strError=str; }
	const std::string& GetError() const { return m_strError; }
	void ResetError() { m_strError.clear(); }

	// [Un]register a C++ function to be called from a script
	bool Register(const ScriptFunction& func);
	bool Unregister(const std::string& strFuncName);

	// Call a function contained in this script
	bool Call(const std::string& strScriptType, const std::string& strFunction,
		const std::vector<ScriptVar>& args, std::vector<ScriptVar>& retvals);

	// Debug dump all info we can
	std::string DebugDump() const;

	// Event handlers
	std::string OnPreparseText(const std::string& str);
	std::string OnSendLine(const std::string& str);
	void OnConnected();
	void OnConnectFailed();
	void OnDisconnected(bool bUserRequested);
	void OnTick();
	std::string OnReceive(const std::string& str);
	bool OnCommand(const std::string& strCommand, const std::string& strParams);
	void OnReceivePML(const std::string& str);

private:
	// Basic host API functions
	DECLARE_FUNCTION(HAPI_LoadScript);
	DECLARE_FUNCTION(HAPI_RegisterCommand);
	DECLARE_FUNCTION(HAPI_UnregisterCommand);
	DECLARE_FUNCTION(HAPI_RegisterEvent);
	DECLARE_FUNCTION(HAPI_UnregisterEvent);
	DECLARE_FUNCTION(HAPI_RegisterTimer);
	DECLARE_FUNCTION(HAPI_UnregisterTimer);

	struct TimerFunc
	{
		bool operator==(const TimerFunc& rhs) const
		{
			return strScriptType==rhs.strScriptType &&
				strCallback == rhs.strCallback &&
				nTimerFrequency == rhs.nTimerFrequency &&
				dwNextTick == rhs.dwNextTick;
		}
		std::string strScriptType;
		std::string strCallback;
		int nTimerFrequency;
		unsigned long dwNextTick;
	};

	struct ScriptSystem
	{
		ScriptContextInterface* pContext;
		std::string strExtension;
		std::string strFilename;
		HMODULE hDll;
	};

	struct CallbackInfo
	{
		CallbackInfo() {}
		CallbackInfo(const std::string& scriptType, const std::string& callback) :
			strScriptType(scriptType), strCallback(callback) {}

		std::string strScriptType;
		std::string strCallback;
	};

	typedef std::map<std::string, CallbackInfo> CommandList;
	typedef CommandList::iterator CommandListIter;
	typedef CommandList::const_iterator CommandListConstIter;

	ScriptContextInterface* GetScriptSystem(const std::string& strSystemName);

	std::string					m_strError;
	CommandList					m_mapCommands;
	std::vector<CallbackInfo>	m_vCallbacks[EVENT_Count];
	std::vector<TimerFunc>		m_vTimers;

	std::vector<ScriptSystem>	m_vScriptSystems;
};

#endif // __SCRIPTMANAGER_H__
