//=================================================================================================
// HTTPGet.cpp - Standard library function: http_get()
//=================================================================================================
#include "StandardLib.h"
#include "Plugin/ScriptContext.h"
#include "Socket/Socket.h"
#include <stdio.h>
#include <sstream>
#include <algorithm>

extern StandardLibState* g_pState;

struct HTTPGetThreadParam
{
	HTTPGet* pGetRequest;
	HANDLE hTerminate;
	HANDLE hStarted;
};

static const DWORD s_nRecvTimeout = 15000;				// 15 seconds
static const size_t s_nMaxDownloadSize = 1024*1024;		// 1MB
static const DWORD s_nTransferSpeedSampleRate = 500;	// 0.5 seconds

//=================================================================================================

static void ParseHeader(HTTPGet* pGet)
{
	// Get header in lower case
	std::string strHeader(pGet->strHeader.length(), ' ');
	std::transform(pGet->strHeader.begin(), pGet->strHeader.end(), strHeader.begin(), tolower);

	// Find HTTP response code in header
	if(strncmp(strHeader.c_str(), "http/1.", 7) == 0)
	{
		size_t nPos = strHeader.find(' ', 9);
		if(nPos != std::string::npos)
		{
			std::istringstream is(strHeader.substr(9, nPos-9));
			int nResponse;
			is >> nResponse;
			pGet->nResponseCode = nResponse;
		}
	}

	// Find content length
	size_t nPos = strHeader.find("content-length:");
	if(nPos != std::string::npos)
	{
		nPos += 15;
		while(strHeader.c_str()[nPos] == ' ')
			++nPos;
		size_t nEnd = strHeader.find("\r\n", nPos);
		std::string strLen;
		if(nEnd == std::string::npos)
			strLen = strHeader.substr(nPos);
		else
			strLen = strHeader.substr(nPos, nEnd-nPos);
		std::istringstream is(strLen);
		int nLen;
		is >> nLen;
		pGet->nContentLength = nLen;
	}
}

//=================================================================================================

static DWORD WINAPI HTTPGetThread(void* pParam)
{
	// Copy local vars and signal that we're started
	HTTPGet* pGet = ((HTTPGetThreadParam*)pParam)->pGetRequest;
	HANDLE hTerminate = ((HTTPGetThreadParam*)pParam)->hTerminate;
	SetEvent(((HTTPGetThreadParam*)pParam)->hStarted);

	// First, parse address
	std::string strAddress = pGet->strAddress;
	if(_strnicmp(strAddress.c_str(), "http://", 7) == 0)
		strAddress.erase(strAddress.begin(), strAddress.begin()+7);

	// Extract server and page
	std::string strServer, strPage;
	size_t nPos = strAddress.find('/');
	if(nPos == std::string::npos)
	{
		strServer = strAddress;
		strPage = "/";
	}
	else
	{
		strServer = strAddress.substr(0, nPos);
		strPage = strAddress.substr(nPos);
	}

	// Does server include port?
	unsigned short nPort;
	nPos = strServer.find(':');
	if(nPos == std::string::npos)
		nPort = 80;
	else
	{
		std::istringstream is(strServer.substr(nPos+1));
		if(!(is >> nPort))
			nPort = 80;
		strServer.erase(strServer.begin()+nPos, strServer.end());
	}

	// Connect to server on correct port
	CSocket sock;
	if(!sock.StartConnect(strServer.c_str(), nPort))
	{
		std::stringstream str;
		str << "Failed to connect to " << strServer << ":" << nPort <<
			" for httpGet(). " << sock.GetError();
		pGet->strError = str.str();
		return 1;
	}

	// Wait for connect to complete
	CBaseSocket::ConnectStatus connectStatus;
	do {
		if(WaitForSingleObject(hTerminate, 50) == WAIT_OBJECT_0)
			return 0;
		connectStatus = sock.PollConnectComplete();
	} while(connectStatus == CBaseSocket::STATUS_Connecting);

	// Check for error
	if(connectStatus == CBaseSocket::STATUS_ConnectError)
	{
		std::stringstream str;
		str << "Failed to connect to " << strServer << ":" << nPort <<
			" for httpGet(). " << sock.GetError();
		pGet->strError = str.str();
		return 1;
	}

	// Send GET request
	std::string strRequest = "GET " + strPage + " HTTP/1.1\r\n"
		"Host: " + strServer + "\r\n"
		"User-Agent: DruinkClientv2 HTTP GET\r\n"
		"Connection: Close\r\n"
		"\r\n";
	if(!sock.Send(strRequest.c_str(), strRequest.length()))
		sock.Disconnect();

	// Read data
	CBaseSocket::SocketStatus status;
	DWORD dwTimeLastRecv = GetTickCount();
	sock.ClearError();
	bool bReadingHeader = true;
	size_t nBytesRead = 0;

	DWORD dwTimeLastSpeedSample = GetTickCount();
	int nBytesReadThisSample = 0;

	std::vector<char> sockData;
	do {
		// Time to quit?
		if(WaitForSingleObject(hTerminate, 50) == WAIT_OBJECT_0)
			return 0;

		// Read data and update last recv time
		status = sock.Poll();
		if(sock.GetBufferSize() > 0 || status != CBaseSocket::STATUS_Ok)
		{
			nBytesRead += sock.GetBufferSize();
			sockData.insert(sockData.end(), sock.GetBuffer(), sock.GetBuffer()+sock.GetBufferSize());
			sock.ClearBuffer();

			// If we're reading the header, has it finished (contains "\r\n\r\n")?
			if(bReadingHeader)
			{
				size_t nPos = sockData.size()+1;
				if(sockData.size() > 4)
				{
					for(size_t i=0; i<sockData.size()-3; ++i)
					{
						if(sockData[i+0]=='\r' && sockData[i+1]=='\n' &&
							sockData[i+2]=='\r' && sockData[i+3]=='\n')
						{
							nPos = i;
							break;
						}
					}
				}

				// If we didn't find the end of the header, and we're disconnected, treat
				// the whole buffer as the header
				if(nPos == sockData.size()+1 && status != CBaseSocket::STATUS_Ok)
					nPos = sockData.size();

				// Copy the buffer, remove all NULL bytes (since that's bound to go
				// horribly wrong), and convert to a string
				if(nPos != sockData.size()+1)
				{
					std::vector<char> vBuffer(sockData.begin(), sockData.begin()+nPos);
					for(size_t i=0; i<vBuffer.size();)
					{
						if(vBuffer[i] == 0)
							vBuffer.erase(vBuffer.begin()+i);
						else
							++i;
					}
					vBuffer.push_back(0);

					pGet->strHeader = &vBuffer[0];
					if(status != CBaseSocket::STATUS_Ok)
						sockData.clear();
					else
						sockData.erase(sockData.begin(), sockData.begin()+nPos+4);
					bReadingHeader = false;

					// Parse anything relevant from the header
					ParseHeader(pGet);
					pGet->bHeaderComplete = true;
				}
			}

			// If we're reading content, write it to the file or append it to the content string
			if(!bReadingHeader && !sockData.empty())
			{
				// Update download speed
				nBytesReadThisSample += (int)sockData.size();
				DWORD dwCurrTime = GetTickCount();
				if(dwCurrTime-dwTimeLastSpeedSample > s_nTransferSpeedSampleRate)
				{
					float fElapsed = (float)(dwCurrTime-dwTimeLastSpeedSample) / 1000.0f;
					pGet->nTransferSpeed = (int)(nBytesReadThisSample / fElapsed);
					nBytesReadThisSample = 0;
					dwTimeLastSpeedSample = dwCurrTime;
				}

				// Update total transferred
				pGet->nBytesTransferred += sockData.size();

				// Store content
				if(pGet->pContentFile)
				{
					fwrite(&sockData[0], 1, sockData.size(), pGet->pContentFile);
				}
				else
				{
					for(size_t i=0; i<sockData.size();)
					{
						if(sockData[i] == 0)
							sockData.erase(sockData.begin()+i);
						else
							++i;
					}
					sockData.push_back(0);
					pGet->strContent += &sockData[0];

					// Buffer getting too big?
					if(pGet->strContent.length() >= s_nMaxDownloadSize)
					{
						sock.Disconnect();
						break;
					}
				}
				sockData.clear();
			}
			dwTimeLastRecv = GetTickCount();
			if(status != CBaseSocket::STATUS_Ok)
				break;
		}
	} while(GetTickCount()-dwTimeLastRecv < s_nRecvTimeout);

	// Close file if it's open
	if(pGet->pContentFile)
		fclose(pGet->pContentFile);

	// Any error?
	if(nBytesRead == 0)
	{
		std::stringstream str;
		str << "Failed to read data from " << strServer << ":" << nPort <<
			" for httpGet(). ";
		if(sock.GetError()[0] != '\0')
			str << "Disconnected from server with no data read";
		else
			str << sock.GetError();
		pGet->strError = str.str();
		return 1;
	}

	// Return
	pGet->strError = "";
	return 0;
}

//=================================================================================================

bool Cmd_HTTPGet(ScriptContext& context, void*)
{
	// Get args
	if(context.GetNumArgs() < 2 || context.GetNumArgs() > 3)
	{
		context.SetError("httpGet() expects two or three parameters");
		return false;
	}
	ScriptVar address = context.GetArg(0);
	if(address.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("httpGet() expects a string as the first parameter");
		return false;
	}
	ScriptVar callback = context.GetArg(1);
	if(callback.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("httpGet() expects a string as the second parameter");
		return false;
	}
	std::string strAddress = address.ToString();
	std::string strCallback = callback.ToString();
	std::string strFilename;

	// Get optional filename
	if(context.GetNumArgs() >= 3)
	{
		ScriptVar filename = context.GetArg(2);
		if(filename.GetType() != ScriptVar::TYPE_String)
		{
			context.SetError("httpGet() expects a string as the optional third parameter");
			return false;
		}
	}

	// See if there's already a request for this
	for(size_t i=0; i<g_pState->vGets.size(); ++i)
	{
		if(g_pState->vGets[i]->strAddress == strAddress &&
			g_pState->vGets[i]->strCallback == strCallback)
		{
			// Already exists, bail
			return true;
		}
	}

	// Create HTTP GET request
	std::string strError;
	HTTPGet* pGet = StartHTTPGet(strAddress, strFilename, g_pState->hThreadTerminate, strError);
	if(!pGet)
	{
		context.SetError(strError.c_str());
		return false;
	}
	pGet->strScriptType = context.GetScriptType();
	pGet->strCallback = strCallback;

	// Add thread to pending list
	g_pState->vGets.push_back(pGet);

	// All done
	return true;
}

//=================================================================================================

HTTPGet* StartHTTPGet(const std::string& strAddress, const std::string& strFilename,
	HANDLE hTerminate, std::string& strError)
{
	// Open file if provided with a filename
	FILE* pFile = NULL;
	if(!strFilename.empty())
	{
		pFile = fopen(strFilename.c_str(), "wb");
		if(!pFile)
		{
			strError = "Failed to open file \"" + strFilename + "\" for writing";
			return NULL;
		}
	}

	// Create HTTP GET request
	HTTPGet* pGet = new HTTPGet;
	pGet->strAddress = strAddress;
	pGet->strContentFilename = strFilename;
	pGet->pContentFile = pFile;
	pGet->bHeaderComplete = false;
	pGet->nResponseCode = 0;
	pGet->nContentLength = 0;
	pGet->nBytesTransferred = 0;
	pGet->nTransferSpeed = 0;
	pGet->nUserData = 0;

	// Spawn thread
	HTTPGetThreadParam param;
	param.hTerminate = hTerminate;
	param.pGetRequest = pGet;
	param.hStarted = CreateEvent(NULL, FALSE, FALSE, NULL);
	DWORD dwThreadID;
	pGet->hThread = CreateThread(NULL, 0, HTTPGetThread, &param, 0, &dwThreadID);
	if(!pGet->hThread)
	{
		delete pGet;
		if(pFile)
			fclose(pFile);
		strError = "Failed to spawn worker thread for httpGet()";
		return NULL;
	}

	// Wait for the thread to signal it's started. Need to do this so we know
	// the worker thread has copied the HTTPGetThreadParam to it's own stack,
	// because the parameter passed in is about to go out of scope.
	WaitForSingleObject(param.hStarted, INFINITE);

	// Close the "started" event handle and return thread info
	CloseHandle(param.hStarted);
	return pGet;
}
