//============================================================================
// StandardLib.cpp - Standard library of functions
//============================================================================
#include "StandardLib.h"
#include "ScriptManager/ScriptManager.h"
#include "Common/Mutex.h"
#include <windows.h>

//============================================================================
// Function prototypes
bool Cmd_GetOS(ScriptContext& context, void*);
bool Cmd_GetTime(ScriptContext& context, void*);
bool Cmd_HTTPGet(ScriptContext& context, void*);
bool Cmd_SetIcon(ScriptContext& context, void*);
bool Cmd_SetWindowTitle(ScriptContext& context, void*);

//============================================================================
// Global state
// NOTE: Used in other standard lib compilation units
StandardLibState* g_pState;

//============================================================================

bool RegisterStandardLibrary(CScriptManager& sm)
{
	// Init standard lib state
	delete g_pState;
	g_pState = new StandardLibState;
	if(!g_pState) return false;
	g_pState->hThreadTerminate = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(!g_pState->hThreadTerminate) return false;

	ScriptFunction func;
	func.pParam = NULL;
	func.pfnFunction = Cmd_GetOS;
	strcpy(func.szName, "getOS");
	sm.Register(func);
	func.pfnFunction = Cmd_GetTime;
	strcpy(func.szName, "getTime");
	sm.Register(func);
	func.pfnFunction = Cmd_HTTPGet;
	strcpy(func.szName, "httpGet");
	sm.Register(func);
	func.pfnFunction = Cmd_SetIcon;
	strcpy(func.szName, "setIcon");
	sm.Register(func);
	func.pfnFunction = Cmd_SetWindowTitle;
	strcpy(func.szName, "setWindowTitle");
	sm.Register(func);
	return true;
}
//============================================================================

void TickStandardLibrary(CScriptManager& sm)
{
	// Check for HTTP GET completion
	for(size_t i=0; i<g_pState->vGets.size();)
	{
		HTTPGet* pGet = g_pState->vGets[i];

		// Thread is signalled when get complete (and thread has exited)
		if(WaitForSingleObject(pGet->hThread, 0) == WAIT_OBJECT_0)
		{
			// Get complete, call callback
			std::vector<ScriptVar> vArgs;
			std::vector<ScriptVar> vRetvals;
			vArgs.push_back(pGet->strAddress.c_str());
			vArgs.push_back(pGet->nResponseCode);
			vArgs.push_back(pGet->strHeader.c_str());
			if(pGet->pContentFile)
				vArgs.push_back(pGet->strContentFilename.c_str());
			else
				vArgs.push_back(pGet->strContent.c_str());
			vArgs.push_back(pGet->strError.c_str());
			bool bRet = sm.Call(pGet->strScriptType, pGet->strCallback, vArgs, vRetvals);
			if(!bRet)
			{
				sm.SetError("Failed to call callback function \"" + pGet->strCallback +
					"\" for httpGet(): " + sm.GetError());
			}

			// Remove this get
			CloseHandle(pGet->hThread);
			delete pGet;
			g_pState->vGets.erase(g_pState->vGets.begin()+i);

			// Return now if there was an error, so the text is saved. Poll other threads next tick
			if(!bRet)
				return;
		}
		else
			++i;
	}
}

//============================================================================

void CleanupStandardLibrary(CScriptManager&)
{
	if(!g_pState)
		return;

	// Signal the thread terminate event if needed
	if(!g_pState->vGets.empty())
	{
		// Signal thread termination
		SetEvent(g_pState->hThreadTerminate);

		// Wait for threads to exit
		std::vector<HANDLE> vHandles;
		vHandles.reserve(g_pState->vGets.size());
		for(size_t i=0; i<g_pState->vGets.size(); ++i)
			vHandles.push_back(g_pState->vGets[i]->hThread);
		DWORD dwRet = WaitForMultipleObjects((DWORD)vHandles.size(), &vHandles[0], TRUE, 5000);
		if(dwRet == WAIT_TIMEOUT)
		{
			// One or more threads not responding, force kill them
			for(size_t i=0; i<g_pState->vGets.size(); ++i)
			{
				if(WaitForSingleObject(vHandles[i], 0) != WAIT_OBJECT_0)
					TerminateThread(g_pState->vGets[i]->hThread, 0xffffffff);
			}
		}
	}

	// Cleanup handles
	if(g_pState->hThreadTerminate)
		CloseHandle(g_pState->hThreadTerminate);
	g_pState->hThreadTerminate = NULL;
	for(size_t i=0; i<g_pState->vGets.size(); ++i)
		CloseHandle(g_pState->vGets[i]->hThread);

	// Kill state
	delete g_pState;
	g_pState = 0;
}

//============================================================================
