//============================================================================
// TA2Stream.h - Stringstream-like class for composing TA2 tag strings
//============================================================================

#ifndef __TA2STRING_H__
#define __TA2STRING_H__

#include <string>

//============================================================================
// Stream class
class CTA2Stream
{
public:
	CTA2Stream();
	~CTA2Stream();

	struct Reset {};
	struct Bold {};
	struct Italic {};
	struct Underline {};
	struct Colour
	{
		Colour(unsigned int n)
		{
			if(n > 0xffffff) n = 0;
			sprintf(colour, "%06x", n);
		}

		char colour[7];
	};
	struct Strikeout {};

	friend CTA2Stream& operator<<(CTA2Stream& stream, const char* sz);
	friend CTA2Stream& operator<<(CTA2Stream& stream, const std::string& str);
	friend CTA2Stream& operator<<(CTA2Stream& stream, Reset bit);
	friend CTA2Stream& operator<<(CTA2Stream& stream, Bold bit);
	friend CTA2Stream& operator<<(CTA2Stream& stream, Italic bit);
	friend CTA2Stream& operator<<(CTA2Stream& stream, Underline bit);
	friend CTA2Stream& operator<<(CTA2Stream& stream, Colour bit);
	friend CTA2Stream& operator<<(CTA2Stream& stream, Strikeout bit);

	const std::string& str() const {return m_strBuffer;}

private:
	std::string	m_strBuffer;
};


//============================================================================
// Functions
// Convert tags to bytecode (E.g. "<-b>foobar<-0>" => "\xE0foobar\xBF")
std::string ParseTA2Tags(const std::string& str);


//============================================================================
// Stream functions
inline CTA2Stream& operator<<(CTA2Stream& stream, const char* sz)
{
	stream.m_strBuffer += sz;
	return stream;
}

inline CTA2Stream& operator<<(CTA2Stream& stream, const std::string& str)
{
	stream.m_strBuffer += str;
	return stream;
}

inline CTA2Stream& operator<<(CTA2Stream& stream, CTA2Stream::Reset)
{
	// If the buffer is empty, just chuck the bits in
	if(stream.m_strBuffer.empty())
	{
		stream.m_strBuffer += (unsigned char)(0x80 | 0x3e);
		return stream;
	}

	// Buffer isn't empty. Was the last byte the same thing?
	unsigned char& rChar = (unsigned char&)stream.m_strBuffer.end()[-1];
	if((rChar & 0xc0) != 0x80)
		stream.m_strBuffer += (unsigned char)(0x80 | 0x3e);
	return stream;
}

inline CTA2Stream& operator<<(CTA2Stream& stream, CTA2Stream::Bold)
{
	// If the buffer is empty, just chuck the bits in
	if(stream.m_strBuffer.empty())
	{
		stream.m_strBuffer += (unsigned char)(0xc0 | 0x20);
		return stream;
	}

	// Buffer isn't empty. Was the last byte the same thing?
	unsigned char& rChar = (unsigned char&)stream.m_strBuffer.end()[-1];
	if((rChar & 0xc0) == 0xc0)
		rChar |= 0x20;
	else
		stream.m_strBuffer += (unsigned char)(0xc0 | 0x20);
	return stream;
}

inline CTA2Stream& operator<<(CTA2Stream& stream, CTA2Stream::Italic)
{
	// If the buffer is empty, just chuck the bits in
	if(stream.m_strBuffer.empty())
	{
		stream.m_strBuffer += (unsigned char)(0xc0 | 0x10);
		return stream;
	}

	// Buffer isn't empty. Was the last byte the same thing?
	unsigned char& rChar = (unsigned char&)stream.m_strBuffer.end()[-1];
	if((rChar & 0xc0) == 0xc0)
		rChar |= 0x10;
	else
		stream.m_strBuffer += (unsigned char)(0xc0 | 0x10);
	return stream;
}

inline CTA2Stream& operator<<(CTA2Stream& stream, CTA2Stream::Underline)
{
	// If the buffer is empty, just chuck the bits in
	if(stream.m_strBuffer.empty())
	{
		stream.m_strBuffer += (unsigned char)(0xc0 | 0x08);
		return stream;
	}

	// Buffer isn't empty. Was the last byte the same thing?
	unsigned char& rChar = (unsigned char&)stream.m_strBuffer.end()[-1];
	if((rChar & 0xc0) == 0xc0)
		rChar |= 0x08;
	else
		stream.m_strBuffer += (unsigned char)(0xc0 | 0x08);
	return stream;
}

inline CTA2Stream& operator<<(CTA2Stream& stream, CTA2Stream::Colour bit)
{
	stream.m_strBuffer += (unsigned char)(0xc0 | 0x04);
	stream.m_strBuffer += bit.colour;
	return stream;
}

inline CTA2Stream& operator<<(CTA2Stream& stream, CTA2Stream::Strikeout)
{
	// If the buffer is empty, just chuck the bits in
	if(stream.m_strBuffer.empty())
	{
		stream.m_strBuffer += (unsigned char)(0xc0 | 0x02);
		return stream;
	}

	// Buffer isn't empty. Was the last byte the same thing?
	unsigned char& rChar = (unsigned char&)stream.m_strBuffer.end()[-1];
	if((rChar & 0xc0) == 0xc0)
		rChar |= 0x02;
	else
		stream.m_strBuffer += (unsigned char)(0xc0 | 0x02);
	return stream;
}

#endif // __TA2STRING_H__
