//============================================================================
// ScriptContextLua.h - Script context for Lua
//============================================================================
#include "ScriptContextLua.h"
#include <algorithm>
extern "C"
{
	#include "Lua/lua.h"
	#include "Lua/lualib.h"
}

struct DataChunk
{
	const void* pData;
	size_t nOffset;
	size_t nLength;
};

//============================================================================
// Exported function
SCRIPT_LINKAGE ScriptContextInterface* CreateScriptInterface(int nAPIVersion)
{
	if(nAPIVersion != SCRIPT_API_VERSION)
		return NULL;
	ScriptContextLua* pContext = new ScriptContextLua;
	if(pContext)
	{
		if(!pContext->Init())
		{
			delete pContext;
			return NULL;
		}
	}
	return pContext;
}

//============================================================================
// Lua allocator function
static void* Lua_Allocator(void*, void* ptr, size_t, size_t nSize)
{
	if(nSize == 0)
	{
		free(ptr);
		return NULL;
	}
	else
		return realloc(ptr, nSize);
}

//============================================================================

ScriptContextLua::ScriptContextLua() :
	m_pState(NULL),
	m_nRetvals(0),
	m_nArgs(0),
	m_bOwnState(true)
{
}

ScriptContextLua::ScriptContextLua(lua_State* pState) :
	m_pState(pState),
	m_nRetvals(0),
	m_bOwnState(false)
{
	m_nArgs = (size_t)lua_gettop(pState);
}

ScriptContextLua::~ScriptContextLua()
{
	for(size_t i=0; i<m_vScriptFunctions.size(); ++i)
		delete m_vScriptFunctions[i];
	m_vScriptFunctions.clear();

	if(m_pState && m_bOwnState)
		lua_close(m_pState);
}

//============================================================================
// ScriptContext functions
size_t ScriptContextLua::GetNumArgs() const
{
	return m_nArgs;
}

ScriptVar ScriptContextLua::GetArg(size_t nIdx) const
{
	if(nIdx >= m_nArgs)
		return ScriptVar(0);
	return GetLuaVar(m_pState, (int)nIdx+1);
}

void ScriptContextLua::AddRetval(const ScriptVar& var)
{
	PushVar(m_pState, var);
	++m_nRetvals;
}

void ScriptContextLua::AddArg(const ScriptVar& var)
{
	PushVar(m_pState, var);
	++m_nArgs;
}

void ScriptContextLua::SetError(const char* szError)
{
	if(!m_strError.empty()) m_strError += "\r\n";
	m_strError += szError;
}

const char* ScriptContextLua::GetScriptType()
{
	return "lua";
}

//============================================================================
// ScriptContextInterface:
void ScriptContextLua::Destroy()
{
	delete this;
}

ScriptLoadResult ScriptContextLua::LoadScript(const char* szScriptName)
{
	m_strError = "";

	// Make a better filename
	std::string scriptName = szScriptName; 
	std::transform(scriptName.begin(), scriptName.end(), scriptName.begin(), tolower);
	std::string strFilepath = scriptName;
	strFilepath = "scripts/" + strFilepath ;

	// Bail out if script is already loaded
	for(size_t i=0; i<m_vLoadedScripts.size(); ++i)
	{
		if(m_vLoadedScripts[i] == scriptName)
			return SLR_AlreadyLoaded;
	}

	// Open file
	FILE* pFile = fopen(strFilepath.c_str(), "r");
	if(!pFile)
	{
		m_strError = "Could not open file \""+strFilepath+"\" for reading";
		return SLR_FileNotFound;
	}

	// Call the chunkreader to read the file
	int nRet = lua_load(m_pState, StaticChunkReader, pFile, scriptName.c_str(), "bt");
	fclose(pFile);
	if(nRet != 0)
	{
		// Get the error from the top of the stack
		if(lua_gettop(m_pState) >= 1)
		{
			const char* szError = lua_tostring(m_pState, -1);
			lua_pop(m_pState, 1);
			if(szError)
				m_strError = szError;
		}
		return SLR_ScriptError;
	}

	// Run the chunk
	m_strError = "";
	nRet = lua_pcall(m_pState, 0, 0, 0);
	if(nRet != 0)
	{
		// Get the error from the top of the stack
		if(lua_gettop(m_pState) >= 1)
		{
			const char* szError = lua_tostring(m_pState, -1);
			lua_pop(m_pState, 1);
			if(szError)
				m_strError = szError;
		}
		else
			m_strError = "Could not run script\""+std::string(szScriptName)+"\"";
		return SLR_ScriptError;
	}

	// If the error string isn't empty, it failed
	if(!m_strError.empty())
		return SLR_UnknownError;

	// Add to list
	m_vLoadedScripts.push_back(scriptName);
	return SLR_Success;
}

bool ScriptContextLua::LoadScript(const void* pData, size_t nLen)
{
	DataChunk chunk;
	chunk.pData = pData;
	chunk.nLength = nLen;
	chunk.nOffset = 0;

	// Call the chunkreader to read the file
	int nRet = lua_load(m_pState, StaticDataReader, &chunk, "", "bt");
	if(nRet != 0)
	{
		// Get the error from the top of the stack
		if(lua_gettop(m_pState) >= 1)
		{
			const char* szError = lua_tostring(m_pState, -1);
			lua_pop(m_pState, 1);
			if(szError)
				m_strError = szError;
		}
		return false;
	}

	// Run the chunk
	m_strError = "";
	nRet = lua_pcall(m_pState, 0, 0, 0);
	if(nRet != 0)
	{
		// Get the error from the top of the stack
		if(lua_gettop(m_pState) >= 1)
		{
			const char* szError = lua_tostring(m_pState, -1);
			lua_pop(m_pState, 1);
			if(szError)
				m_strError = szError;
		}
		return false;
	}

	// If the error string isn't empty, it failed
	if(!m_strError.empty())
		return false;
	return true;
}

//============================================================================

bool ScriptContextLua::ReloadScripts()
{
	// Make a copy of all loaded scripts, clear the list with loaded scripts and then reload them
	std::vector<std::string> prevScripts = m_vLoadedScripts;
	m_vLoadedScripts.clear();

	// Add autoexec to the scripts list, if it's not there
	bool bFound = false;
	for(size_t i=0; i<prevScripts.size(); ++i)
	{
		if(prevScripts[i] == "autoexec.lua")
		{
			bFound = true;
			break;
		}
	}
	if(!bFound)
		prevScripts.push_back("autoexec.lua");

	// Load each script
	std::string strError;
	for(size_t i=0; i<prevScripts.size(); ++i)
	{
		if (LoadScript(prevScripts[i].c_str()) != SLR_Success)
		{
			if(!strError.empty()) strError += "\r\n";
			strError += m_strError;
		}
	}
	if(!strError.empty())
		m_strError = strError;
	return m_strError.empty();
}

bool ScriptContextLua::Register(LPScriptFunction pfnFunc, void* pParam, const char* szName)
{
	// Check for bad params
	if(!pfnFunc || !szName || !szName[0])
	{
		m_strError = "Bad parameter";
		return false;
	}
	if(strlen(szName) >= ms_nMaxScriptNameLen)
	{
		m_strError = "Function name too long";
		return false;
	}

	// Copy to internal structure
	ScriptFunction* pFunction = new ScriptFunction;
	pFunction->pfnFunction = pfnFunc;
	pFunction->pParam = pParam;
	strcpy(pFunction->szName, szName);
	pFunction->pThis = this;

	// Register with lua
	lua_pushlightuserdata(m_pState, pFunction);
	lua_pushcclosure(m_pState, StaticFunctionStub, 1);
	lua_setglobal(m_pState, pFunction->szName);

	// Save function and return
	m_vScriptFunctions.push_back(pFunction);
	return true;
}

bool ScriptContextLua::Unregister(const char* szFuncName)
{
	for(size_t i=0; i<m_vScriptFunctions.size(); ++i)
	{
		if(strcmp(m_vScriptFunctions[i]->szName, szFuncName) == 0)
		{
			lua_pushnil(m_pState);
			lua_setglobal(m_pState, m_vScriptFunctions[i]->szName);

			delete m_vScriptFunctions[i];
			m_vScriptFunctions.erase(m_vScriptFunctions.begin()+i);
			return true;
		}
	}
	return false;
}

bool ScriptContextLua::Call(const char* szFunction, size_t nArgs, const ScriptVar* pArgs,
							ScriptVar** ppRetvals, size_t& rnRetvals)
{
	// Safety
	if(*ppRetvals)
		*ppRetvals = NULL;
	rnRetvals = 0;

	// Get function and push arguments
	int nStartSize = lua_gettop(m_pState);
	lua_getglobal(m_pState,szFunction);
	for(size_t i=0; i<nArgs; ++i)
		PushVar(m_pState, pArgs[i]);

	// Call the function
	int nRet = lua_pcall(m_pState, (int)nArgs, LUA_MULTRET, 0);
	if(nRet != 0)
	{
		const char* szError = lua_tostring(m_pState, -1);
		if(szError) m_strError = szError;
		else m_strError = "Unknown error";
		lua_pop(m_pState, 1);
		return false;
	}

	// Convert from Lua var to ScriptVar
	int nRetvals = lua_gettop(m_pState) - nStartSize;
	m_vReturnValues.clear();
	m_vReturnValues.reserve(nRetvals);
	for(int i=0; i<nRetvals; ++i)
		m_vReturnValues.push_back(GetLuaVar(m_pState, nStartSize+i+1));
	if(ppRetvals)
	{
		if(nRetvals != 0)
			*ppRetvals = &m_vReturnValues[0];
		else
			*ppRetvals = NULL;
	}
	rnRetvals = (size_t)nRetvals;

	// Pop return values and return
	lua_pop(m_pState, nRetvals);
	return true;
}

const char* ScriptContextLua::GetError() const
{
	return m_strError.c_str();
}

//========================================================================
// ScriptContextLua:
bool ScriptContextLua::Init()
{
	// Create lua state
	m_pState = lua_newstate(Lua_Allocator, NULL);
	if(!m_pState)
		return false;

	// Open misc libs
	luaL_openlibs(m_pState);

	// Register hang detection callback
	// 100 million opcodes should be enough for anyone...
	lua_sethook(m_pState, StaticHookProc, LUA_MASKCOUNT, 100000000);

	// Done
	return true;
}

//============================================================================
// Internal functions
int ScriptContextLua::StaticFunctionStub(lua_State* pState)
{
	// Get the function
	int nIdx = lua_upvalueindex(1);
	ScriptFunction* pFunction = (ScriptFunction*)lua_topointer(pState, nIdx);
	if(!pFunction)
	{
		// Lua longjmp's! Will never return
		lua_pushstring(pState, "Internal error: Function does not exist");
		lua_error(pState);
		return 0;
	}

	// Create a new context for the function
	// Needs to be shielded in a block due to lua's longjmp fun
	int nRetvals;
	bool bRet;
	{
		ScriptContextLua context(pState);

		// Call it
		bRet = pFunction->pfnFunction(context, pFunction->pParam);

		// Get number of return values
		nRetvals = context.m_nRetvals;

		// Store any error in ourself
		if(!bRet)
			pFunction->pThis->m_strError = context.GetError();
	}

	// Return number of arguments or error out
	if(!bRet)
	{
		{
			std::string strFunc(pFunction->szName);
			const std::string& strScriptError = pFunction->pThis->m_strError;
			std::string strError;
			if(!strScriptError.empty())
				strError = "Host function \"" + strFunc + "\" returned an error: " + strScriptError;
			else
				strError = "Host function \"" + strFunc + "\" returned an error";
			lua_pushstring(pState, strError.c_str());
			pFunction->pThis->m_strError = "";
		}

		// Lua longjmp's! Will never return
		lua_error(pState);
	}
	return nRetvals;
}

void ScriptContextLua::StaticHookProc(lua_State* _pState, lua_Debug* _pDebug)
{
	// Get stack
	{
		std::string error = "Hang detected!\nStack:\n";
		for(int i=0; i<16; ++i)
		{
			if(!lua_getstack(_pState, i, _pDebug))
				break;
			lua_getinfo(_pState, "nlS", _pDebug);

			char line[512];
			sprintf(line, " %s [%s]", _pDebug->name ? _pDebug->name : "<unknown>",
				_pDebug->what);
			error += line;
			if(_pDebug->source && _pDebug->currentline >= 0)
			{
				sprintf(line, " (%s:%d)\n", _pDebug->source, _pDebug->currentline);
				error += line;
			}
			else
				error += "\n";
		}
		if(!error.empty())
			error.erase(error.end()-1);
		lua_pushstring(_pState, error.c_str());
	}

	lua_error(_pState);
}

ScriptVar ScriptContextLua::GetLuaVar(lua_State* pState, int nIdx)
{
	int nType = lua_type(pState, nIdx);
	switch(nType)
	{
		case LUA_TBOOLEAN:
		{
			return lua_toboolean(pState, nIdx);
		}

		case LUA_TNUMBER:
		{
			return (float)lua_tonumber(pState, nIdx);
		}

		case LUA_TSTRING:
		{
			return lua_tostring(pState, nIdx);
		}

		case LUA_TTABLE:
		{
			int nCount = (int)lua_rawlen(pState, nIdx);
			std::vector<ScriptVar> ret;
			ret.reserve(nCount);
			for(int i=1; i<=nCount; ++i)
			{
				lua_rawgeti(pState, nIdx, i);
				ret.push_back(GetLuaVar(pState, -1));
				lua_pop(pState, 1);
			}
			if(ret.empty())
				return ScriptVar(ScriptVar::TYPE_Array);
			return ScriptVar(&ret[0], ret.size());
		}
	}

	return 0;
}

void ScriptContextLua::PushVar(lua_State* pState, const ScriptVar& var)
{
	switch(var.GetType())
	{
		case ScriptVar::TYPE_Float:
		{
			lua_pushnumber(pState, (lua_Number)var.ToFloat());
		}
		break;

		case ScriptVar::TYPE_Int:
		{
			lua_pushnumber(pState, (lua_Number)var.ToInt());
		}
		break;

		case ScriptVar::TYPE_String:
		{
			std::string strVal = var.ToString();
			lua_pushstring(pState, strVal.c_str());
		}
		break;

		case ScriptVar::TYPE_Array:
		{
			lua_newtable(pState);
			for(size_t i=0; i<var.GetArraySize(); ++i)
			{
				PushVar(pState, var.GetArrayItem(i));
				lua_rawseti(pState, -2, (int)i+1);
			}
		}
		break;
	}
}

const char* ScriptContextLua::StaticChunkReader(lua_State*, void* pData, size_t* pSize)
{
	static char szBuffer[4096];
	FILE* pFile = (FILE*)pData;
	size_t nRet = fread(szBuffer, 1, sizeof(szBuffer), pFile);
	*pSize = nRet;
	return szBuffer;
}

const char* ScriptContextLua::StaticDataReader(lua_State*, void* pData, size_t* pSize)
{
	DataChunk* pChunk = (DataChunk*)pData;
	size_t nLen = pChunk->nLength - pChunk->nOffset;
	pChunk->nOffset = pChunk->nLength;
	*pSize = nLen;
	return (const char*)pChunk->pData;
}

//============================================================================
